/*
 * JBoss, the OpenSource J2EE webOS
 * 
 * Distributable under LGPL license. See terms of license at gnu.org.
 */
package org.jboss.mail;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import org.jboss.logging.Logger;

/**
 * Abstract Protocol provides a configurable stateful abstraction for
 * protocols. State is the state of an instance. Properties are
 * configurationation parameters.
 * 
 * @author Andrew C. Oliver
 */
public abstract class AbstractProtocol implements Protocol {
    private static final Logger log = Logger.getLogger(AbstractProtocol.class);
    /**
     * the serverThread we're assigned to (we may need to release it)
     */
    private ServerThread serverThread;
    /**
     * State is the temporary status of this instance.
     */
    private Map state;
    /**
     * properties are setup parameters...things that don't ususally change very
     * often...like the server name..
     */
    protected Map properties;

    /**
     * determines if the connection is secure, that is, the SSL handshake has completed.
     * This happens after the client says STARTTLS
     */
    protected boolean secure = false;
    
    /**
     * Constructs an abstract protocol with empty properties and state.
     *  
     */
    public AbstractProtocol() {
        this.state = new HashMap();
        this.properties = new HashMap();
    }

    /**
     * Constructs an abstract protocol with the passed in configuration
     * properties
     * 
     * @param properties
     *            which govern this protocol isntance's configuration
     *            (generally systemwide)
     */
    public AbstractProtocol(Map properties) {
        this.state = new HashMap();
        this.properties = properties;
    }

    /**
     * Retrieve a configuration property
     * 
     * @param name
     *            of the property
     * @return property
     */
    public Object getProperty(String name) {
        return properties.get(name);
        //     return "property named "+name+" would be here";
    }

    /**
     * retrieve a configuration property as a long primitive
     * 
     * @param name
     *            of the property to return
     * @return the property as a primitive long value
     */
    public long getPropertyLong(String name) {
        Object val = properties.get(name);
        long lval = 0;
        if (val != null && val instanceof Long) {
            lval = ((Long) val).longValue();
        }
        return lval;
    }

    /**
     * retrieve a configuration property as a boolean primitive
     * 
     * @param name
     *            of the property to return
     * @return the property as a primitive boolean value
     */
    public boolean getPropertyBool(String name) {
        Object val = properties.get(name);
        boolean bval = false;
        if (val != null && val instanceof Boolean) {
            bval = ((Boolean) val).booleanValue();
        }
        return bval;
    }

    /**
     * retrieve a state value for this protocol instance (generally specific
     * conversational state)
     * 
     * @param name
     *            of state value
     * @return state value
     */
    public Object getState(String name) {
        return state.get(name);
    }

    /**
     * set a state variable.
     * 
     * @param name
     *            of state variable
     * @param value
     *            of state variable
     */
    public void setState(String name, Object value) {
        state.put(name, value);
    }

    /**
     * reset the conversational state (ditch all state variables)
     */
    public void resetState() {
        state = new HashMap();
    }

    /**
     * set the configuration (should only be done upon construction)
     * 
     * @param properties
     *            map containing key value pairs
     */
    public void setProperties(Map properties) {
        this.properties = properties;
    }

    /**
     * Reads the input stream until it encounters endline charecters, or the
     * input stream closes.
     * 
     * @param stream
     *            InputStream to read
     * @return input command without endline characters.
     * @throws IOException
     */
    protected String readCommand(InputStream stream) throws IOException {
        boolean firstchar = true;
        byte[] buffer = new byte[1];
        StringBuffer commstring = new StringBuffer();
        while (stream.read(buffer) != -1) {
            char c = (char) buffer[0];
            //System.out.print(HexDump.toHex(buffer[0]));
            commstring.append(c);
            String check = commstring.substring(commstring.length() - 1);
            //            System.out.print(check);
            if (check.equals("\n") || check.equals("\r")) {
                if (firstchar == false) {
                    break;
                } else {
                    commstring.deleteCharAt(commstring.length() - 1);
                }
            } else {
                firstchar = false;
            }
        }
        stream.mark(2);
        if (log.isDebugEnabled()) {
            log.debug("ReadCommand outputting " + commstring);
        }
        return commstring.toString();
    }

    /**
     * seperates the argument from the command
     * 
     * @param commandstring
     *            argument + command
     * @return string array with the first element being the command,
     *         subsequent (always only 1) being the arguments
     */
    protected String[] parseCommand(String commandstring) {
        String command = commandstring;
        String argument = null;
        int spaceIndex = command.indexOf(" ");
        if (spaceIndex > 0) {
            argument = command.substring(spaceIndex + 1);
            command = command.substring(0, spaceIndex);
        }
        command = command.toUpperCase(Locale.US);
        return new String[] { command, argument };
    }

    /**
     * @return ServerThread associated with this instance (we'll need to release it in the pool)
     */
    public ServerThread getServerThread() {
        return serverThread;
    }
    /**
     * used by the pool to associate our serverThread instance
     * @param serverThread associated with this instance
     */
    public void setServerThread(ServerThread serverThread) {
        this.serverThread = serverThread;
    }

    /**
         * @return whether or not we're using TLS
         */
    public boolean isSecure() {
        return secure;
    }

    /**
     * whether to use TLS
     * @param secure use TLS or not
     */
    public void setSecure(boolean secure) {
        this.secure = secure;
    }

}
