/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.nio.channels.SocketChannel;
import java.util.Map;

/**
 * The protocol is an abstraction for attaching bi-directional stateful protocols to the
 * "Server".  The two most important artifacts of this contract are the properties and 
 * states as well as the difference.  Properties are intended to be passed in upon construction
 * of a protocol instance, if the instance is pooled then these are only passed in upon construction
 * or in the event of a server reconfiguration of some kind.  State variables are used throughout
 * the lifecycle of a user connection and change or are cleared as needed.  These represent the 
 * conversational state with the client and are much like session variables.
 * 
 * @author Andrew C. Oliver
 */
public interface Protocol {

    /**
     * @return get the key name for the protocol
     */
    String getName();

	/**
	 * greet the client upon connection
	 * @param stream output stream of the client
	 * @throws IOException in the event the OutputStream cannot be written to for some reason
	 */
	void greet(OutputStream stream) throws IOException;
	
	/**
	 * greet the client upon connection (NIO version)
	 * @param ioChannel
	 * @throws IOException in the event the OutputStream cannot be written to for some reason
	 */
	void greet(SocketChannel ioChannel) throws IOException;

	/**
	 * parse an incoming request by pulling the "command" off of the input stream and constructing 
	 * a request object of some kind.
	 * @param stream the input stream to process the request from
	 * @param socket the socket to retrieve user information from (remote address, etc)
	 * @return Request object representing the incomming request.
	 * @throws IOException denoting a failure to read the request from the input stream
	 */
	Request parseRequest(InputStream stream, Socket socket) throws IOException;
	
	/**
	 * handle a request including any subsequent conversation required on the output stream then 
	 * construct a response object representing any information about the process.
	 * @param stream OutputStream in which to contact the client.
	 * @param request representing the protocol command from the client. 
	 * @return Response object providing information about the processing
	 */
	Response handleRequest(OutputStream stream, Request request);
	
	/**
	 * In the event an IOError is thrown, the server will delegate teh "what to do" to this method.
	 * It cannot assume that the outputstream is valid and cannot throw an exception (it needs to handle
	 * the error possibly by relinquishing resources and giving up)
	 * @param stream in which to contact the client 
	 * @return Response indicating any information about processing the error.
	 */
	Response handleIOError(OutputStream stream);
	
	/**
	 * retrieve a configuration property
	 * @param name of the property
	 * @return value of property
	 */
	Object getProperty(String name);
	
	/**
	 * retrieve a numeric configuration property
	 * @param name of the configuration property
	 * @return long primitive value
	 */
	long getPropertyLong(String name);
	
	/**
	 * retrieve a boolean configuration property
	 * @param name of the configuration property
	 * @return boolean primitive value
	 */
	boolean getPropertyBool(String name);
	
	/**
	 * retrieve a conversational state variable (session variable) 
	 * @param name of the state variable
	 * @return value of the variable
	 */
	Object getState(String name);
	
	/**
	 * set the value of a conversational state variable (session variable)
	 * @param name of the variable
	 * @param value of the variable
	 */
	void setState(String name, Object value);
	
	/**
	 * reset all state to nothing
	 */
	void resetState();
	
	/**
	 * set configruation information for this protocol (should only need to do this on construction or 
	 * server reconfig)
	 * @param properties
	 */
	void setProperties(Map properties);


}
