/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;

import org.jboss.mail.util.JBMailJMXUtil;

import org.apache.log4j.NDC;
import org.apache.log4j.Logger;

/**
 * ServerThread is a pooled handler which actually does the work for a connection.
 * @author Andrew C. Oliver
 */
public class ServerThread implements Runnable {

    Logger logger = Logger.getLogger(ServerThread.class);

    // state attached to a protocol when a client connects
    public static final String STATE_CLIENT_ADDRESS = "client address";
    public static final String STATE_LOCAL_ADDRESS = "local address";

    // instance and invocation statistics:
    private static Long instances = new Long(0);
    private long id, clientsHandled;
    
    private Socket socket;
    private InputStream input;
    private OutputStream output;
    private Protocol protocol;
    private String protocolName;
    private long timeout;
    private long life;

    private Thread myThread;
    private ServerThreadPool pool;

    public long getId() {
	return id;
    }

    public long getClientsHandled() {
	return clientsHandled;
    }

    /**
     * @return ServerThreadPool 
     */
    public ServerThreadPool getPool() {
        return pool;
    }

    /**
     * sets up a callback for the pool containing this thread
     * @param pool
     */
    protected void setPool(ServerThreadPool pool) {
        this.pool = pool;
    }

    /**
     * @return java.lang.Thread instance presently assigned to this ServerThread
     */
    protected Thread getMyThread() {
        return myThread;
    }

    /**
     * This sets the underlying Java.lang.thread for this serverThread (done by the ServerThreadPool)
     * @param myThread java.lang.Thread instance
     */
    protected void setMyThread(Thread myThread) {
        this.myThread = myThread;
    }

    /**
     * @return input stream assigned to this ServerThread
     */
    protected InputStream getInput() {
        return input;
    }

    /**
     * sets the InputStream for the thread (ServerThreadPool)
     * @param input stream for this thread
     */
    protected void setInput(InputStream input) {
        this.input = input;
    }

    /**
     * @return life for this ServerThread - we'll kill it and return it to the pool when its life is up
     */
    protected long getLife() {
//        System.out.println("currtime="+ System.currentTimeMillis() +" life="+life + " diff="+(System.currentTimeMillis()-life));
        return life;
    }

    /**
     * set how long we want this thread to live before we kill it and return it to the library.
     * @param life milliseconds to live regardless of operation
     */
    protected void setLife(long life) {
        this.life = life;
    }

    /**
     * @return out put stream for this ServerThread
     */
    protected OutputStream getOutput() {
        return output;
    }

    /**
     * @param output stream for this ServerThread
     */
    protected void setOutput(OutputStream output) {
        this.output = output;
    }

    /**
     * @return Protocol instance presently assigned for this ServerThread
     */
    protected Protocol getProtocol() {
        return protocol;
    }

    /**
     * ServerThread is responsible for looking up the Protocol instance.  This is a performance
     * optimization.  If Server did it then he'd get behind on answering the connections and passing
     * them off.
     * @param protocol to lookup
     */
    protected void setProtocol(String protocolName) {
        this.protocolName = protocolName;
    }

    /**
     * @return Socket assigned to this ServerThread
     */
    public Socket getSocket() {
        return socket;
    }

    /**
     * set the socket to assign to this ServerThread
     * @param socket to assing to thsi ServerThread
     */
    public void setSocket(Socket socket) throws IOException {
        this.socket = socket;
        input = new BufferedInputStream(socket.getInputStream());
        input.mark(2);
        output = socket.getOutputStream();

    }

    /**
     * @return timeout (milliseconds) between operations
     */
    public long getTimeout() {
        return timeout;
    }

    /**
     * set the time in ms between operations.  This is the time between when an operation has 
     * completed and another one is initiated.  Meaning if I say HELO localhost and the command
     * takes 10 minutes to return, provided that "life" is 11 minutes...This takes no effect..  If it
     * returns and I sit around for 10 minutes and the timeout is set to 5...then after 5 minutes of 
     * me being idle...the connection is killed and the ServerThread is returned to the pool...and 
     * thanks for playing.
     * @param timeout between operations (in milliseconds)
     */
    public void setTimeout(long timeout) {
        this.timeout = timeout;
    }

    /**
     * construct a ServerThread (done by the ServerThreadPool at startup)
     * @param socket to handle the connection...
     * @param protocolName is the JMX name (or classname) to look up and use as the Protocol...
     * @param timeout time between operations (ms)
     * @param life time to live (ms)
     * @param pool callback for the ServerThreadPool which contains this...
     */
    public ServerThread(
        Socket socket,
        String protocolName,
        long timeout,
        long life,
        ServerThreadPool pool) {
            
	this();

        this.socket = socket;
        this.protocolName = protocolName;
        this.timeout = timeout;
        this.life = life;
        this.pool = pool;
        
    }
    
    /**
     * an inner class that expresses the identity of a particular invocation of a server thread
     */
    static public class NDCIdentity {
	
	private long threadId, clientsHandled;
	private Socket socket;
	
	public NDCIdentity(ServerThread thread) {
	    this.threadId = thread.getId();
	    this.clientsHandled = thread.getClientsHandled();
	    this.socket = thread.getSocket();
	}

	public String toString() {
	    return "addr="+socket.getInetAddress().getHostAddress()+",thread=" + threadId + ",clients=" + clientsHandled;
	}
	
    }

    /**
     * default constructor...use the setters before starting the thread up..
     */
    public ServerThread() {
	synchronized (instances) {
	    id = instances.longValue();
	    instances = new Long(instances.longValue() + 1);
	    clientsHandled = 0;
	}
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run()
     */
    public void run() {
	synchronized (instances) {
	    clientsHandled++;
	}

        try {

	    // use NDC's for conversational logging; *MUST* call NDC.remove() sometime later
	    // to reclaim resources... [log4j warns about this, although NDC.remove()'s need
	    // not be called in a 1-1 correspondence with NDC.push()'s]
	    NDC.push(new NDCIdentity(this).toString());
 
            this.protocol = (Protocol) JBMailJMXUtil.getService(protocolName);
	    
	    // set state of socket
	    protocol.setState(STATE_CLIENT_ADDRESS,socket.getInetAddress());
	    protocol.setState(STATE_LOCAL_ADDRESS,socket.getLocalAddress());
	    
            input = new BufferedInputStream(socket.getInputStream());
            output = socket.getOutputStream();
            protocol.greet(output);
            ((AbstractProtocol)protocol).setServerThread(this);
            boolean loop = true;
            input.mark(2);            
            while (loop && input.read() != -1 && life > System.currentTimeMillis()) {
                input.reset();
                // create Request object and parse
                Request request = protocol.parseRequest(input, socket);
                input.mark(2); //we've got our command...mark our new spot

                // create Response object
                Response response = protocol.handleRequest(output, request);
                input.mark(2);
                loop = !response.isFinish();
		if (loop) {
		    // MIKEA: avoid delay on SMTP QUIT, etc. (not sure why we need delay anyway?)
		    Thread.sleep(1000);
		}
            }
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            try {
                pool.free(this);
                this.socket.close();
            } catch (IOException ioe) {
                ioe.printStackTrace();
            } finally {
		NDC.remove();
	    }
        }

    }

}
