/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail;

import java.net.Socket;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.io.IOException;

import org.jboss.logging.Logger;

/**
 * ServerThreadPool contains a set number of instances of ServerThread.  It handles the real work
 * for Server.  Its also responsible for cleaning things up.
 * @author Andrew C. Oliver
 */
public class ServerThreadPool implements Runnable {
    private static final Logger log = Logger.getLogger(ServerThreadPool.class);
    /**
     * list of threads available and not in use
     */
    private List availthreads;
    /**
     * list of threads presently in use
     */
    List usedthreads;
    /**
     * any time we sleep (say in the reaper thread) this is how long we sleep for
     */
    private long interval;
    /**
     * false for debugging purposes...true for normal operation.  We want to pool...pooling is good
     * so set this to true... False means we start a new thread each time.
     */
    private boolean pool;
    /**
     * number of threads in the pool
     */
    private int numthreads;

    /**
     * create a new ServerThreadPool
     * @param pooled true or false --true is good...false is bad...don't set this to true unless you're really smart and are hacking on this...deal?
     * @param num of threads we want in the pool
     * @param interval how long to take a coffee break between reaping unsuspecting threads (they deserve it...they're dead on the inside and don't know it)
     */
    protected ServerThreadPool(boolean pooled, int num, long interval) {
        this.pool = pooled;
        if (pooled) {
            availthreads = Collections.synchronizedList(new ArrayList(num));
            this.numthreads = num;
            createThreads();
        }
        usedthreads = Collections.synchronizedList(new ArrayList(num));
        this.interval = interval;      
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run()
     */
    public void run() {
        try {
            //iterators are bad here....don't use them... long story...concurrentModification and all of that...
	    //
	    // MIKEA'S NOTE: WHETHER OR NOT WE GET concurrentModification EXCEPTIONS FROM ITERATOR IS NOT
	    // THE MAIN PROBLEM --- ITS THAT WE DON'T SYNCHRONIZE ON usedthreads TO START WITH!
	    // THUS, IF WE DON'T USE ITERATORS, THAT MAY JUST BE MASKING A REAL CONCURRENCY PROBLEM SINCE THE 
	    // FOR-LOOP TENDS TO AVOID THROWING THAT EXCEPTION... FOR INSTANCE, AFTER EVALUATING
	    // THE 'k < usedthreads.size()', but before evaluating 'usedthreads.get(k)', another
	    // thread may remove an element and thus the 'get' method will may throw an IndexOutOfBoundsException.
	    // thus, we merely replace a concurrentModification exception with a slightly less 
	    // likely IndexOutOfBoundsException (less likely by a factor of usedthreads.size())
	    //
            while (true) {
                Thread.sleep(interval);
                //System.out.println("availthreads = "+this.availthreads.size()+" usedthreads="+this.usedthreads.size());
                for (int k = 0; k < usedthreads.size(); k++) {
                    ServerThread thread = (ServerThread) usedthreads.get(k);
                    if (thread.getLife() < System.currentTimeMillis()) {
                        cleanThread(k, thread);
                    }

                }
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
            for (int k = 0; k < usedthreads.size(); k++) {
                ServerThread st = (ServerThread) usedthreads.get(k);
              cleanThread(k, st);
            }
        }
    }

    /**
     * create the threads to stick in our pool...  Done when we start up..
     */
    protected void createThreads() {
        for (int k = 0; k < this.numthreads; k++) {
            ServerThread st = new ServerThread();
            Thread thread = new Thread(st);
            st.setMyThread(thread);
            availthreads.add(st);
        }
    }

    /**
     * assign a thread to handle a socket connection
     * @param socket to handle
     * @param protocolName to look up
     * @param timeout between operations (ms)
     * @param life of the thread (ms)
     * @return Thread instance containing our ServerThread
     * @throws InterruptedException
     */
    protected synchronized Thread assign(
        Socket socket,
        String protocolName,
        long timeout,
        long life)
        throws InterruptedException, IOException {
        Thread thread =
            this.pool
                ? pooledThread(socket, protocolName, timeout, life)
                : newThread(socket, protocolName, timeout, life);
        return thread;
    }

    /**
     * Used for debugging purpsoses...peer to pooledThread...creates a new ServerThread instead  of
     * pooling...no limit!  meaning this would be AWESOME for DoS attacks!...don't use this mode unless
     * you're debugging the SMTP server or something
     * @param socket to handle 
     * @param protocolName to lookup
     * @param timeout between operations
     * @param life of the connection
     * @return new java.lang.Thread instance
     */
    private Thread newThread(
        Socket socket,
        String protocolName,
        long timeout,
        long life) {
        ServerThread serverthread =
            new ServerThread(
                socket,
                protocolName,
                timeout,
                life,
                this);
        Thread thread = new Thread(serverthread);
        serverthread.setMyThread(thread);
        return thread;
    }

    /**
     * get a pooledThread to handle the connection
     * @param socket to handle
     * @param protocolName to lookup
     * @param timeout between operations
     * @param life of the whole thing
     * @return a new java.lang.Thread assigned to a pooled ServerThread
     * @throws InterruptedException
     */
    private Thread pooledThread(
        Socket socket,
        String protocolName,
        long timeout,
        long life)
        throws InterruptedException, IOException {
        awaitFreeThread();
        ServerThread serverthread = null;
        synchronized (availthreads) {
            serverthread =
                (ServerThread) availthreads.remove(availthreads.size() - 1);
        this.usedthreads.add(serverthread);
        }
        serverthread.setSocket(socket);
        serverthread.setProtocol(protocolName);
        serverthread.setLife(life);
        serverthread.setTimeout(timeout);
        serverthread.setPool(this);
        return serverthread.getMyThread();
    }

    /**
     * if we've run out of availthreads then we'll have to hang out until one is available...
     */
    private void awaitFreeThread() throws InterruptedException {
        while (availthreads.size() < 1) {
            Thread.sleep(interval);
        }
    }

    /**
     * cleans a thread by index.
     * @param index number in the usedthreads...
     * @param thread instance to clean
     */
    private void cleanThread(int index, ServerThread thread) {
        //i.remove();
        ServerThread st = (ServerThread) usedthreads.remove(index);
        Thread javathread = thread.getMyThread();
        javathread.interrupt();
       // availthreads.add(thread);
        if (this.pool == true) {
            Thread mt = new Thread(thread);
            thread.setMyThread(mt);
            availthreads.add(st);
        }
    }

    /**
     * frees a thread and returns it to the pool...
     * @param thread to clean
     */
    public void free(ServerThread thread) {
        log.debug("freeing " + thread);
        boolean success = usedthreads.remove(thread);
        if (pool && success) {
            Thread mt = new Thread(thread);
            thread.setMyThread(mt);
            availthreads.add(thread);
        }
    }
}
