/*
 * JBoss, the OpenSource J2EE webOS
 * 
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.dev;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.jboss.mail.Base64;

/**
 * Central concept to the Script runner.  Represents the first S:to last C:xxEND.
 * Its based on a composite pattern of sorts.  One condition contains subconditions for the various
 * branches.  C: clauses can be multiline, S: clauses presently cannot.  
 * @author Andrew C. Oliver
 */
public class Condition {
    List subconditions;
    String condition;
    String response;
	
    /**
     * Compose a condition from a group of lines.
     * @param group list containing lines from the file. Only those lines belonging to this condition.
     */
	public Condition(List group) {
		Iterator i = group.iterator();
		boolean first = true;
		boolean hasSubConditions = false;
		String[] last = null;
		int lineno = 0;
		while (i.hasNext()) {
			String[] pieces = parseLine((String)i.next());
			if (first) {
				if (!pieces[0].equals("S")) {
					throw new RuntimeException("first in group must be S: or Sx:");
				}
				if (pieces[1] != null) {
					if (!pieces[1].equals("1")) {
						throw new RuntimeException("first condition must be S1:");
					}
					hasSubConditions = true;
					this.subconditions = new ArrayList();
					this.condition = pieces[2];
				} else {
					hasSubConditions = false;
					this.condition = pieces[2];
					String[] res = parseLine((String)i.next());
					if (!res[0].equals("C")) {
						throw new RuntimeException("an S must be followed by a C or have a branch #");
					}
					if (res[1]!= null) {
						throw new RuntimeException("if an S has no branch number then no C can have a branch number");
					}
					this.response = parseFunctions(res[2]);
					
					if (i.hasNext()) {
						throw new RuntimeException("Un branched groups cannot have more than 2 elements");
					}
				}
			} else {
				if (pieces[0].equals("S")) {
					Condition sub = null;
					if (pieces[1] == null || pieces[1].equals("1")) {
						throw new RuntimeException(">1 element in the group must not be 1 or null");
					}
					
					last = (last == null) ? parseLine((String)group.get(group.size()-1)) : last;
					if (!last[0].equals("C")){
						throw new RuntimeException("The last thing must be a C (response)");
					} 
					
					if (last[1] == null) {
						this.response = parseFunctions(last[2]); //also set the first condition to have the same response
						sub = new Condition(pieces[2], last[2]);
						if (lineno == group.size()-2) {
							i.next();
						}
					} else {
						String[] next = parseLine((String)i.next());
						if (!next[0].equals("C")) {
							throw new RuntimeException("Either the last response must have no branch number or "+
									                    "each conditional must have a response under it");
						}
						if (next[1] == null ||!pieces[1].equals(next[1])) {
							throw new RuntimeException(">1 response must match the conditional above it");
						}
						sub = new Condition(pieces[2], next[2]);
					}
					this.subconditions.add(sub);
				} else if (pieces[0].equals("C") && lineno == 1){
				  if (pieces[1] != null && !pieces[1].equals("1")) {
				  	throw new RuntimeException("if Cx follows S1 then it must be C1");
				  }
				  response = parseFunctions(pieces[2]);
				} else {
					throw new RuntimeException("Two Cs can not follow an S");
				}
			}
			lineno++;
			first = false;
		}
	}
	
	/**
	 * Constructor for SubConditions
	 * @param condition to check
	 * @param response to give for that condition
	 */
	private Condition(String condition, String response) {
		this.subconditions = null;
		this.condition = condition;
		this.response = parseFunctions(response);
	}
	
	/**
	 * parses functions in the response. Presently we only have %encode(bla) which encodes things 
	 * BASE64
	 * @param response
	 * @return munged responds
	 */
	public static String parseFunctions(String response) {
		if (response.indexOf("%encode(") > -1) {
			int containBrace = response.indexOf("(");
			int closeBrace = response.indexOf(")");
			String contents = response.substring(containBrace+1,closeBrace);
			try {
				return Base64.encodeAsString(contents);
			} catch (Exception e) {
				e.printStackTrace();
				return response;
			}
		}
		return response;
	}

	/**
	 * respond (or don't) to a condition.
	 * @param value from the server
	 * @return response or null if its not worthy.
	 */
	public String respond(String value){
		if (test(value,this.condition)) {
			if (this.response.equals("%NOTHING%")) {
				return null;
			} else {
				return this.response;
			}
		} else if (subconditions != null && subconditions.size()> 0){
			Iterator i = subconditions.iterator();
			while (i.hasNext()) {
				Condition subcon = (Condition)i.next();
				String response = parseFunctions(subcon.respond(value));
				if (response != null) {
					return response;
				}
			}
		}
//TODO:throw an exception instead.
		return null;
	}
	
	/**
	 * helper to the main test function
	 * @param value to respond
	 * @param condition string
	 * @return whether or not we can respond..
	 */
	private boolean test(String value, String condition) {
		if (this.condition.equals(value)) {
			return true;
		}
		if (this.condition.trim().equals("%ANYTHING%")) {
			return true;
		}
		String[] split = this.condition.split("\\%ANYTHING\\%");
		if (split.length <2 && this.condition.indexOf("%ANYTHING%") < 0) {
			return false;
		} else if (split.length < 2){
			String[] temp = new String[] {
				split[0],
				""
			};
			split = temp;
		}
		String pre = split[0];
		String post = split[1];
		try { //dirty hack should check bounds 
			if (value.substring(0,pre.length()).equals(pre) &&
			    value.substring(value.length()-post.length(),value.length()).equals(post)) {
				return true;
			}
		} catch (Exception e) {
			return false;
		}
		return false;
	}

	/**
	 * parse out the C: or S: into an array
	 * @param line to parse
	 * @return array containing the subject (C/S), number (i.e. C2 the subject would be C number=2) and the rest
	 */
	String[] parseLine(String line) {
		String number = null;
		String subject = line.split("\\:\\s")[0];
		String check = line.split(subject+"\\:\\s")[1];
		
		number = subject.substring(1,subject.length());
		subject = ""+subject.charAt(0);
		if (!isInt(number)){
			number = null;
		}
		
		return new String[]{subject,number,check};
	}

	/**
	 * normal kludgy hack to check for numerics
	 * @param string potentially containing a numeric
	 * @return whether it is numeric or not
	 */
	private boolean isInt(String string) {
		try {
			Integer.parseInt(string);
		} catch (Exception e) {
			return false;
		}
		return true;
	}
}
