/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.dev.prot;

import java.lang.reflect.*;
import org.w3c.dom.*;
import java.util.*;

import org.jboss.mail.dev.prot.inf.*;

import org.apache.log4j.*;

/** 
 * @author <a href='mailto:mikea@xoba.com'>mike andrews</a>
 */	
public class ConditionalFactoryBuilder {

    private static final Logger logger = Logger.getLogger(ConditionalFactoryBuilder.class);

    /**
     * create composite conditional for an xml element, or null if the element is not a conditional
     */
    public static ConditionalFactory parseConditionalElement(Element e, Script script) {
	
	final Manager manager = script.getManager();

	// an  'always true' conditional factory for debugging purposes only...
	ConditionalFactory defaultFactory = ConditionalFactoryBuilder.newConditionalFactory(e,script,TrueConditional.class);
	
	final String tag = e.getTagName();
	final String text = XmlUtilities.getElementText(e);

	Map booleans = new HashMap();
	booleans.put("not",NotConditional.class);
	booleans.put("or",OrConditional.class);
	booleans.put("and",AndConditional.class);

	if (tag.equals("client")) {
	    ModuleFactory client = manager.getModuleFactory(e.getAttribute("module"));
	    if (client == null) client = script.getDefaultClientModule();
	    return wrap(client,text);
	} else if (tag.equals("server")) {
	    ModuleFactory server = manager.getModuleFactory(e.getAttribute("module"));
	    if (server == null) server = script.getDefaultServerModule();
	    return wrap(server,text);
	} else if (booleans.containsKey(tag)) {
	    return ConditionalFactoryBuilder.newConditionalFactory(e,script,(Class) booleans.get(tag));
	} else if (tag.equals("connect")) {
	    return wrap(ModuleFactoryBuilder.newModuleFactory(manager,"connect",ConnectModule.class),text);
	} else if (tag.equals("disconnect")) {	
	    return wrap(ModuleFactoryBuilder.newModuleFactory(manager,"disconnect",DisconnectModule.class),text);
	} else if (tag.equals("state-match")) {
	    return wrap(ModuleFactoryBuilder.newModuleFactory(manager,"state-match",StateMatchModule.class),text);
	} else if (tag.equals("description")) {
	    return null;
	} else {
	    throw new RuntimeException("unknown element: " + tag);
	}
	
    }
    
    public static ConditionalFactory wrap(final ModuleFactory fact, final String patternText) {
	if (fact == null) throw new RuntimeException("can't wrap a null factory");
	return new ConditionalFactory() {
		public String toString() {
		    if (fact == null) {
			return "null";
		    } else {
			return "ConditionalFactory.wrap(" + fact.toString() + ")";
		    }
		}
		public Conditional createConditional() throws InstantiationException {
		    return fact.createModule(patternText);
		}
	    };
	
    }	

    public static ConditionalFactory[] parseConditionalElementChildren(Element e, Script script) {
	
	final Manager manager = script.getManager();

	List conditionalFactories = new Vector();
		
	NodeList list = e.getChildNodes();
	for (int j=0; j<list.getLength(); j++) {
	    Node n = list.item(j);
	    if (n.getNodeType() == Node.ELEMENT_NODE) {
		Element child = (Element) n;
		ConditionalFactory fact = ConditionalFactoryBuilder.parseConditionalElement(child,script);
		if (fact != null) conditionalFactories.add(fact);
	    }
	}
	
	ConditionalFactory[] array = new ConditionalFactory[conditionalFactories.size()];
	conditionalFactories.toArray(array);
	return array;
    }

    /**
     * the conditionalClass should implement MutableConditional
     */
    public static MutableConditionalFactory newConditionalFactory(final Element e, final Script script, final Class conditionalClass) {
	
	final Manager manager = script.getManager();

	return new MutableConditionalFactory() {
		public Conditional createConditional() throws InstantiationException {
		    return createMutableConditional();
		}

		public MutableConditional createMutableConditional() throws InstantiationException {
		    try {
			Constructor cons = conditionalClass.getConstructor(new Class[0]);
			MutableConditional c = (MutableConditional) cons.newInstance(new Object[0]);
			// now see if there are any sub-elements that are conditionals
			ConditionalFactory[] facts = parseConditionalElementChildren(e,script);
			c.addSubConditionals(facts);
			return c;
		    } catch (NoSuchMethodException e) {
			throw new RuntimeException(e);
		    } catch (InvocationTargetException e) {
			throw new RuntimeException(e);
		    } catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		    }
		}	
	    };

    }
}
