/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.domaingroup;

import java.util.Set;
import java.util.TreeSet;
import java.util.Comparator;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

import java.net.InetAddress;
import java.net.NetworkInterface;

import org.jboss.mail.ServiceProxy;
import org.jboss.system.ServiceMBeanSupport;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.apache.log4j.Logger;

/**
 * represents a group of domains (such as for using as a required "to" domain in verify identity, authentication
 * required, etc).
 *
 * mikea: domains should be case insensitive, and should actually be a set (not a list)
 *
 * @author Andrew C. Oliver
 * @author <a href='mailto:mikea@xoba.com'> mike andrews </a>
 */
public class DomainGroup extends ServiceMBeanSupport implements DomainGroupMBean, ServiceProxy {
 
    private static final Logger logger = Logger.getLogger(DomainGroup.class);
    private static final Pattern ipAddrPattern = 
	Pattern.compile("^\\[(\\d+\\.\\d+\\.\\d+\\.\\d+)\\]$");
    
    private boolean includesLocalInterfaces;

    public DomainGroup(boolean includesLocalInterfaces) {
	this.includesLocalInterfaces = includesLocalInterfaces;
    }

    public DomainGroup() {  }

    /**
     * a comparator for case-insensitive string representations
     *
     * @author <a href='mikea@xoba.com'>mike andrews</a>
     */
    static public class CaseInsensitiveStringComparator implements Comparator {
	public int compare(Object a, Object b) {
	    return a.toString().toLowerCase().compareTo(b.toString().toLowerCase());
	}
	public boolean equals(Object c) {
	    if (c instanceof CaseInsensitiveStringComparator) {
		return true;
	    } else {
		return false;
	    }
	}
    }

    private Set domains = new TreeSet(new CaseInsensitiveStringComparator());
    
    /* (non-Javadoc)
     * @see org.jboss.mail.domaingroup.DomainGroupMBean#add(java.lang.String)
     */
    public void add(String domain) {
        domains.add(domain.trim());
    }

    public String[] listDomains() {
        String[] retval = new String[domains.size()];
        retval = (String[])domains.toArray(retval);
        return retval;
    }

    /* (non-Javadoc)
     * @see org.jboss.mail.domaingroup.DomainGroupMBean#isInGroup(java.lang.String)
     */
    public boolean isInGroup(String domain) {
	boolean isInGroup = domains.contains(domain);
	if (isInGroup) {
	    return true;
	} else if (includesLocalInterfaces) {
	    Matcher m = ipAddrPattern.matcher(domain);
	    if (m.matches()) {
		try {
		    InetAddress addr = InetAddress.getByName(m.group(1));
		    // returns true if a network interface is bound to address
		    return NetworkInterface.getByInetAddress(addr) != null;
		} catch (Exception e) {
		    // a problem processing inet address not worth failing over...
		    logger.warn("problem processing address domain: " + domain,e);
		    return false;
		}
	    } else {
		return false;
	    }
	} else {
	    return false;
	}
    }

    public boolean getIncludesLocalInterfaces() {
	return includesLocalInterfaces;
    }

    /* (non-Javadoc)
     * @see org.jboss.mail.ServiceProxyFactory#produceServiceProxy()
     */
    public ServiceProxy produceServiceProxy() {
        return this;
    }

    /* (non-Javadoc)
     * @see org.jboss.mail.domaingroup.DomainGroupMBean#setDomains(org.w3c.dom.Element)
     */
    public void setDomains(Element element) {
	String attr = element.getAttribute("includes-local-interfaces");
	if (attr.equals("true")) {
	    includesLocalInterfaces = true;
	} else {
	    includesLocalInterfaces = false;
	}

        NodeList list = element.getElementsByTagName("domain");        
	domains.clear();
        for (int k = 0; k < list.getLength(); k++) {
            Node n = list.item(k);
            domains.add(getNodeText(n).trim());
        }  
    }
    
    private static String getNodeText(Node e) {
	StringBuffer buf = new StringBuffer();
	int type = e.getNodeType();
	if (type == Node.CDATA_SECTION_NODE || type == Node.TEXT_NODE) {
	    buf.append(e.getNodeValue());
	} else if (type == Node.ELEMENT_NODE) {
	    NodeList list = e.getChildNodes();
	    for (int i=0; i<list.getLength(); i++) {
		Node child = list.item(i);
		buf.append(getNodeText(child));
	    }
	}
	return buf.toString();
   }
    
}
