/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.mailbox;

import org.jboss.mail.message.Mail;
import org.apache.log4j.Logger;

/**
 * Provides an implementation of methods that are common
 * across multiple implementations of the Folder interface.
 *
 * @author Eric Daugherty
 */
public abstract class AbstractFolder implements Folder {

    protected boolean locked = false;

    /** Log4j logger */
    private static final Logger log = Logger.getLogger( AbstractFolder.class );

    /**
     * Indicates whether this mailbox is already locked.
     *
     * @return ...
     */
    public boolean isLocked() {
        return locked;
    }

    /**
     * Attempts to obtain an exclusive lock on this mailbox.
     * True is returned if the lock is obtained.
     *
     * @return true if the mailbox is now locked by the caller.
     */
    public synchronized boolean lock() {
        if( !locked )
        {
            log.debug( "Mailbox Locked" );
            locked = true;
            return true;
        }
        return false;
    }

    public abstract boolean commit();

    public abstract void release();

    /**
     * Returns an array of Folders that are stored
     * in this folder.  This implementation just returns
     * and empty array always.
     *
     * @return Empty Folder array.
     */
    public Folder[] getFolders() {
        // Sub-folders are not supported in this version.
        return new Folder[0];
    }

    public abstract MailboxMessage[] getMessages();

    /**
     * Returns the message with the specified id.
     *
     * @param messageId
     * @return null if the message id is invalid
     */
    public MailboxMessage getMessage(String messageId) {

        MailboxMessage[] messages = getMessages();
        MailboxMessage message = null;
        try {
            int index = Integer.parseInt( messageId );
            // Convert 1 based to 0 based.
            index--;
            if( index >= 0 && index < messages.length )
            {
                message = messages[index];
                if( message.isDeleted() )
                {
                    message = null;
                }
            }
        }
        catch (NumberFormatException e) {
        }

        return message;
    }

    public abstract void addMessage(Mail message);
}
