/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.mailbox;

import org.jboss.mail.message.Mail;

/**
 * A folder represents a container for messages and other folders.
 * Some folders may only contain messages, while others may only
 * contain other folders.
 *
 * @author Eric Daugherty
 */
public interface Folder {

    // Access/Transaction control methods

     /**
     * Indicates whether this mailbox is already locked.
     *
     * @return ...
     */
    boolean isLocked();

    /**
     * Attempts to obtain an exclusive lock on this mailbox.
     * True is returned if the lock is obtained.
     *
     * @return true if the mailbox is now locked by the caller.
     */
    boolean lock();

    /**
     * Commits any outstanding changes to the folder state.  This
     * should be called before releasing the folder.  The folder lock
     * will also be released when this method is called.
     *
     * @return returns true if the changes were commited.  False if any error occured.
     */
    boolean commit();

    /**
     * Unlocks the mailbox.  Any uncommited changes are lost.
     */
    void release();

    // Manipulation Methods.

    /**
     * Returns an array of Folders that are stored
     * in this folder.
     *
     * @return ...
     */
    Folder[] getFolders();

    /**
     * Returns an array of Messages that are stored
     * in this folder.
     *
     * @return ...
     */
    MailboxMessage[] getMessages();

    /**
     * Returns the message with the specified id.
     *
     * @param messageId
     * @return null if the message id is invalid
     */
    MailboxMessage getMessage( String messageId );

    /**
     * Store a new email message in this folder.
     * <p>
     * This method should be callable without obtaining
     * a lock on the folder.
     *
     * @param message
     */
    void addMessage( Mail message );

}
