/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.mail.mailbox;

/**
 * Defines an interface that that Mailbox implementations must
 * implement for messages retrieved from storage.
 * <p>
 * This interface is defined seperate from the {@link org.jboss.mail.message.Message} and 
 * {@link org.jboss.mail.message.Mail} classes
 * due to the possible implementation needs of mailboxes, as well as
 * the additional interface requirements unique to the mailbox.
 *
 * @author Eric Daugherty
 */
public interface MailboxMessage {

    /**
     * Each MailboxMessage must have a unique ID.  This ID must be consistent
     * across mailbox sessions.
     *
     * @return the unique id of this message.
     */
    String getId();

    /**
     * Indicates whether this message has been marked for deletion.  The message
     * will not be purged until the commit method is called on the mailbox
     * that holds this message, but deleted messages may be ignored by
     * mailbox access implementations.
     *
     * @return true if the messages is marked for deletion.
     */
    boolean isDeleted();

    /**
     * Indicates whether this message has been marked for deletion.  The message
     * will not be purged until the commit method is called on the mailbox
     * that holds this message, but deleted messages may be ignored by
     * mailbox access implementations.
     *
     * @param deleted true if the messages is marked for deletion.
     */
    void setDeleted(  boolean deleted );

    /**
     * Returns the total size of the message, including headers, seperator, and body.
     * NOTE FROM MIKEA: does this include EOL characters? according to
     * the implementation of {@link org.jboss.mail.mailbox.SimpleMailboxMessage},
     * it doesn't.
     *
     * @return total message size.
     */
    long getSize();

    /**
     * Returns the headers as a string array.  Individual strings do not
     * contain EOL characters at the end (although they may contain them in the middle).
     *
     * @return array of Message Headers as strings.
     */
    String[] getHeaders();

    /**
     * Returns the body as a string array.  Individual strings do not
     * contain EOL characters.
     *
     * @return array of message body lines.
     */
    String[] getBody();
}
