/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.mailbox;

import java.util.ArrayList;
import java.util.Iterator;

import org.apache.log4j.Logger;
import org.jboss.mail.message.Mail;

/**
 * A simple implementation for the MailboxMessage interface.  This implementation
 * just stores the data in local arrays, and provides several constructors to easily
 * convert a Mail message.
 *
 * @author Eric Daugherty
 * @author Andrew C. Oliver <acoliver ot jboss dat org>
 */
public class SimpleMailboxMessage implements MailboxMessage {

    protected String id;
    protected boolean isDeleted;
    protected String[] headers;
    protected String[] body;

     /** Log4j logger */
    private static final Logger log = Logger.getLogger( SimpleMailboxMessage.class );

    /**
     * Creates a new SimpleMailboxMessage with the required data.
     * @param id a unique identifier for this message.
     * @param headers array of internet header.  Strings should not end with EOL characters.
     * @param body array of body lines.  Strings should not end with EOL characters.
     */
    public SimpleMailboxMessage( String id, String[] headers, String[] body ) 
    {
	this(id);
        this.headers = headers;
        this.body = body;
    }

    public SimpleMailboxMessage( String id, Mail message ) 
    {
	this(id);
        Iterator headerEnumeration = message.getMailHeaders().getAllHeaderLines();
        ArrayList lines = new ArrayList();
        while( headerEnumeration.hasNext() )
        {
            lines.add( headerEnumeration.next() );
        }

        // Convert to array.
        headers = new String[lines.size()];
        headers = (String[]) lines.toArray( headers );

        body = new String( message.getBody() ).split( "\r\n" );
    }

    /**
     * Protected constructor for use by other implementations.
     *
     * @param id unique ID of this message.
     */
    protected SimpleMailboxMessage( String id ) 
    {
        this.id = id;
    }

    /**
     * Each MailboxMessage must have a unique ID.  This ID must be consistent
     * across mailbox sessions.
     *
     * @return the unique id of this message
     */
    public String getId() {
        return id;
    }

    /**
     * Indicates whether this message has been marked for deletion.  The message
     * will not be purged until the commit method is called on the mailbox
     * that holds this message, but deleted messages may be ignored by
     * mailbox access implementations.
     *
     * @return true if the messages is marked for deletion.
     */
    public boolean isDeleted() {
        return isDeleted;
    }

    /**
     * Indicates whether this message has been marked for deletion.  The message
     * will not be purged until the commit method is called on the mailbox
     * that holds this message, but deleted messages may be ignored by
     * mailbox access implementations.
     *
     * @param deleted true if the messages is marked for deletion.
     */
    public void setDeleted(boolean deleted) {
        this.isDeleted = deleted;
    }

    /**
     * Returns the total size of the message, including headers, seperator, and body.
     *
     * @return total message size.
     */
    public long getSize() {
        long size = 0;

        // Add the headers
        for( int index = 0; index < headers.length; index++ )
        {
            size += headers[index].length();
        }

        // Add the 'emtpy line' between headers and body.
        size += 2;

        for( int index = 0; index < body.length; index++ )
        {
            size+= body.length;
        }

        return size;
    }

    /**
     * Returns the headers as a string array.  Individual strings do not
     * contain EOL characters at the end (although they may contain them in the middle).
     *
     * @return array of Message Headers as strings.
     */
    public String[] getHeaders() {
        return headers;
    }

    /**
     * Returns the body as a string array.  Individual strings do not
     * contain EOL characters.
     *
     * @return array of message body lines.
     */
    public String[] getBody() {
        return body;
    }
}
