/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.mailbox.entity;

import java.util.Collection;
import java.util.Iterator;

import javax.ejb.CreateException;
import javax.ejb.DuplicateKeyException;
import javax.ejb.FinderException;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.jboss.logging.Logger;
import org.jboss.mail.mailbox.AbstractFolder;
import org.jboss.mail.mailbox.Folder;
import org.jboss.mail.mailbox.MailboxMessage;
import org.jboss.mail.mailbox.entity.message.LocalMessageSB;
import org.jboss.mail.mailbox.entity.message.LocalMessageSBHome;
import org.jboss.mail.message.Mail;

/**
 * Provides an implementation of the Folder interface that uses
 * Entity beans to store individual messages.
 *
 * @author Eric Daugherty
 * @author Andrew C. Oliver <acoliver ot jboss dat org>
 */
public class EntityFolder extends AbstractFolder implements Folder {

    private String folderName;
    private EntityMessage[] messages;

    /** Log4j logger */
    private static final Logger log = Logger.getLogger( EntityFolder.class );

    public EntityFolder( String folderName )
    {
        this.folderName = folderName;
        log.warn("folder requested for Entityfolder="+folderName);
        try {
            InitialContext context = new InitialContext();
            LocalMessageSBHome messageHome = (LocalMessageSBHome) context.lookup( "jbossmail.MessageSB" );
            LocalMessageSB messageBean = messageHome.create();
            EntityMessage[] mailboxMessages = messageBean.getMessagesForFolder(folderName);
            this.messages = mailboxMessages;
        } catch( NamingException namingException ) {
            log.error( "Error loading messages for mailbox: " + folderName + ".  Unable to lookup Entity Bean Home: " + namingException, namingException );
            this.messages = new EntityMessage[0];
        } catch( CreateException createException ) {
            log.error( "Error loading messages for mailbox: " + folderName + ".  Unable to create Entity Bean : " + createException, createException );
            this.messages = new EntityMessage[0];
        }
    }

    /**
     * Commits any outstanding changes to the folder state.  This
     * should be called before releasing the folder.  The folder lock
     * will also be released when this method is called.
     *
     * @return returns true if the changes were commited.  False if any error occured.
     */
    public boolean commit() {
       try {
           InitialContext context = new InitialContext();
           LocalMessageSBHome messageHome = (LocalMessageSBHome) context.lookup( "jbossmail.MessageSB" );
           LocalMessageSB messageBean = messageHome.create(); 
           messageBean.deleteMessages( messages );

           locked = false;

           log.debug( "Mailbox Commited" );
        } catch( NamingException namingException ) { 
            log.error( "Error loading messages for mailbox: " + folderName + ".  Unable to lookup Entity Bean Home: " + namingException, namingException );
            this.messages = new EntityMessage[0];
        } catch( CreateException createException ) {
            log.error( "Error loading messages for mailbox: " + folderName + ".  Unable to create Entity Bean : " + createException, createException );
            this.messages = new EntityMessage[0];
        }

        return true;
    }

    /**
     * Unlocks the mailbox.  Any uncommited changes are lost.
     */
    public void release() {
        // Nothing to do, just
    }

    /**
     * Returns an array of Messages that are stored
     * in this folder.
     *
     * @return ...
     */
    public MailboxMessage[] getMessages() {
        return messages;
    }

    /**
     * Store a new email message in this folder.
     * <p/>
     * This method should be callable without obtaining
     * a lock on the folder.
     *
     * @param message
     */
    public void addMessage(Mail message) {

        try {

            InitialContext context = new InitialContext();
            LocalMessageSBHome messageSBHome = (LocalMessageSBHome) context.lookup( "jbossmail.MessageSB" );
            LocalMessageSB messageBean = messageSBHome.create();
            messageBean.createMessage( folderName, message);

        }
        catch( NamingException namingException )
        {
            log.error( "Error storing messages for mailbox: " + folderName + ".  Unable to lookup Entity Bean Home: " + namingException, namingException );
        }
        catch (CreateException createException) {
            log.error( "Error storing new message for mailbox: " + folderName + ".  Create Exception: " + createException, createException );
        }
    }
}
