/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.mailbox.memory;

import org.jboss.mail.mailbox.Folder;
import org.jboss.mail.mailbox.MailboxMessage;
import org.jboss.mail.mailbox.SimpleMailboxMessage;
import org.jboss.mail.mailbox.AbstractFolder;
import org.jboss.mail.message.Mail;
import org.apache.log4j.Logger;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Random;

/**
 * Provides a memory only based implementation of the Folder
 * interface.
 *
 * @author Eric Daugherty
 */
public class MemoryFolder extends AbstractFolder implements Folder {

    private List messages = new ArrayList();
    private List queuedMessages = new ArrayList();

    private static Random uniqueId = new Random( System.currentTimeMillis() );

    /** Log4j logger */
    private static final Logger log = Logger.getLogger( MemoryFolder.class );

    /**
     * Commits any outstanding changes to the folder state.  This
     * should be called before releasing the folder.  The folder lock
     * will also be released when this method is called.
     *
     * @return returns true if the changes were commited.  False if any error occured.
     */
    public boolean commit() {
        int messageCount = messages.size();
        MailboxMessage message;
        for( int index = 0; index < messageCount; index++ )
        {
            message = (MailboxMessage) messages.get( index );
            if( message.isDeleted() )
            {
                messages.remove( index );
                index--;
                messageCount--;
            }
        }

        mergeQueuedMessages();

        locked = false;

        log.debug( "Mailbox Commited" );

        return true;
    }

    /**
     * Unlocks the mailbox.  Any uncommited changes are lost.
     */
    public void release() {

        // Since the folder is stored in memory, we need to make sure we clear
        // the deleted flags.

        int messageCount = messages.size();
        MailboxMessage message;
        for( int index = 0; index < messageCount; index++ )
        {
            message = (MailboxMessage) messages.get( index );
            message.setDeleted( false );
        }

        mergeQueuedMessages();

        locked = false;

        log.debug( "Mailbox Released" );
    }

    /**
     * Returns an array of Messages that are stored
     * in this folder.
     *
     * @return ...
     */
    public MailboxMessage[] getMessages() {
        MailboxMessage[] messageArray = new MailboxMessage[messages.size()];
        return (MailboxMessage[]) messages.toArray( messageArray );
    }

    /**
     * Store a new email message in this folder.
     *
     * @param mail
     */
    public synchronized void addMessage(Mail mail ) {
        SimpleMailboxMessage message = new SimpleMailboxMessage( String.valueOf( uniqueId.nextLong() ), mail );

        // If it is locked, add it to the queued message array
        // until the lock is releated.
        if( locked )
        {
            queuedMessages.add( message );
        }
        else
        {
            messages.add( message );
        }

        log.debug( "Message added." );
    }

    /**
     * Copy the queued messages into the normal queue.
     */
    private void mergeQueuedMessages()
    {
        Iterator queuedIterator = queuedMessages.iterator();
        while( queuedIterator.hasNext() )
        {
            messages.add( queuedIterator.next() );
        }
        queuedMessages = new ArrayList();
    }
}
