/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.mailhandler.localmailbox;

import java.util.ArrayList;

import javax.ejb.MessageDrivenBean;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.management.ObjectName;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.Logger;
import org.jboss.mail.mailbox.Folder;
import org.jboss.mail.mailbox.Mailbox;
import org.jboss.mail.mailbox.MailboxManager;
import org.jboss.mail.maillistener.JMXMailListenerProxy;
import org.jboss.mail.message.EnvelopedAddress;
import org.jboss.mail.message.Mail;
import org.jboss.mail.message.MailAddress;
import org.jboss.mail.message.MailRetryWrapper;
import org.jboss.mail.userrepository.UserRepository;
import org.jboss.mail.util.JBMailJMXUtil;

/**
 * Processes messages from the local delivery queue
 * and delivers them to the local mailbox.  If the local
 * mailbox does not exist, a new error message is created and sent
 * to the sender.
 *
 * @author Eric Daugherty
 */
public class LocalDeliveryMessageDrivenBean implements MessageDrivenBean, MessageListener {

    /** Log4j logger */
    private static final Logger log = Logger.getLogger( LocalDeliveryMessageDrivenBean.class );
	private Context ctx;
	/** Name of the MailListener MBean */
	private ObjectName listenerName;
	/** Name of the UserRepository MBean */
	private ObjectName userRepositoryName;

    /**
     * Initialize the MessageDrivenBean
     */
    public LocalDeliveryMessageDrivenBean () {

        log.debug( "LocalDeliveryMessageDrivenBean initialized." );
    }

    /**
     * Handles incoming email messages from the queue and delivers them
     * to the appropriate mailbox(es).
     *
     * @param jmsMessage the incoming JMS Message
     */
    public void onMessage( javax.jms.Message jmsMessage ) {
    	
    	ArrayList failed = new ArrayList();
        try {
            // Convert the incoming message to a Mail message.
            ObjectMessage jmsObjectMessage = (ObjectMessage) jmsMessage;
            Object messageObject = jmsObjectMessage.getObject();
            if( !(messageObject instanceof Mail) )
            {
                log.error( "Invalid Object placed in Local Delivery queue.  Expected type is: " + Mail.class + " but was: " + messageObject.getClass() + "; Message will be ignored.");
                return;
            }
            Mail smtpMessage = (Mail) jmsObjectMessage.getObject();

            // Access the mailbox manager.
            MailboxManager mailboxManager = Mailbox.getMailboxManager();

            // Deliver each recipient. Only deliver to local users 
            MailAddress[] addresses = smtpMessage.getTo();
            for( int index = 0; index < addresses.length; index++ )
            {
            	EnvelopedAddress ea = new EnvelopedAddress(addresses[index]);
            	if (ea.getLocal()) {
            		if (!deliver( mailboxManager, addresses[index], smtpMessage )){
            			failed.add(ea);
            		}
            	}
            }
            
            //Retry the failed recipients
			if (failed.size() >0){
				MailRetryWrapper mailWrp = (smtpMessage instanceof MailRetryWrapper) 
							? (MailRetryWrapper)smtpMessage
							: new MailRetryWrapper(smtpMessage, MailRetryWrapper.REDELIVERY_LOCAL);
							
				mailWrp.setTos(failed);
								
				JMXMailListenerProxy listener =
					 (JMXMailListenerProxy) JBMailJMXUtil.getService(listenerName);
            	
				listener.send(mailWrp);
			}
 			
        }
        catch ( Exception e ) {
            log.error( "Error handling message, message lost! " + e, e );
        }
    }

    public void setMessageDrivenContext(javax.ejb.MessageDrivenContext context ) {
    }

    public void ejbCreate() {
		try {
			ctx = new InitialContext();
			ctx = (Context) ctx.lookup("java:comp/env");
		} catch (NamingException e) {
			log.error("Could not get initial context", e);
			throw new RuntimeException("Could not get initial context!");
		}
		try {
			listenerName = new ObjectName((String)ctx.lookup("MailListenerService"));
		} catch (Exception e) {
			log.error("Could not get the JMX name of the MailListener service (*:MailListener:* usually) "+ e, e);
			throw new RuntimeException("Could not get the JMX name of the delivery service");
		}
		try {
			userRepositoryName = new ObjectName((String)ctx.lookup("UserRepositoryService"));
		} catch (Exception e) {
			log.error("Could not get the JMX name of the UserRepository service (*:UserRepository:* usually) "+ e, e);
			throw new RuntimeException("Could not get the JMX name of the delivery service");
		}
	
    }

    public void ejbRemove() {
    }

    /**
     * Delivers the email to the specified local address.
     *
     * @param manager MailboxManger instance
     * @param address the specific address to deliver to.
     * @param message the message to deliver.
     */
    private boolean deliver( MailboxManager manager, MailAddress address, Mail message )
    {
    	if (log.isDebugEnabled()){
    		log.debug("Attempting local delivery to: " + address);
    	}
    	
		try{
			String user = address.getUser();
			UserRepository userRepository = (UserRepository)JBMailJMXUtil.getService(userRepositoryName);
			
	 		if (userRepository.test(user)){
		    	Folder mailbox = manager.getMailbox(user);
		        if( mailbox != null )
		        {
		            mailbox.addMessage( message );
					return true;
		        }
	 		}

			log.error( "Unable to deliver message to local user: " + address.getUser() + ".  Unknown user." );
		}catch(Exception e){
			 log.error("Could not deliver local mail", e);  
     	}
        return false;
    }
}
