/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.mailhandler.maillist;

import java.io.Serializable;
import java.util.ArrayList;

import javax.ejb.MessageDrivenBean;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.management.ObjectName;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.Logger;
import org.jboss.mail.MailListener;
import org.jboss.mail.maillist.MailList;
import org.jboss.mail.maillist.MailListManagerProxy;
import org.jboss.mail.message.EnvelopedAddress;
import org.jboss.mail.message.Mail;
import org.jboss.mail.message.MailAddress;
import org.jboss.mail.message.MailHeaders;
import org.jboss.mail.message.Message;
import org.jboss.mail.util.JBMailJMXUtil;

/**
 * Processes messages from the local delivery queue 
 * and delivers them to the local mailbox.  If the local
 * mailbox does not exist, a new error message is created and sent
 * to the sender.
 *
 * @author Eric Daugherty
 */
public class MailListMessageDrivenBean implements MessageDrivenBean, MessageListener {

    /** Log4j logger */
    private static final Logger log = Logger.getLogger( MailListMessageDrivenBean.class );
    private Context ctx;
    /** Name of the MailListener MBean */
    private MailListener listener;
    
    MailListManagerProxy proxy;

    /**
     * Initialize the MessageDrivenBean
     */
    public MailListMessageDrivenBean () {

        log.debug( "LocalDeliveryMessageDrivenBean initialized." );
    }

    /**
     * Handles incoming email messages from the queue and delivers them
     * to the appropriate mailbox(es).
     *
     * @param jmsMessage the incoming JMS Message
     */
    public void onMessage( javax.jms.Message jmsMessage ) {
        
        ArrayList failed = new ArrayList();
        try {
            // Convert the incoming message to a Mail message.
            ObjectMessage jmsObjectMessage = (ObjectMessage) jmsMessage;
            Object messageObject = jmsObjectMessage.getObject();
            if( !(messageObject instanceof Mail) )
            {
                log.error( "Invalid Object placed in MailList queue.  Expected type is: " + Mail.class + " but was: " + messageObject.getClass() + "; Message will be ignored.");
                return;
            }
            Mail smtpMessage = (Mail) jmsObjectMessage.getObject();

            // Deliver each recipient. Only deliver to local users 
            MailAddress[] addresses = smtpMessage.getTo();
            for( int index = 0; index < addresses.length; index++ )
            {   
                MailList list = proxy.findList(addresses[index]); 
                if (list != null) {
                    MailAddress[] members = list.getMembers();
                    for (int k = 0; k < members.length; k++) {
                        deliver(addresses[index],members[k],smtpMessage);
                    }
                }
            }
            
             
        }
        catch ( Exception e ) {
            log.error( "Error handling message, message lost! " + e, e );
        }
    }

    public void setMessageDrivenContext(javax.ejb.MessageDrivenContext context ) {
    }

    public void ejbCreate() {
        try {
            ctx = new InitialContext();
            ctx = (Context) ctx.lookup("java:comp/env");
        } catch (NamingException e) {
            log.error("Could not get initial context", e);
            throw new RuntimeException("Could not get initial context!");
        }
        try {
            listener = (MailListener)JBMailJMXUtil.getService(new ObjectName((String)ctx.lookup("MailListenerService")));
            
        } catch (Exception e) {
            log.error("Could not get the JMX name of the MailListener service (*:MailListener:* usually) "+ e, e);
            throw new RuntimeException("Could not get the JMX name of the delivery service");
        }
        try {
            this.proxy = (MailListManagerProxy) JBMailJMXUtil.getService(new ObjectName((String)ctx.lookup("MailListManager")));
        } catch (Exception e) {
            log.error("Could not get the JMX name of the MailListMgr service "+ e, e);
            throw new RuntimeException("Could not get the JMX name of the mail list service");
        }
    
    }

    public void ejbRemove() {
    }

    /**
     * Delivers the email to the specified local address.
     *
     * @param manager MailboxManger instance
     * @param address the specific address to deliver to.
     * @param message the message to deliver.
     */
    private Message deliver( final MailAddress listaddress, final MailAddress address, final Mail message )
    {

        if (log.isDebugEnabled()){
            log.debug("Attempting delivery to: " + address);
        }

        return listener.send(new WrappedMail(listaddress, address, message) );
        
    }
}

class WrappedMail extends Mail implements Serializable {
        EnvelopedAddress eal;
        EnvelopedAddress ear;
        Mail message;
        public WrappedMail(MailAddress listaddress, MailAddress address, Mail message) {
        	  super(message);
            this.eal = new EnvelopedAddress(listaddress,javax.mail.Message.RecipientType.TO);
            this.ear = new EnvelopedAddress(address, javax.mail.Message.RecipientType.BCC);
            this.message = message;
        }
    
        public byte[] getBody() {
            return message.getBody();
        }

        public MailAddress getFrom() {
            return message.getFrom();
        }

        public MailHeaders getMailHeaders() {
            return message.getMailHeaders();
        }

        public MailAddress[] getTo() {
        	  //Only return the real address, we don't want to try sending to the list address again 
            return new MailAddress[]{ear};
        }
}
