/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.mailhandler.remote;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import javax.ejb.MessageDrivenBean;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.management.ObjectName;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.Logger;
import org.jboss.mail.mailbox.Mailbox;
import org.jboss.mail.mailbox.MailboxManager;
import org.jboss.mail.maillistener.JMXMailListenerProxy;
import org.jboss.mail.message.BouncedMail;
import org.jboss.mail.message.EnvelopedAddress;
import org.jboss.mail.message.Mail;
import org.jboss.mail.message.MailAddress;
import org.jboss.mail.message.MailRetryWrapper;
import org.jboss.mail.smtp.sender.SMTPResult;
import org.jboss.mail.smtp.sender.SMTPSender;
import org.jboss.mail.util.JBMailJMXUtil;

/**
 * Processes messages from the remote delivery queue
 * and delivers them to the appropriate SMTP Server.
 *
 * @author Eric Daugherty
 * @author Andrew C. Oliver
 */
public class RemoteDeliveryMessageDrivenBean implements MessageDrivenBean, MessageListener {

    /** Log4j logger */
    private static final Logger log = Logger.getLogger( RemoteDeliveryMessageDrivenBean.class );
    private Context ctx;
    private ObjectName serviceName;
    private ObjectName listenerName;

    /**
     * Initialize the MessageDrivenBean
     */
    public RemoteDeliveryMessageDrivenBean () {

        log.debug( "RemoteDeliveryMessageDrivenBean initialized." );
    }

    /**
     * Handles incoming email messages from the queue and delivers them
     * to the appropriate mailbox(es).
     *
     * @param jmsMessage the incoming JMS Message
     */
    public void onMessage( javax.jms.Message jmsMessage ) {
        log.debug("attempting to deliver the message to remote receivers");
        try {
        	//Map domains = new HashMap(); //hack because JavaMail sends mails to all recipients on a given
        	                             // host whether you want it to or not.
        	                             //KAB: I'm not so sure if this is actually true, rather I think it was 
        	                             //a symptom of other bugs....
        	                             
            // Convert the incoming message to a Mail message.
            ObjectMessage jmsObjectMessage = (ObjectMessage) jmsMessage;
            Object messageObject = jmsObjectMessage.getObject();
            if( !(messageObject instanceof Mail) )
            {
                log.error( "Invalid Object placed in Remote Delivery queue.  Expected type is: " + Mail.class + " but was: " + messageObject.getClass() + " Message will be ignored.");
                return;
            }
            Mail smtpMessage = (Mail) jmsObjectMessage.getObject();

            // Access the mailbox manager.
            MailboxManager mailboxManager = Mailbox.getMailboxManager();

			//Get all addresses by domain
            MailAddress[] addresses = smtpMessage.getTo();
			ArrayList excluded = new ArrayList();						
			for( int index = 0; index < addresses.length; index++ )
			{
				EnvelopedAddress ea = new EnvelopedAddress(addresses[index]); // MIKEA: SEEMS LIKE WE CAN AVOID CLASSCASTEXCEPTIONS THIS WAY...

				if (ea.getLocal()) {
					excluded.add(ea);
				}
			}
			
			//Kab: If it seems funny that I am stripping out the local addresses and passing them 
			//in to the remote SMTPSender, it is because I'm going with the original SMTPSender interface.
			//I can change it if required
			SMTPResult[] results = deliver(mailboxManager, 
									(MailAddress[])excluded.toArray(new MailAddress[excluded.size()]), 
									smtpMessage);	
									
			//Retry the failed recipients
			ArrayList failed = new ArrayList();
		
			for (int i = 0 ; i < results.length ; i++){
				if (!results[i].isOk()){
					failed.add(results[i].getAddress());
				}
			}
	
			if (failed.size() > 0){
				MailRetryWrapper mailWrp = (smtpMessage instanceof MailRetryWrapper) 
							? (MailRetryWrapper)smtpMessage
							: new MailRetryWrapper(smtpMessage, MailRetryWrapper.REDELIVERY_REMOTE);

				mailWrp.setTos(failed);
							
				JMXMailListenerProxy listener =
					 (JMXMailListenerProxy) JBMailJMXUtil.getService(listenerName);
        	
				listener.send(mailWrp);
			}
        }
        catch ( Exception e ) {
            log.error( "Error handling message, message lost! " + e, e );
        }
    }

    public void setMessageDrivenContext(javax.ejb.MessageDrivenContext context ) {
    }

    public void ejbCreate() {
        try {
            ctx = new InitialContext();
            ctx = (Context) ctx.lookup("java:comp/env");
        } catch (NamingException e) {
            log.error("Could not get initial context", e);
            throw new RuntimeException("Could not get initial context!");
        }
		try {
			serviceName = new ObjectName((String)ctx.lookup("DeliveryService"));
		} catch (Exception e) {
			log.error("Could not get the JMX name of the Delivery service (*:SMTPSender:* usually) "+ e, e);
			throw new RuntimeException("Could not get the JMX name of the delivery service");
		}
		try {
			listenerName = new ObjectName((String)ctx.lookup("MailListenerService"));
		} catch (Exception e) {
			log.error("Could not get the JMX name of the MailListener service (*:MailListener:* usually) "+ e, e);
			throw new RuntimeException("Could not get the JMX name of the delivery service");
		}
    }

    public void ejbRemove() {
    }

    /**
     * Delivers the email to the specified address.
     *
     * @param manager MailboxManger instance
     * @param address the specific address to deliver to.
     * @param message the message to deliver.
     */
    private SMTPResult[] deliver( MailboxManager manager, MailAddress[] excluded, Mail message )
    {
		log.debug("deliver - exclude:" + excluded);
    	try {
            SMTPSender sender = (SMTPSender)JBMailJMXUtil.getService(serviceName);
            return sender.send(message, excluded);
        } catch (Exception e) {
        	//This should never happen, and here we don't know what messages have failed and what have not
            log.error("Error sending, no data available for bounce messages", e);
            return new SMTPResult[]{};
    	}
    }
}

