/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
 
package org.jboss.mail.maillist.memory;

import java.util.HashMap;

import org.jboss.mail.ServiceProxy;
import org.jboss.mail.maillist.MailList;
import org.jboss.mail.maillist.MailListManagerProxy;
import org.jboss.mail.maillist.MailListProperties;
import org.jboss.mail.maillist.MailListPropertyConstants;
import org.jboss.mail.message.MailAddress;
import org.jboss.system.ServiceMBeanSupport;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/** Memory implementation of MailListManager MBean
*
* @author <a href="kabirkhan@bigfoot.com">Kabir Khan</a>
*/
public class MemoryMailListManager extends ServiceMBeanSupport implements MailListPropertyConstants, MemoryMailListManagerMBean {
	HashMap mailingLists = new HashMap();
	
	/** Creates a MailList with the passed in address as its name
	 * 
	 * @param listAddress Address of list to create
	 * @param properties of list we are creating
	 * @return false if a list already exists with that name
	 */
	public boolean createList(MailAddress listAddress, MailListProperties properties) {
		//TODO: add concurrency stuff
		MailList mailingList = (MailList)mailingLists.get(listAddress);
		
		if ( mailingList != null ){
			return false;
		}
		mailingList = new MemoryMailList(listAddress, properties);
		mailingLists.put(listAddress, mailingList);

		return true;
	}

	/** Deletes the MailList with the passed in address as its name
	 * 
	 * @param listAddress Address of list to delete
	 */
	public void deleteList(MailAddress listAddress) {
		//TODO: add concurrency stuff
		mailingLists.remove(listAddress);

	}

	/** Finds the MailList with the passed in address as its name
	 * 
	 * @param listAddress Address of list to find
	 * @return The created list. If the list already existed the existing list is returned
	 */
	public MailList findList(MailAddress address) {
		//TODO: add concurrency stuff
		return (MailList)mailingLists.get(address);
	}

	/** Initialise the MemoryMailListManager MBean with xml data
	 * 
	 * @param mailingLists
	 */
	public void setMailingLists(Element mailingLists){
		System.out.println("setMailingLists()");
		NodeList mailListNodes = mailingLists.getElementsByTagName("mailingList");
		int length = mailListNodes.getLength();
		for (int i = 0 ; i < length ; i++){
			Element mailingList = (Element)mailListNodes.item(i);
			loadMailingList(mailingList);
		}
	}
	
	private void loadMailingList(Element mailingListElement){
		
		//Get the list name
		NodeList listAddrNodes = mailingListElement.getElementsByTagName("listAddress");
		
		if (listAddrNodes.getLength() != 1){
			throw new RuntimeException("Each mailingLists/mailingList element must have exactly one listAddress child element");
		}
		String listAddr = listAddrNodes.item(0).getFirstChild().getNodeValue();
		MailAddress listAddress = MailAddress.parseSMTPStyle(listAddr);
		
		if (!listAddress.isValid()){
			throw new RuntimeException("mailingLists/mailingList/listAddress must be a well-formed email address");
		}
		
		//Get the list properties
		MailListProperties properties = new MailListProperties();
		NodeList propertyNodes = mailingListElement.getElementsByTagName("property");
		int length = propertyNodes.getLength();
		
		for (int i = 0 ; i < length ; i++){
			String pname = null;
			String pval = null;
			Node property = propertyNodes.item(i); 
			NodeList subNodes = property.getChildNodes();
			int subLength = subNodes.getLength();

			for (int j = 0 ; j < subLength ; j++){

				Node subNode = subNodes.item(j);
				if (subNode.getNodeName().equals("name")) {
					pname = subNode.getFirstChild().getNodeValue();
				} else if (subNode.getNodeName().equals("value")) {
					pval = subNode.getFirstChild().getNodeValue();
				} else {
					throw new RuntimeException("only permitted child elements of mailingLists/mailingList/property are name and value");
				}
			}
			
			if (pname == null || pval == null){
				throw new RuntimeException("Need to specify both name and value child elements for mailingLists/mailingList/property");
			}
			
			if (pname.equals(REPLY_TO_LIST)){
				properties.setProperty(pname, new Boolean(pval));
			}else if (pname.equals(SUBJECT_PREFIX)){
				properties.setProperty(pname, pval);
			}else if (pname.equals(PREFIX_AUTO_BRACKETED)){
				properties.setProperty(pname, new Boolean(pval));
			}else if (pname.equals(ATTACHMENT_ALLOWED)){
				properties.setProperty(pname, new Boolean(pval));
			}else if (pname.equals(MEMBERS_ONLY)){
				properties.setProperty(pname, new Boolean(pval));
			}
		}
		
		//Create the list		
		createList(listAddress, properties);
		MailList mailingList = findList(listAddress);
		
		
		//Add members to the list
		NodeList membersNodes = mailingListElement.getElementsByTagName("members");
		
		if (membersNodes.getLength() > 1){
			throw new RuntimeException("Each mailingLists/mailingList element can only have one members child element");
		}
		else if (membersNodes.getLength() == 1){
			NodeList memberNodes = ((Element)membersNodes.item(0)).getElementsByTagName("member");
			int memberLength = memberNodes.getLength();

			for (int i = 0 ; i < memberLength ; i++){
				String addr = memberNodes.item(i).getFirstChild().getNodeValue();
				MailAddress address = MailAddress.parseSMTPStyle(addr);
				
				if (address.isValid()){
					mailingList.addMember(address);
				}else{
					log.debug("Member " + addr + " in list " + listAddr 
						+ " is not a well formed email address and has not been added to the list");
				}
			}
		}
	}

	/* (non-Javadoc)
	 * @see org.jboss.mail.ServiceProxyFactory#produceServiceProxy()
	 */
	public ServiceProxy produceServiceProxy() {
		// TODO Auto-generated method stub
		return new MailListManagerProxy(this.getServiceName());
	}

}
