/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.maillistener;

import java.util.ArrayList;

import javax.jms.ObjectMessage;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.management.ObjectName;
import javax.naming.InitialContext;

import org.apache.log4j.Logger;
import org.jboss.mail.MailListener;
import org.jboss.mail.ServiceProxy;
import org.jboss.mail.ServiceProxyFactory;
import org.jboss.mail.maillist.MailListManagerProxy;
import org.jboss.mail.message.Mail;
import org.jboss.mail.message.MailAddress;
import org.jboss.mail.message.Message;
import org.jboss.mail.util.JBMailJMXUtil;
import org.jboss.system.ServiceMBeanSupport;


/**
 * MailListJMSListener - sends mails to a special queue for de-listifying mail
 * 
 * At current this needs to be early in the chain.
 *   
 * @author Andrew C. Oliver
 */
public class MailListJMSMailListener
    extends ServiceMBeanSupport
    implements MailListJMSMailListenerMBean, ServiceProxyFactory, MailListener{

    private static final Logger jblog = Logger.getLogger(MailListJMSMailListener.class);
    private ObjectName mlmgr;

    /**
     * queue or topic
     */
    private String destType;
    /**
     * destination queue/topic name
     */
    private String destination;
    /**
     * QCF instance from JNDI (null unless destType="queue")
     */
    private QueueConnectionFactory qcf;
    /**
     * TCF instance from JNDI (null unless destType="topic")
     */
    private TopicConnectionFactory tcf;
    /**
     * destination queue (null unless destType="queue")
     */
    private Queue destinationQueue;
    /**
     * destination topic (null unless destType="topic")
     */
    private Topic destinationTopic;

    private String queueConnectionFactoryName;
    private String topicConnectionFactoryName;

    public MailListJMSMailListener(String queueConnectionFactoryName, String topicConnectionFactoryName) {
        this.queueConnectionFactoryName = queueConnectionFactoryName;
        this.topicConnectionFactoryName = topicConnectionFactoryName;
    }

    /* (non-Javadoc)
     * @see org.jboss.mail.ServiceProxyFactory#produceServiceProxy()
     */
    public ServiceProxy produceServiceProxy() {
//        try {Class clazz = Class.forName("org.jboss.mail.maillistener.JMXMailListenerProxy");
//        JMXMailListenerProxy proxy1 = (JMXMailListenerProxy)clazz.newInstance();}
//        catch (Exception e) { e.printStackTrace();}
        JMXMailListenerProxy proxy = new JMXMailListenerProxy(this.getServiceName());
        return proxy;
    }

    public void startService() throws Exception {
        InitialContext ctx = new InitialContext();
        if (destType.equals("topic")) {
          //TODO: make configurable
          tcf = (TopicConnectionFactory)ctx.lookup(topicConnectionFactoryName);
          destinationTopic = (Topic)ctx.lookup("topic/"+destination);
        } else {
          //TODO: make configurable
          qcf = (QueueConnectionFactory)ctx.lookup(queueConnectionFactoryName);
          destinationQueue = (Queue)ctx.lookup("queue/"+destination);
        }
    }

    public void stopService() throws Exception {

    }

   public void setDestinationType(String destType) {
        this.destType = destType;
    }

    public String getDestinationType() {
        return this.destType;
    }

    public void setDestination(String postDest) {
        this.destination = postDest;
    }


    public String getDestination() {
        return destination;
    }

    public Message send(Message msg) {
        boolean retval = true;
        try { 
            MailListManagerProxy mgr = (MailListManagerProxy)JBMailJMXUtil.getService(mlmgr);
            Mail mail = (Mail)msg;
            MailAddress[] addrs = mail.getTo();
            
            ArrayList notHandledAddrs = new ArrayList();
            boolean isMailList = false;
            
            for (int k = 0; k < addrs.length; k ++) {
                if (mgr.findList(addrs[k]) == null) {
                    notHandledAddrs.add(addrs[k]);
                }else{
                	isMailList = true;
                }
            }
            
            if (isMailList){
                //Message was for list
                if (destType.equals("topic")) {
                    sendMessageTopic(msg);
                }   else {
                    sendMessageQueue(msg);
                }
                
                int size = notHandledAddrs.size();
                if (mail.getTo().length == size){
                	//We handled all the addresses
                	//TODO: return null, or WrappedMail with empty to array?
                	return null;
                }else{
                	//Create new Mail with all addresses not for list
                	MailAddress[] newAddrs = (MailAddress[])notHandledAddrs.toArray(new MailAddress[size]);
                	WrappedMail wmail = new WrappedMail(mail, newAddrs);
                	return wmail;
                }            
            }else{
                //Message was not for the list, return the original message
                return msg;
            }
            
            
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException (e);
        }
    }

    /**
     * does the real work for sendMessage in the event of a "queue" destType
     * @param msg the message to be sent
     * @throws Exception if there is any problem (JMS related or otherwise)
     */
    private void sendMessageQueue(Message msg) throws Exception {
        Mail mail = (Mail)msg;
                
        QueueConnection qc = qcf.createQueueConnection();
        QueueSession qs = qc.createQueueSession(true, 0);
        
        try{
            putMessageOnQueue(qs, destinationQueue, mail);
        } finally{
            qs.commit();
        }
    }    

    private void putMessageOnQueue(QueueSession qs, Queue queue, Mail mail) throws Exception
    {
        QueueSender sender = qs.createSender(queue);
        ObjectMessage om = qs.createObjectMessage(mail);
            
        sender.send(om);
        sender.close();
    }


    /**
     * does the real work for sendMessage in the event of a "topic" destType
     * @param msg the message to be sent
     * @throws Exception if there is any problem (JMS related or otherwise)
     */
    private void sendMessageTopic(Message msg) throws Exception {
        Mail mail = (Mail)msg;
        
        
        TopicConnection tc = tcf.createTopicConnection();
        TopicSession ts = tc.createTopicSession(false, 0);
        
        try{
            putMessageOnTopic(ts, destinationTopic, mail);        
        } finally {
            ts.commit();
        }
    }


    private void putMessageOnTopic(TopicSession ts, Topic topic, Mail mail) throws Exception
    {
        TopicPublisher pub = ts.createPublisher(topic);
        ObjectMessage om = ts.createObjectMessage(mail);
                     
        pub.publish(om);
        pub.close();
    }

    /* (non-Javadoc)
     * @see org.jboss.mail.maillistener.MailListJMSListenerMBean#getMailListManager()
     */
    public ObjectName getMailListManager() {
        return mlmgr;
    }

    /* (non-Javadoc)
     * @see org.jboss.mail.maillistener.MailListJMSListenerMBean#setMailListManager(javax.management.ObjectName)
     */
    public void setMailListManager(ObjectName on) {
        this.mlmgr = on;
    }

}

/**
 * Contains the original mail, with all addresses that were not for maillist
 * @author <a href="kabirkhan@bigfoot.com">Kabir Khan</a>
 */
class WrappedMail extends Mail{
    MailAddress[] to;        
    WrappedMail(){
    	
    }
    WrappedMail(Mail mail, MailAddress[] to){
        super(mail);
        this.to = to;
    }
    
    public MailAddress[] getTo() {
        return to;
    }
}

