/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.maillistener;

import java.util.ArrayList;

import javax.jms.ObjectMessage;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.naming.InitialContext;

import org.apache.log4j.Logger;
import org.jboss.mail.MailListener;
import org.jboss.mail.ServiceProxy;
import org.jboss.mail.ServiceProxyFactory;
import org.jboss.mail.domaingroup.DomainGroup;
import org.jboss.mail.message.Mail;
import org.jboss.mail.message.MailAddress;
import org.jboss.mail.message.Message;
import org.jboss.mail.util.JBMailJMXUtil;
import org.jboss.system.ServiceMBeanSupport;


/**
 * MailListJMSListener - sends mails to a special queue for posting replies to Nukes
 *
 * @author <a href="kabirkhan@bigfoot.com">Kabir Khan</a>
 *
 */
public class NukesPosterJMSMailListener
    extends ServiceMBeanSupport
    implements NukesPosterJMSMailListenerMBean, ServiceProxyFactory, MailListener {

    private static final Logger jblog = Logger.getLogger(NukesPosterJMSMailListener.class);

    /**
     * queue or topic
     */
    private String destType;
    /**
     * destination queue/topic name
     */
    private String destination;
    /**
     * QCF instance from JNDI (null unless destType="queue")
     */
    private QueueConnectionFactory qcf;
    /**
     * TCF instance from JNDI (null unless destType="topic")
     */
    private TopicConnectionFactory tcf;
    /**
     * destination queue (null unless destType="queue")
     */
    private Queue destinationQueue;
    /**
     * destination topic (null unless destType="topic")
     */
    private Topic destinationTopic;

    private String queueConnectionFactoryName;
    private String topicConnectionFactoryName;

    /**
     *  Name of the NukesPoster MBean
     */
    private ObjectName nukesPosterName;

    /**
     *  Name of the DomainGroup MBean
     */
    private ObjectName domainGroupName;

    /**
     *  DomainGroup ServiceProxy
     */
    private DomainGroup domainGroup;

    

    public NukesPosterJMSMailListener(String queueConnectionFactoryName, String topicConnectionFactoryName) {
        this.queueConnectionFactoryName = queueConnectionFactoryName;
        this.topicConnectionFactoryName = topicConnectionFactoryName;
    }

    /* (non-Javadoc)
     * @see org.jboss.mail.ServiceProxyFactory#produceServiceProxy()
     */
    public ServiceProxy produceServiceProxy() {
        JMXMailListenerProxy proxy = new JMXMailListenerProxy(this.getServiceName());
        return proxy;
    }

    public void startService() throws Exception {
        InitialContext ctx = new InitialContext();
        if (destType.equals("topic")) {
          //TODO: make configurable
          tcf = (TopicConnectionFactory)ctx.lookup(topicConnectionFactoryName);
          destinationTopic = (Topic)ctx.lookup("topic/"+destination);
        } else {
          //TODO: make configurable
          qcf = (QueueConnectionFactory)ctx.lookup(queueConnectionFactoryName);
          destinationQueue = (Queue)ctx.lookup("queue/"+destination);
        }
        
        domainGroup = (DomainGroup)JBMailJMXUtil.getService(domainGroupName);
    }

    public void stopService() throws Exception {

    }

   public void setDestinationType(String destType) {
        this.destType = destType;
    }

    public String getDestinationType() {
        return this.destType;
    }

    public void setDestination(String postDest) {
        this.destination = postDest;
    }


    public String getDestination() {
        return destination;
    }

    public void setNukesForumPoster(ObjectName name){
        nukesPosterName = name;
    }

    public ObjectName getNukesForumPoster(){
        return nukesPosterName;
    }

    public void setDomainGroup(ObjectName on){
        domainGroupName = on;
    }

    public ObjectName getDomainGroup(){
        return domainGroupName;
    }
    
    public Message send(Message msg) {
        log.debug("send()");

        try {
            MBeanServer server = JBMailJMXUtil.locateJBoss();
            Mail mail = (Mail)msg;
               MailAddress fromAddr = mail.getFrom();
            if (fromAddr.isEmpty()){
                //This is a bounce message, ignore
                //return false;
                return msg;
            }

            MailAddress[] toAddrs = mail.getTo();
            ArrayList toUsers = new ArrayList();
            //for each entry in toUsers, contains the index of the corresponding entry in toAddrs  
            ArrayList usersLookup = new ArrayList();

            ArrayList notForNukes = new ArrayList();
            
            //Get the usernames for all local users
            for (int i = 0 ; i < toAddrs.length ; i++){
                if (domainGroup.isInGroup(toAddrs[i].getDomain())){
                    toUsers.add(toAddrs[i].getUser());
                    usersLookup.add(new Integer(i));
                }
                else {
                	notForNukes.add(toAddrs[i]);
                }
            }
            
            String[] users = (String[])toUsers.toArray(new String[toUsers.size()]);
            String[] usersForNukes = getUsersForNukes(users);
            
            if (usersForNukes.length == 1)
            {
                //Only makes sense to be able to post to one nukes topic at once
                sendMessageToNukes(toAddrs, mail, users, usersForNukes, usersLookup);
                
                if (usersForNukes.length == users.length){
               	//We handled all the addresses
                	//TODO: return null, or WrappedMail with empty to array?
                	return null;
                }else{
                	//Create new Mail with all addresses not for list for returning
                	return getMailForChain(toAddrs, mail, users, usersForNukes, usersLookup, notForNukes);
                }
            }
            else{
                //Nothing for Nukes, return original mail
                return msg;
            }
            	
            
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException (e);
        }        
    }
    
    private String[] getUsersForNukes(String[] to) throws Exception{
          
          if (to.length == 0){
              return new String[0];
          }
          
          Object[] params = new Object[]{to};
          String[] sig = new String[]{to.getClass().getName()};
          return (String[])server.invoke(nukesPosterName, "getUsersForNukes", params, sig);
      }

    private void sendMessageToNukes(MailAddress[] toAddrs, Mail mail, String[] users, String[] usersForNukes, ArrayList usersLookup)throws Exception{
        MailAddress nukesAddr = null;
        for (int i = 0 ; i < users.length ; i++){
            if (users[i].equals(usersForNukes[0])){
                int index = ((Integer)usersLookup.get(i)).intValue();                    	
                nukesAddr = toAddrs[index];
            }
        }

        //Create mail for Nukes
        NukesWrappedMail wmail = new NukesWrappedMail(mail, new MailAddress[]{nukesAddr});
        if (destType.equals("topic")) {
            sendMessageTopic(wmail);
        } else {
            sendMessageQueue(wmail);
        }
    }

    private Mail getMailForChain(MailAddress[] toAddrs, Mail mail, String[] users, String[] usersForNukes, ArrayList usersLookup, ArrayList notForNukes)
    {
        for (int i = 0 ; i < users.length ; i++){
            boolean found = false;
            if (usersForNukes[0].equals(users[i])){
                found = true;
            }
                    
            if (!found){
                int index = ((Integer)usersLookup.get(i)).intValue();                    	
                notForNukes.add(toAddrs[index]);
            }
        }
            	
            	
        MailAddress[] newAddrs = (MailAddress[])notForNukes.toArray(new MailAddress[notForNukes.size()]);
        return new NukesWrappedMail(mail, newAddrs);
    }
    
    /**
     * does the real work for sendMessage in the event of a "queue" destType
     * @param msg the message to be sent
     * @throws Exception if there is any problem (JMS related or otherwise)
     */
    private void sendMessageQueue(Message msg) throws Exception {
        Mail mail = (Mail)msg;

        QueueConnection qc = qcf.createQueueConnection();
        QueueSession qs = qc.createQueueSession(true, 0);

        try{
            putMessageOnQueue(qs, destinationQueue, mail);
        } finally{
            qs.commit();
        }
    }

    private void putMessageOnQueue(QueueSession qs, Queue queue, Mail mail) throws Exception
    {
        QueueSender sender = qs.createSender(queue);
        ObjectMessage om = qs.createObjectMessage(mail);

        sender.send(om);
        sender.close();
    }


    /**
     * does the real work for sendMessage in the event of a "topic" destType
     * @param msg the message to be sent
     * @throws Exception if there is any problem (JMS related or otherwise)
     */
    private void sendMessageTopic(Message msg) throws Exception {
        Mail mail = (Mail)msg;


        TopicConnection tc = tcf.createTopicConnection();
        TopicSession ts = tc.createTopicSession(false, 0);

        try{
            putMessageOnTopic(ts, destinationTopic, mail);
        } finally {
            ts.commit();
        }
    }


    private void putMessageOnTopic(TopicSession ts, Topic topic, Mail mail) throws Exception
    {
        TopicPublisher pub = ts.createPublisher(topic);
        ObjectMessage om = ts.createObjectMessage(mail);

        pub.publish(om);
        pub.close();
    }
}

class NukesWrappedMail extends Mail{
    MailAddress[] to;        
    NukesWrappedMail(Mail mail, MailAddress[] to){
        super(mail);
        this.to = to;
    }
    
    public MailAddress[] getTo() {
        return to;
    }
}    
