/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.message;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.jboss.logging.Logger;
import org.jboss.mail.smtp.SMTPProtocolInstance;

/**
 * SMTP uses a Mail message.  This is him.
 *
 * @author Andrew C. Oliver
 * @author Eric Daugherty
 */
public class Mail implements Message, Serializable{
    private Logger log = Logger.getLogger(Mail.class);

    MailHeaders ih;
    /**
     * envelope from address
     */
    MailAddress from;
    /**
     * envelope to addresses
     */
    List to;

    /**
     * body of mail after the headers are parsed
     */
    byte[] body;

    int retries = 0;
    
    static private class Buffer {
	private int length;
	private byte[] data;
	private Buffer(int length, byte[] data) {
	    this.length = length;
	    this.data = data;
	}
    }

    protected Mail(){
    	
    }
    
    protected Mail(Mail mail){
        this.body = mail.body;
        this.from = mail.from;
        this.ih = mail.ih;
        this.retries = mail.retries;
        this.to = mail.to;
    }
    
    /**
     * Create a org.jboss.org.jboss.mail off of a given input stream
     * @param msgIn the input stream we'll read the mail off of
     * @param buffersize buffer size (block size of sorts) - ideally close to the size of the message, maybe a little bigger
     * @param from who is sending it
     * @param to who all they are sending it to
     * @param cc who is being copied on it
     * @param bcc who is being blind cc'd on it
     * @param smtpProtocol instance
     */
    private Mail(
        InputStream msgIn,
        int buffersize,
        MailAddress from,
        MailAddress[] to,
        SMTPProtocolInstance smtpProtocol) throws LoopDetectedException{
    	List bodyList = new ArrayList();
        log.debug("MAIL constructor called");
        try {
            this.from = from;
            //this.to = new ArrayList(to.length);
            this.to = new ArrayList(Arrays.asList(to));
            ih = smtpProtocol.getMailHeaders();
            ih.load(msgIn, smtpProtocol, from, to);


	    log.info("all headers after loading: " + ih);

	    // MIKEA: i implemented the following differently, since i noticed
	    // the recipient would get an email with lots
	    // of binary '0's at the end of the message (presumably total size
	    // in integral bunches of 'buffersize').
	    // so basically i add Buffer objects to the list, instead of byte[],
	    // to keep track of the actual amount of data.
	    if (false) {
		// ORIGINAL FAULTY IMPLEMENTATION (DON'T USE)
		byte[] b = new byte[buffersize];
		while (msgIn.read(b) > 0) {
		    bodyList.add(b);
		    b = new byte[buffersize];
		}
	    } else {
		boolean done = false;
		byte[] b = new byte[buffersize];
		while (!done) {
		    int len = msgIn.read(b);
		    if (len > 0) {
			bodyList.add(new Buffer(len,b));
			b = new byte[buffersize];
		    } else {
			done = true;
		    }
		}
	    }

            envelopeTos();
            ih.removeHeader("To:");
            ih.removeHeader("CC:");
            body = new byte[calculateBodySize(bodyList)];
            body = makeBody(bodyList,body);
		} catch (LoopDetectedException e) {
			throw e;
		} catch (Exception e) {
			log.error(e.toString(),e);
			throw new RuntimeException("Error",e);
		}
        if (log.isDebugEnabled()) {
            debugPrintHeaders(ih);
        }

    }

    /**
     * determines the enveloping based on the headers.
     */
    private void envelopeTos() {
        String[] tos = ih.getHeader(StandardMailHeaders.TO);
        String[] ccs = ih.getHeader(StandardMailHeaders.CC);
        tos = tos != null ? parseMailHeaderAddress(tos) : null;
        ccs = ccs != null ? parseMailHeaderAddress(ccs) : null;
        if (tos == null) {
            tos = new String[0];
        }
        if (ccs == null) {
            ccs = new String[0];
        }
        EnvelopedAddress ea = null;
        for (int k = 0; k < this.to.size(); k++) {
            if(isIn(tos, (MailAddress)to.get(k))) {
                ea = new EnvelopedAddress((MailAddress)to.get(k), javax.mail.Message.RecipientType.TO);
                ea.setRcpt(true);
                to.set(k, ea);
            } else if (isIn(ccs, (MailAddress)to.get(k))) {
                ea = new EnvelopedAddress((MailAddress)to.get(k),javax.mail.Message.RecipientType.CC);
                ea.setRcpt(true);
                to.set( k, ea );
            } else {
                ea = new EnvelopedAddress((MailAddress)to.get(k),javax.mail.Message.RecipientType.BCC);
                ea.setRcpt(true);
                to.set(k, ea);
            }
        }        
    }


    /**
     * parse the addresses from the set of a header (all TOs for instance)
	 * @param strings from TO, CC, or BCC header
	 * @return addresses as seperate strings
	 */
	private String[] parseMailHeaderAddress(String[] strings) {
		List stringsList = new ArrayList();
		for(int k = 0; k <strings.length; k ++){
			String[] temp = parseMailHeaderAddress(strings[k]);
			for (int x = 0; x <temp.length; x++) {
				stringsList.add(temp[x]);
			}
		}
		return (String[])stringsList.toArray(new String[stringsList.size()]);
	}

	/**
	 * parse only ONE string from the TOs
	 * @param string containing multiple addresses
	 * @return string array of the parsed addresses
	 */
	private String[] parseMailHeaderAddress(String string) {
		String[] result = string.split("\\,\\s");
		for (int k = 0; k < result.length; k++) {
			result[k] = result[k].trim();
		}

		return result;
	}


	/**
	 * determine if the passed in address is in the list of addresses.
     * @param to list of addresses
     * @param address to check for
     * @return present or not
     */
    private boolean isIn(List tos, String address) {
        for (int k = 0; k < tos.size(); k++) {
            MailAddress temp = MailAddress.parseSMTPStyle(address);
            if (temp.equals((MailAddress)tos.get(k))) {
                return true;
            }
        }
        return false;
    }


    /**
	 * determine if the passed in address is in the array of addresses.
     * @param tos array of addresses
     * @param address to check for
     * @return present or not
     */
    private boolean isIn(String[] tos, MailAddress address) {
        for (int k = 0; k < tos.length; k++) {
            MailAddress temp = MailAddress.parseSMTPStyle(tos[k]);
            if (temp.equals(address)) {
                return true;
            }
        }
        return false;
    }


    /**
     * create a contiguous body from a list of byte arrays
	 * @param bodyList list of byte arrays
	 * @param body contiguous byte array equal to cumulative size of byte arrays
	 * @return body
	 */
	private byte[] makeBody(List bodyList, byte[] body) {
		int offset = 0;
		for (int k = 0; k < bodyList.size(); k++) {
		    Buffer buf = (Buffer) bodyList.get(k);
		    byte[] b = buf.data;
			System.arraycopy(b,0,body,offset,buf.length);
			offset += buf.length;
		}
        log.debug("MAIL BODY: "+new String(body));
		return body;
	}

	/**
	 * calculate the total size of the body from the list.
	 * @param bodyList containing byte arrays
	 * @return size for the body list
	 */
      private int calculateBodySize(List bodyList) {
	int retval = 0;
	for (int k = 0; k < bodyList.size(); k++) {
	    retval += ((Buffer)bodyList.get(k)).length;
	}
	return retval;
    }

	/**
     * Static constructor for org.jboss.org.jboss.mail objects.  Later we may pool these, for now it just calls the
     * private constructor.
     * @param msgIn input stream containing the org.jboss.org.jboss.mail, post DATA command
     * @return Mail object
     */
    public static Mail create(
        InputStream msgIn,
        int buffersize,
        MailAddress from,
        MailAddress[] to,
        SMTPProtocolInstance smtpProtocol)  throws LoopDetectedException{
        //log.debug("Mail CREATE called");
        Mail mail = new Mail(msgIn, buffersize, from, to, smtpProtocol);
        return mail;
    }

    /**
     * should print the org.jboss.org.jboss.mail headers.
     * @param ih internet headers object parsed from the input
     */
    private void debugPrintHeaders(MailHeaders ih) {
        java.util.Iterator e = ih.getAllHeaderLines();
        while (e.hasNext()) {
            log.debug((String) e.next());
        }
    }

    /**
     * WARNING: sender is returned by copy, thus modifications do not affect the original object
     * @return sends or the mail
     */
    public MailAddress getFrom() {
        MailAddress from = new MailAddress(this.from);
        return from;
    }

    /*
     * WARNING: receipients are returned by copy, thus modifications do not affect the original object.
     * @return copy of the receipients
     */
//    public MailAddress[] getTo() {
//        MailAddress[] to = new MailAddress[this.to.size()];
//        for (int k = 0; k < to.length; k++) {
//            if (this.to.get(k) instanceof EnvelopedAddress) {
//                to[k] = new EnvelopedAddress((EnvelopedAddress)this.to.get(k));
//            } else {
//                to[k] = new MailAddress((MailAddress)this.to.get(k));
//            }
//        }
//        return to;
//    }

    public MailAddress[] getTo() {
        return (MailAddress[])to.toArray(new MailAddress[to.size()]);
    }

    /**
     * Returns the MailHeaders that contain all the individual header entries.
     *
     * @return the original MailHeaders instance.
     */
    public MailHeaders getMailHeaders() {
        // TODO: Should this be cloned?
        return ih;
    }

    /**
     * WARNING: receipients are returned by copy, thus modifications do not affect the original object.
     * @return copy of the body
     */
    public byte[] getBody() {
    	byte[] retval = new byte[body.length];
    	System.arraycopy(body,0,retval,0,body.length);
    	return retval;
    }
    
	public void increaseRetries(){
		retries++;
	}
	
	public int getRetryNumber(){
		return retries;
	}
    
    public byte[] getRawBytes()throws IOException{
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		BufferedOutputStream out = new BufferedOutputStream(baos);
		MailHeaders headers = getMailHeaders();
		
		for (Iterator it = headers.getAllHeaderLines(); it.hasNext() ; ){
			//Headers should be encoded in US-ASCII
			out.write(it.next().toString().getBytes("US-ASCII"));
			out.write("\r\n".getBytes());
		}
		
		out.write("\r\n".getBytes());
		out.write(body);
		out.close();
		return baos.toByteArray();
    }
}
