/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.message;

import java.io.Serializable;

/**
 * Represents a org.jboss.org.jboss.mail address, be it from, to, cc or otherwise
 * @author Andrew C. Oliver
 */
public class MailAddress implements Serializable {
    private String prettyName;
    /**
     * User (that which preceeds the at symbol)
     */
    private String user;
    /**
     * domain (after the at symbol)
     */
    private String domain;
    /**
     * is this user actually valid or is this address bogus (generally syntax related)
     */
    private boolean isvalid;

	/**
	 * Whether this address is empty or not
	 */
	private boolean isempty;
    /**
     * Construct a new org.jboss.org.jboss.mail address for the user and domain, not to be called by the 
     * client class, but rather the result of parsing an address
     * @param user from standard org.jboss.org.jboss.mail string preceeding the at symbol
     * @param domain from standard org.jboss.org.jboss.mail string following the at symbol
     */
    private MailAddress(String user, String domain, boolean isfrom) {
        this.user = user;
        this.domain = domain;
        if (user != null
            && user.length() > 0
            && domain != null
            && domain.length() > 0) {
            isvalid = true;
        }
        
        if ((user == null || user.length() == 0) 
        	&& (domain == null || domain.length() == 0)){
        	isempty = true;        	
        	isvalid = isfrom;
        }
    }
    
    /**
     * Construct a new org.jboss.org.jboss.mail address for the user and domain, not to be called by the 
     * client class, but rather the result of parsing an address
     * @param user from standard org.jboss.org.jboss.mail string preceeding the at symbol
     * @param domain from standard org.jboss.org.jboss.mail string following the at symbol
     * @param prettyName the name before the &gt; &lt;
     */
    private MailAddress(String user, String domain, String prettyName, boolean isfrom) {
        this(user, domain, isfrom);
        this.prettyName = prettyName;
    }
    
    /**
     * copy constructor.  Will create a new copy of Mail Address including validity.
     * @param orig source MailAddress object
     */
    public MailAddress(MailAddress orig) {
        this.user = new String(orig.getUser());
        this.domain = new String(orig.getDomain());
        this.prettyName = orig.getPrettyName() != null ? new String(orig.getPrettyName()) : null;
        this.isvalid = orig.isValid();
        this.isempty = orig.isempty;
    }

    /**
     * @return the name before &lt;bla@bla.bla&gt;
     */
    private String getPrettyName() {
        return prettyName;
    }

	/**
	 * If a valid address is used to construct the instance, it should be valid.  If the parsing fails 
	 * for some reason (say no at symbol) then this will be false.
	 * @return whether this instance is valid
	 */
	public boolean isValid() {
		return isvalid;
	}

	/**
	 * If a valid address is used to construct the instance, it should be valid.  If the parsing fails 
	 * for some reason (say no at symbol) then this will be false.
	 * @return whether this instance is valid
	 */
	public boolean isEmpty() {
		return isempty;
	}

    /**
     * @return username (preceeding the at sign)
     */
    public String getUser() {
        return user;
    }

    /**
     * @return domain name (following the at sign)
     */
    public String getDomain() {
        return domain;
    }

    /**
     * parse a standard org.jboss.org.jboss.mail string from at domain.com and return a org.jboss.org.jboss.mail address.
     * @param addr a string representing a org.jboss.org.jboss.mail address
     * @return MailAddress object representing the passed in string.
     */
    public static MailAddress parseSMTPStyle(String addr) {
    	boolean isfrom =  false;
    	return parseSMTPStyle(addr, isfrom);
    }
    	
    public static MailAddress parseSMTPStyle(String addr, boolean isfrom) {
        String address = addr;
        int atloc = 0;
        MailAddress retval = null;
        String prettyName = null;
        try {
        if (address.indexOf("<") > 0) {
            prettyName = address.substring(0,address.indexOf("<"));
            address = address.substring(address.indexOf("<"),address.length());
        }
        if (address.charAt(0) == '<' && address.charAt(address.length()-1) == '>') {
            address = address.substring(1, address.length() - 1);

            if (address.length() == 0) {
            } else {
                if (address.indexOf("@") < 0) {
                    address = address + "@localhost";
                }
            }

            atloc = address.indexOf("@");
        } else if (address.indexOf("<") < 0 && address.indexOf("@") > 0) {
            atloc = address.indexOf("@");
        }

        
            retval =  new MailAddress(
                address.substring(0, atloc),
                address.substring(atloc + 1, address.length()),
                prettyName,
                isfrom);
        } catch (Exception e) {
            retval = new MailAddress("","", isfrom);    
        }
        
        return retval;
    }
    
    /**
     * determines if this mail address represents the same as another address.  This is a simple textual compare,
     * we have no idea if first@jboss.org is the same as First.Last@jboss.org or if first@jboss.org is the same as 
     * first@mail.jboss.org.  
     * @param address to test
     * @return true if its equal or false if it isnt
     */
	public boolean equals(MailAddress address) {
		if (this.getUser().equals(address.getUser()) &&
			this.getDomain().equals(address.getDomain())) {
			return true;
		}
		return false;
	} 
    
    
    //TODO: make support prettyNames 
    public String toString() {
    	if (isempty){
    		return "<>";
    	}
    	else{
			return "<"+getUser()+"@"+getDomain()+">";
    	}
    }
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		//Added to be able to use Set functionality
		return user.hashCode();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		//Added to be able to use Set functionality
		if (obj instanceof MailAddress)
			return equals((MailAddress)obj);
		else
			return false;
	}

}
