/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.message;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import javax.mail.Header;
import javax.mail.MessagingException;

import org.jboss.mail.ServerThread;
import org.jboss.mail.smtp.SMTPConstants;
import org.jboss.mail.smtp.SMTPProtocolInstance;
import org.jboss.mail.util.DateUtil;

import com.sun.mail.util.LineInputStream;



/**
 * Implementation for basic mail-header functionality.
 * @author Michael Krause 
 * @author Andrew C. Oliver <acoliver ot jboss dat org>
 */
public class MailHeadersImpl implements MailHeaders, Serializable
{
    private List headers;
    
    
    
    

    public MailHeadersImpl()
    {
        headers = new ArrayList();
    }

    public MailHeadersImpl(InputStream inputstream)
    throws MessagingException
    {
        headers = new ArrayList();
        //KAB - I was not sure if this constructor should take an SMTPProtocol and 
        //MailAddress for calling the modifed load() method, or if I 
        //should call the internal loadFromStream() method (the old load()
        //method). This constructor does not seem to be called thoug, so it shouldn't
        //matter too much.        
        loadFromStream(inputstream);
    }

    public String toString() {
	return "MailHeadersImpl ("+headers.size()+"): " + headers.toString();
    }

	String localHostAddress; 
	String remoteHost;
	String remoteIP;
	String serverName;

    public void load(InputStream inputstream, 
    		SMTPProtocolInstance smtpProtocol, 
			MailAddress from,
			MailAddress[] to)
	throws MessagingException, LoopDetectedException
	{
		//Initialise addressing members
		this.serverName = (String)smtpProtocol.getProperty(SMTPConstants.SERVER_NAME);
		InetAddress localAddress = (InetAddress)smtpProtocol.getState(ServerThread.STATE_LOCAL_ADDRESS);
		InetAddress remoteAddress = (InetAddress)smtpProtocol.getState(ServerThread.STATE_CLIENT_ADDRESS);
		this.remoteHost = remoteAddress != null ? remoteAddress.getHostName() : "unknown";
		this.remoteIP = remoteAddress != null ? remoteAddress.getHostAddress() : "unknown";
		this.localHostAddress = localAddress.getHostAddress();

		//Load the headers
    	loadFromStream(inputstream);
    	checkForLoops(smtpProtocol);
        addAdditionalHeaders(smtpProtocol, from, to);
        //Is this necessary?
    	ensureTraceHeadersAtStart();
    }

	/** Checks for loops in headers (6.2 in rfc 2821)
	 * 
	 * @param smtpProtocol The SMTPProtocolInstance 
	 * @throws LoopDetectedException If a loop is detected
	 */
	private void checkForLoops(SMTPProtocolInstance smtpProtocol)throws LoopDetectedException{
		//Check for loops in headers            
		String[] receivedHeaders = getHeader(StandardMailHeaders.RECEIVED);

		if (receivedHeaders != null)
		{
			if (receivedHeaders.length > smtpProtocol.getPropertyLong(SMTPConstants.MAX_RECEIVED_HEADERS)){
				throw new LoopDetectedException("SMTP loop? Too many Received headers. Max allowed is "	
							+ smtpProtocol.getPropertyLong(SMTPConstants.MAX_RECEIVED_HEADERS));
			}
			
			if (receivedHeaders.length > smtpProtocol.getPropertyLong(SMTPConstants.RECEIVED_HEADERS_THRESHOLD)){
				
				//Over the threshold for self inspection, see if we occur more than allowed
				long maxOwn = smtpProtocol.getPropertyLong(SMTPConstants.MAX_OWN_RECEIVED_HEADERS);
	            int count = 0;
				for (int i = 0 ; i < receivedHeaders.length ; i++){
					if ( receivedHeaders[i].indexOf(serverName) > -1 
							&& receivedHeaders[i].indexOf(localHostAddress) > -1 ){
						count++;
						
						if (count > maxOwn){
							throw new LoopDetectedException(
													"SMTP loop? More than " 
													+ maxOwn 
													+ " Received headers contain the name of our server: "
													+ serverName);
						}
					}
				}
			}			
		}
	}

    private void loadFromStream(InputStream inputstream)
    throws MessagingException
    {
        LineInputStream lineinputstream = new LineInputStream(inputstream);
        String s1 = null;
        StringBuffer stringbuffer = new StringBuffer();
        boolean first = true;
        do
            try
            {
            String s = lineinputstream.readLine();
            if( first && s.length() == 0 ) {
                s = lineinputstream.readLine();
                first = false;
            }
            if(s != null && (s.startsWith(" ") || s.startsWith("\t")))
               {
                if(s1 != null)
                   {
                    stringbuffer.append(s1);
                    s1 = null;
                }
                stringbuffer.append("\r\n");
                stringbuffer.append(s);
            } else
               {
                if(s1 != null)
                    addHeaderLine(s1);
                else
                    if(stringbuffer.length() > 0)
                       {
                        addHeaderLine(stringbuffer.toString());
                        stringbuffer.setLength(0);
                    }
                s1 = s;
            }
            if(s == null)
                break;
            if(s.length() <= 0)
                return;
        }
        catch(IOException ioexception)
        {
            throw new MessagingException("Error in input stream", ioexception);
        }
        while(true);
    }
    
    /** The 'From' and 'Date' mail headers are required by rfc 2822 (3.6),
     * these are added if they don't already exist.
     * Also adds the "Return-Path" and "Received" trace headers.
     * @param smtpProtocol The SMTPProtocolInstance used to generate this mail     */
    private void addAdditionalHeaders(SMTPProtocolInstance smtpProtocol, MailAddress from, MailAddress[] to){
    	String date = DateUtil.mailHeaderDate();
    	
		//Add the required fields if this is the originating mail system
		if (smtpProtocol.getState(SMTPConstants.USER) != null){
			//Add the From and  Date fields required by RFC 2821 (3.6)
			if (getHeader(StandardMailHeaders.FROM) == null){
				addHeader(StandardMailHeaders.FROM, from.toString());
			}
			if (getHeader(StandardMailHeaders.DATE) == null) {
				addHeader(StandardMailHeaders.DATE, date);
			}
			
			// Determine the Return-Path and add that RFC 2821 (7.2)
			String[] returnPath = getHeader(StandardMailHeaders.RETURN_PATH);
			if (returnPath == null){
				addHeader(StandardMailHeaders.RETURN_PATH, from.toString());
			}
		}

		//Create the Received header for this server
        //There is quite a lot of vague information on Received in a load of rfs's,
        //and SMTP servers provide a load of different implementations of this, so I 
        // went with one of the options in RFC 2821 (7.2)
        //At the end of the day it is meant to be humanly readable.

        //Generate an id (should this be stored somewhere??)
        StringBuffer id = new StringBuffer();
        id.append(System.currentTimeMillis());
        id.append(Math.random() * 1024);
        
        //TODO: If any smtp extensions have been used, protocol should be ESMTP
        String protocol = "SMTP";
        
        StringBuffer received = new StringBuffer();       
        received.append("from " + smtpProtocol.getState(SMTPConstants.HELO_DOMAIN));
        received.append(" (" + remoteHost + " " + remoteIP + ")");
        received.append(" by " + smtpProtocol.getProperty(SMTPConstants.SERVER_NAME));
        received.append(" (" + localHostAddress + ")");
        received.append("\r\n\twith SMTP id " + id.toString());
		received.append("; " + date);
        addHeader(StandardMailHeaders.RECEIVED, received.toString());
    }

    /** Makes sure that the first header is the Return-Path, followed 
     * by all Received headers, followed by any passed in Return headers
     * followed by all the other headers.
     */
    private void ensureTraceHeadersAtStart(){
		//Is this necessary?
		//From rfc 2822 (3.6): "the trace header fields and resent
		//header fields MUST NOT be reordered, and SHOULD be kept in blocks
		//prepended to the message".

		boolean returnPath = false;
		int size = headers.size();
		int receivedIndex = 0;
		
		for (int i = 0 ; i < size ; i++){
			hdr hdr1 = (hdr)headers.get(i);
			
			if (hdr1.name.equalsIgnoreCase(StandardMailHeaders.RETURN_PATH)){
				if (i == 0){
				}
				else{
					headers.remove(i);
					headers.add(0, hdr1);
				}
				receivedIndex++;
				returnPath = true;
			}
			else if (hdr1.name.equalsIgnoreCase(StandardMailHeaders.RECEIVED)){
				if (i != receivedIndex){
					headers.remove(i);
					headers.add(receivedIndex, hdr1);
				}
				receivedIndex++;
			}
		}

    }

    public String[] getHeader(String s)
    {
        Iterator iterator = headers.iterator();
        List list = new ArrayList();
        while(iterator.hasNext()) 
           {
            hdr hdr1 = (hdr)iterator.next();
            if(s.equalsIgnoreCase(hdr1.name) && hdr1.line != null)
                list.add(hdr1.getValue());
        }
        if(list.size() == 0)
           {
            return null;
        } else
           {
            String as[] = new String[list.size()];
            as = (String[])list.toArray(as);
            return as;
        }
    }

    public String getHeader(String s, String s1)
    {
        String as[] = getHeader(s);
        if(as == null)
            return null;
        if(as.length == 1 || s1 == null)
            return as[0];
        StringBuffer stringbuffer = new StringBuffer(as[0]);
        for(int i = 1; i < as.length; i++)
           {
            stringbuffer.append(s1);
            stringbuffer.append(as[i]);
        }

        return stringbuffer.toString();
    }

    public void setHeader(String s, String s1)
    {
        boolean flag = false;
        for(int i = 0; i < headers.size(); i++)
           {
            hdr hdr1 = (hdr)headers.get(i);
            if(s.equalsIgnoreCase(hdr1.name))
                if(!flag)
                   {
                    int j;
                    if(hdr1.line != null && (j = hdr1.line.indexOf(':')) >= 0)
                        hdr1.line = hdr1.line.substring(0, j + 1) + " " + s1;
                    else
                        hdr1.line = s + ": " + s1;
                    flag = true;
                } else
                   {
                    headers.remove(i);
                    i--;
                }
        }

        if(!flag)
            addHeader(s, s1);
    }

    public void addHeader(String s, String s1)
    {
        int i = headers.size();
        boolean flag = s.equalsIgnoreCase("Received");
        if(flag)
            i = 0;
        for(int j = headers.size() - 1; j >= 0; j--)
           {
            hdr hdr1 = (hdr)headers.get(j);
            if(s.equalsIgnoreCase(hdr1.name))
                if(flag)
                   {
                    i = j;
                } else
                   {
                    headers.add(j + 1,new hdr(s, s1));
                    return;
                }
            if(hdr1.name.equals(":"))
                i = j;
        }

        headers.add(i,new hdr(s, s1));
    }

    public void removeHeader(String s)
    {
        for(int i = 0; i < headers.size(); i++)
           {
            hdr hdr1 = (hdr)headers.get(i);
            if(s.equalsIgnoreCase(hdr1.name))
                hdr1.line = null;
        }

    }

    public Iterator getAllHeaders()
    {
        return new matchEnum(headers, null, false, false);
    }


    public void addHeaderLine(String s)
    {
        try
        {
            char c = s.charAt(0);
            if(c == ' ' || c == '\t')
               {
                hdr hdr1 = (hdr)headers.get(headers.size()-1);
                hdr1.line += "\r\n" + s;
                return;
            } else
               {
                headers.add(new hdr(s));
                return;
            }
        }
        catch(StringIndexOutOfBoundsException _ex)
        {
            return;
        }
        catch(NoSuchElementException _ex)
        {
            return;
        }
    }

    public Iterator getAllHeaderLines()
    {
        return getNonMatchingHeaderLines(null);
    }

    public Iterator getMatchingHeaderLines(String as[])
    {
        return new matchEnum(headers, as, true, true);
    }

    public Iterator getNonMatchingHeaderLines(String as[])
    {
        return new matchEnum(headers, as, false, true);
    }
    
}


class matchEnum
    implements Iterator, Serializable
{

    private Iterator e;
    private String names[];
    private boolean match;
    private boolean want_line;
    private hdr next_header;

    matchEnum(List list, String as[], boolean flag, boolean flag1)
    {
        e = list.iterator();
        names = as;
        match = flag;
        want_line = flag1;
        next_header = null;
    }

    public boolean hasMoreElements()
    {
        if(next_header == null)
            next_header = nextMatch();
        return next_header != null;
    }

    public Object nextElement()
    {
        if(next_header == null)
            next_header = nextMatch();
        if(next_header == null)
            throw new NoSuchElementException("No more headers");
        hdr hdr1 = next_header;
        next_header = null;
        if(want_line)
            return hdr1.line;
        else
            return new Header(hdr1.getName(), hdr1.getValue());
    }

    private hdr nextMatch()
    {
label0:
        while(e.hasNext())
        {
            hdr hdr1 = (hdr)e.next();
            if(hdr1.line == null)
                continue;
            if(names == null)
                if(match)
                    return null;
                else
                    return hdr1;
            for(int i = 0; i < names.length; i++)
            {
                if(!names[i].equalsIgnoreCase(hdr1.name))
                    continue;
                if(match)
                    return hdr1;
                continue label0;
            }

            if(!match)
                return hdr1;
        }
        return null;
    }

    public boolean hasNext() {
        return hasMoreElements();
    }

    public Object next() {
        return nextElement();
    }

    public void remove() {
        e.remove();
        
    }
}


class hdr implements Serializable
{

    String name;
    String line;

    hdr(String s)
    {
        int i = s.indexOf(':');
        if(i < 0)
            name = s.trim();
        else
            name = s.substring(0, i).trim();
        line = s;
    }

    hdr(String s, String s1)
    {
        name = s;
        if(s1 != null)
        {
            line = s + ": " + s1;
            return;
        } else
        {
            line = null;
            return;
        }
    }

    String getName()
    {
        return name;
    }

    String getValue()
    {
        int i = line.indexOf(':');
        if(i < 0)
            return line;
        int j;
        for(j = i + 1; j < line.length(); j++)
        {
            char c = line.charAt(j);
            if(c != ' ' && c != '\t' && c != '\r' && c != '\n')
                break;
        }

        return line.substring(j);
    }

    public String toString() {
	return "hdr("+name+" = '"+line+"')";
    }

}
