/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.pop3;

import java.util.Map;
import java.util.List;
import java.util.StringTokenizer;
import java.io.OutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.InputStream;
import java.net.Socket;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;

import org.jboss.mail.*;
import org.jboss.mail.mailbox.Folder;
import org.jboss.logging.Logger;
import org.jboss.mail.smtp.SMTPProtocolInstance;
import org.jboss.mail.util.DateUtil;
import org.jboss.mail.pop3.handlers.POP3Handler;
import org.w3c.dom.Element;

/**
 * Service Proxy created by the MBean which is used by the Server (actually ServerThread) to implement
 * the protocol.
 *
 * @author Eric Daugherty
 */
public class POP3ProtocolInstance
         extends AbstractProtocol
    implements Protocol, ServiceProxy {

    private Logger log = Logger.getLogger(POP3ProtocolInstance.class);

    //
    // The possible connection states
    //

    /**
     * The connection is in AUTHORIZATION state until
     * the user has successfully authenticated with the server
     */
    public static final int STATE_AUTHORIZATION = 0;

    /**
     * The TRANSACTION state allows the user to execute
     * all commands to manipulate the mailbox state.  No
     * changes are saved until it enters the UPDATE state.
     */
    public static final int STATE_TRANSACTION = 1;

    /**
     * The connection enters teh UPDATE state after
     * the QUIT command.  During this state, the mailbox changes
     * are persisted.
     */
    public static final int STATE_UPDATE = 2;


    /**
     * returned from SMTPHandlers.instance();
     */
    private Map handlers;

    /**
     * The current state of the connection.  Values
     * are: AUTHORIZATION, TRANSACTION, and UPDATE
     */
    private int state;

    /**
     * The username for the current session.
     */
    private String username = null;

    /**
     * The user's ROOT folder.
     */
    private Folder mailbox = null;

    /**
     * create an POP3ProtocolInstance with the proper handlers and the already
     * constructed properties map
     * @param handlers from POP3Handlers.instance();
     * @param properties from POP3Protocol.setProperties()
     */
    public POP3ProtocolInstance(Map handlers, Map properties) {
        super(properties);
        this.handlers = handlers;
        state = STATE_AUTHORIZATION;
    }

    /*
     *  (non-Javadoc)
     * @see org.jboss.mail.Protocol#greet(java.io.OutputStream)
     */
    public void greet(OutputStream stream) throws IOException {
        PrintWriter writer = AbstractResponse.getWrappedWriter(new PrintWriter(stream));
        writer.println(
            "+OK "
                + getProperty(SMTPProtocolInstance.SERVER_NAME) //TODO: Refactor to common code.
                + " POP3 Server (JBMAIL POP3 Server version 0.1) "
                //+ getProperty(STATUS)
	    // THERE WAS NO 'LONG INTERNET DATE' PROPERTY....
                // + " "
//                 + getProperty("long internet date"))
	    );
        writer.flush();
    }

    /*
     *  (non-Javadoc)
     * @see org.jboss.mail.Protocol#parseRequest(java.io.InputStream, java.net.Socket)
     */
    public Request parseRequest(InputStream stream, Socket socket)
        throws IOException {

        MutablePOP3Request request = new MutablePOP3Request( this, stream );
        String commandString = readCommand(stream);
        if (log.isDebugEnabled()){
          log.debug("Got commandString: " + commandString);
        }

        String[] command = parseCommand(commandString);

        request.setCommand(command[0].trim());
        if( command[1] != null && command[1].trim().length() > 0 ){
            //KAB: POP3 commands can have more than one argument (e.g. TOP).
        	//Arguments are separated by a single space.
        	//request.addArgument(command[1].trim());
        	StringTokenizer tok = new StringTokenizer(command[1].trim(), " ");
			while (tok.hasMoreTokens()){
				String arg = tok.nextToken();
				request.addArgument(arg);
			}
        }
           
        request.setSocket(socket);
        return request;
    }

    /*
     *  (non-Javadoc)
     * @see org.jboss.mail.Protocol#handleRequest(java.io.OutputStream, org.jboss.mail.Request)
     */
    public Response handleRequest(OutputStream stream, Request request) {
        if (log.isDebugEnabled()) {
           log.debug(
            "HandleRequest called with request = "
                + (request == null ? "true" : "false"));
            log.debug(
            "HandleRequest called with command = "
                + (request == null
                    ? "null"
                    : ((POP3Request) request).getCommand()));
        }
        POP3Handler handler =
            (POP3Handler) handlers.get(((POP3Request) request).getCommand());
        POP3Response response = null;
        if (handler != null) {
            try {
                response =
                    handler.handleRequest(stream, (POP3Request) request, this);
            } catch (IOException e) {
                handleIOError(stream);
            }
        } else {
            try {
                response =
                    new POP3Response(request, stream, this);
                PrintWriter writer = response.getWriter();
                writer.println(
                    "-ERR "
                        + getProperty(POP3Constants.SERVER_NAME)
                        + " Syntax error, command unrecognized: "
                        + ((POP3Request) request).getCommand());
                writer.flush();
            } catch (IOException ioe) {
                response = (POP3Response) handleIOError(stream);
            }
        }
        return response;
    }

    /**
     * gets a response string for a request
     * @param request which is being processed
     * @return response string
     * @throws IOException in the event the response cannot be read due to some communication exception
     */
    public String readResponse(Request request) throws IOException {
        return readCommand( ((POP3Request) request).getInputStream());
    }

    /**
     * in the event an IO Error occurs (generally fatal to the request) this is called
     */
    public Response handleIOError(OutputStream stream) {
        log.error("Handle IO Error");
        return null;
    }

    public String getName() {
        return "POP3";
    }

    public void setProperties(List properties) {
        // TODO Auto-generated method stub

    }

    public Element getProperties() {
        // TODO Auto-generated method stub
        return null;
    }

    /**
     * The current state of the transaction.
     *
     * @return STATE_AUTHORIZATION, STATE_TRANSACTION, or STATE_UPDATE
     */
    public int getState() {
        return state;
    }

    /**
     * The current state of the transaction.
     *
     * @param state STATE_AUTHORIZATION, STATE_TRANSACTION, or STATE_UPDATE
     */
    public void setState(int state) {
        this.state = state;
    }

    /**
     * The username for the current session.
     *
     * @return null if not set.
     */
    public String getUsername() {
        return username;
    }

    /**
     * The username for the current session.
     *
     * @param username valid username, or null to clear.
     */
    public void setUsername(String username) {
        this.username = username;
    }

    /**
     * The user's ROOT folder.
     *
     * @return null if not authenticated.
     */
    public Folder getMailbox() {
        return mailbox;
    }

    /**
     * The user's ROOT folder.
     *
     * @param mailbox a locked Folder.
     */
    public void setMailbox(Folder mailbox) {
        this.mailbox = mailbox;
    }

	/* (non-Javadoc)
	 * @see org.jboss.mail.Protocol#greet(java.nio.channels.SocketChannel)
	 */
	public void greet(SocketChannel ioChannel) throws IOException {

				
		char[] chr = new String(         
			   "OK " 
		        + getProperty(SMTPProtocolInstance.SERVER_NAME) //TODO: Refactor to common code.
		        + " POP3 Server (JBMAIL POP3 Server version 0.1) "
		        //+ getProperty(STATUS)
//		 THERE WAS NO 'LONG INTERNET DATE' PROPERTY....
		        // + " "
//		         + getProperty("long internet date"))
				).toCharArray();
	        ByteBuffer buffer = ByteBuffer.allocate(chr.length * 2);
	        buffer.asCharBuffer().put(chr);
	        ioChannel.write(buffer); 
	}
}
