/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.pop3.handlers;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;

import org.jboss.mail.pop3.POP3Response;
import org.jboss.mail.pop3.POP3Request;
import org.jboss.mail.pop3.POP3ProtocolInstance;
import org.jboss.mail.mailbox.Folder;
import org.jboss.mail.mailbox.MailboxMessage;
import org.apache.log4j.Logger;

/**
 * LIST generates a list of all messages currently
 * stored in the user's mailbox.
 *
 * @author Eric Daugherty
 */
public class CmdLIST extends AbstractCommand implements POP3Handler, POP3UserMessages {

    public final static String COMMAND = "LIST";

    /** Log4j logger */
    private static final Logger log = Logger.getLogger( CmdLIST.class );

    /* (non-Javadoc)
     * @see org.jboss.mail.smtp.handlers.POP3Handler#handleRequest(java.io.OutputStream, org.jboss.mail.smtp.POP3Request, org.jboss.mail.pop3.POP3ProtocolInstance)
     */
    public POP3Response handleRequest(
        OutputStream out,
        POP3Request request,
        POP3ProtocolInstance protocol)
        throws IOException {

        log.debug("LIST command handler called");
        POP3Response response = new POP3Response(request, out, protocol);
        PrintWriter writer = response.getWriter();

        // Get the arguments.
        String[] arguments = request.getArguments();

        // Verify the protocol is in the right state.
        if( verifyState( protocol, writer, POP3ProtocolInstance.STATE_TRANSACTION ) )
        {
            Folder mailbox = protocol.getMailbox();

            // Output a listing of all messages
            if( arguments.length == 0 )
            {
                MailboxMessage[] messages = mailbox.getMessages();

                long totalSize = 0;
                long size = 0;
                StringBuffer messageOutput = new StringBuffer();

                // Calculate the individual message sizes and total size.
                for( int index = 0; index < messages.length; index++ )
                {
                    // Don't included messages marked for deletion.
                    if( !messages[index].isDeleted() )
                    {
                        size = messages[index].getSize();
                        totalSize += size;
                        messageOutput.append( (index + 1) + " " + size + "\r\n" );
                    }
                }

                // Output the response.
                writer.println( "+OK " + messages.length + " messages (" + totalSize + " octets)" );
                writer.print( messageOutput.toString() );
                writer.println( "." );
            }
            // Output info for one message
            else if( arguments.length == 1 )
            {
                MailboxMessage message = mailbox.getMessage( arguments[0] );
                if( message == null || message.isDeleted() )
                {
                    writer.println( MESSAGE_INVALID_MESSAGE_ID );
                }
                else
                {
                    writer.println(  "+OK " + arguments[0] + " " + message.getSize() );
                }

            }
            // Otherwise, too many arguments
            else
            {
                writer.println( MESSAGE_TOO_MANY_ARGUMENTS );
            }
        }

        writer.flush();
        return response;
    }
}
