/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.pop3.handlers;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;

import org.jboss.mail.pop3.POP3Response;
import org.jboss.mail.pop3.POP3Request;
import org.jboss.mail.pop3.POP3ProtocolInstance;
import org.jboss.mail.mailbox.Folder;
import org.jboss.mail.mailbox.MailboxMessage;
import org.apache.log4j.Logger;

/**
 * RETR is a request for the server to
 * send a specific message to the client.
 *
 * @author Eric Daugherty
 */
public class CmdRETR extends AbstractCommand implements POP3Handler, POP3UserMessages {

    public final static String COMMAND = "RETR";

    /** Log4j logger */
    private static final Logger log = Logger.getLogger( CmdRETR.class );

    /* (non-Javadoc)
     * @see org.jboss.mail.smtp.handlers.POP3Handler#handleRequest(java.io.OutputStream, org.jboss.mail.smtp.POP3Request, org.jboss.mail.pop3.POP3ProtocolInstance)
     */
    public POP3Response handleRequest(
        OutputStream out,
        POP3Request request,
        POP3ProtocolInstance protocol)
        throws IOException {

        log.debug("RETR command handler called");
        POP3Response response = new POP3Response(request, out, protocol);
        PrintWriter writer = response.getWriter();

        // Get the arguments.
        String[] arguments = request.getArguments();

        // Verify the protocol is in the right state.
        if( verifyState( protocol, writer, POP3ProtocolInstance.STATE_TRANSACTION ) &&
            verifyArgumentCount( arguments, writer, 1 ) )
        {
            Folder mailbox = protocol.getMailbox();

            // There should be one argument, which is the message number.
            if( arguments.length == 1 )
            {
                // Retrieve the message.
                MailboxMessage message = mailbox.getMessage( arguments[0] );
                if( message == null || message.isDeleted() )
                {
                    writer.println( MESSAGE_INVALID_MESSAGE_ID );
                }
                else
                {
                    // Write the message back to the client.
                    writer.println( MESSAGE_OK );
                    writeMessage( writer, message );
                }
            }
        }

        writer.flush();
        return response;
    }

    /**
     * Outputs the message headers and body to the writer.
     *
     * @param writer the PrintWriter the message is written to.
     * @param message the message to output.
     */
    private void writeMessage( PrintWriter writer, MailboxMessage message )
    {
        // Write the headers.
        String[] headers = message.getHeaders();
        for( int index = 0; index < headers.length; index++ )
        {
            writer.println( headers[index] );
        }

        // Write the empty line between headers and body.
        writer.println( "" );

        // Write the body.
        String[] body = message.getBody();
        for( int index = 0; index < body.length; index++ )
        {
            if (body[index].equals(".")) {
              body[index] = "..";
            } else if (body[index].equals("..")) {
              body[index] = "...";
            }
            writer.println( body[index] );
        }

        writer.println( "." );
    }
}
