/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.pop3.handlers;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;

import org.jboss.mail.pop3.POP3ProtocolInstance;
import org.jboss.mail.pop3.POP3Request;
import org.jboss.mail.pop3.POP3Response;
import org.jboss.mail.mailbox.Folder;
import org.jboss.mail.mailbox.MailboxMessage;
import org.apache.log4j.Logger;

/**
 * TOP Returns the first x lines of the
 * requested message.
 * <p>
 * This is an optional POP3 command.
 *
 * @author Eric Daugherty
 */
public class CmdTOP extends AbstractCommand implements POP3Handler, POP3UserMessages {

    public final static String COMMAND = "TOP";

    /** Log4j logger */
    private static final Logger log = Logger.getLogger( CmdTOP.class );

    /* (non-Javadoc)
     * @see org.jboss.mail.smtp.handlers.POP3Handler#handleRequest(java.io.OutputStream, org.jboss.mail.smtp.POP3Request, org.jboss.mail.pop3.POP3ProtocolInstance)
     */
    public POP3Response handleRequest(
        OutputStream out,
        POP3Request request,
        POP3ProtocolInstance protocol)
        throws IOException {

        log.debug( "TOP command handler called" );
        POP3Response response = new POP3Response(request, out, protocol);
        PrintWriter writer = response.getWriter();

        // Get the arguments.
        String[] arguments = request.getArguments();

        // Verify the protocol is in the right state.
        if( verifyState( protocol, writer, POP3ProtocolInstance.STATE_TRANSACTION ) &&
            verifyArgumentCount( arguments, writer, 2 ) )
        {
            Folder mailbox = protocol.getMailbox();

            try
            {
                int numLines = Integer.parseInt( arguments[1] );

                // Retrieve the message.
                MailboxMessage message = mailbox.getMessage( arguments[0] );
                if( message == null || message.isDeleted() )
                {
                    writer.println( MESSAGE_INVALID_MESSAGE_ID );
                }
                else
                {
                    // Write the message back to the client.
                    writer.println( MESSAGE_OK );
                    writeMessage( writer, message, numLines );
                }
            }
            catch( NumberFormatException numberFormatException )
            {
                writer.println( MESSAGE_ARGUMENT_INVALID );
            }
        }

        writer.flush();
        return response;
    }

    /**
     * Outputs the message headers and body to the writer.
     *
     * @param writer the PrintWriter the message is written to.
     * @param message the message to output.
     * @param numLines the number of lines to output.
     */
    private void writeMessage( PrintWriter writer, MailboxMessage message, int numLines )
    {
        String[] headers = message.getHeaders();
        for( int index = 0; index < headers.length; index++ )
        {
            writer.println( headers[index] );
        }

        // Write the empty line between headers and body.
        writer.println( "" );

        String[] body = message.getBody();
        // Write out the requested number of lines, or the entire message, which ever is shorter
        int lineCount = body.length > numLines ? numLines : body.length;
        for( int index = 0; index < lineCount; index++ )
        {
            writer.println( body[index] );
        }

        writer.println( "." );
    }
}
