/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.pop3.handlers;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;

import org.jboss.mail.pop3.POP3ProtocolInstance;
import org.jboss.mail.pop3.POP3Request;
import org.jboss.mail.pop3.POP3Response;
import org.jboss.mail.mailbox.Folder;
import org.jboss.mail.mailbox.MailboxMessage;
import org.apache.log4j.Logger;

/**
 * UIDL returns the unique id of the
 * requested message, or all messages
 * if no message is specified.
 * <p>
 * This is an optional POP3 command.
 *
 * @author Eric Daugherty
 */
public class CmdUIDL extends AbstractCommand implements POP3Handler, POP3UserMessages {

    public final static String COMMAND = "UIDL";

    /** Log4j logger */
    private static final Logger log = Logger.getLogger( CmdUIDL.class );

    /* (non-Javadoc)
     * @see org.jboss.mail.smtp.handlers.POP3Handler#handleRequest(java.io.OutputStream, org.jboss.mail.smtp.POP3Request, org.jboss.mail.pop3.POP3ProtocolInstance)
     */
    public POP3Response handleRequest(
        OutputStream out,
        POP3Request request,
        POP3ProtocolInstance protocol)
        throws IOException {

        log.debug( "UIDL command handler called" );
        POP3Response response = new POP3Response(request, out, protocol);
        PrintWriter writer = response.getWriter();

        // Verify the protocol is in the right state.
        if( verifyState( protocol, writer, POP3ProtocolInstance.STATE_TRANSACTION ) )
        {
            // Get the arguments.
            String[] arguments = request.getArguments();

            Folder mailbox = protocol.getMailbox();

            // Output all the message ids.
            if( arguments.length == 0 )
            {
                MailboxMessage[] messages = mailbox.getMessages();
                writer.println( MESSAGE_OK );
                for( int index = 0; index < messages.length; index++ )
                {
                    if( !messages[index].isDeleted() )
                    {
                        writer.println( (index + 1) + " " + messages[index].getId() );
                    }
                }
                writer.println( "." );

            }
            // Output a specific message id.
            else if( arguments.length == 1 )
            {

                // Retrieve the message.
                MailboxMessage message = mailbox.getMessage( arguments[0] );
                if( message == null || message.isDeleted() )
                {
                    writer.println( MESSAGE_INVALID_MESSAGE_ID );
                }
                else
                {
                    // Write the message id to the client.
                    writer.println( MESSAGE_OK + " " + arguments[0] + " " + message.getId() );
                }

            }
            // Otherwise, too many arguments
            else
            {
                writer.println( MESSAGE_TOO_MANY_ARGUMENTS );
            }

        }

        writer.flush();
        return response;
    }
}

