/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.smtp;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.management.ObjectName;
import javax.naming.InitialContext;
import javax.net.ssl.SSLSocketFactory;

import org.jboss.mail.MailListener;
import org.jboss.mail.ServiceProxy;
import org.jboss.mail.domaingroup.DomainGroupMBean;
import org.jboss.mail.smtp.handlers.SMTPHandlers;
import org.jboss.mail.util.JBMailJMXUtil;
import org.jboss.security.SecurityDomain;
import org.jboss.system.ServiceMBeanSupport;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * SMTPProtocol implements the SMTP Protocol for Server.  SMTP is the protocol used to receive
 * mails from org.jboss.org.jboss.mail clients and servers.
 *
 * @author Andrew C. Oliver
 * @copyright SuperLink Software, Inc. 2003 All rights reserved unless otherwise granted.
 */
public class SMTPProtocol extends ServiceMBeanSupport implements SMTPConstants, SMTPProtocolMBean {

    Map handlers;
    private static final String SENDER = "SENDER";
    Map properties;
    private ObjectName domainGroup;

	/** The name of the security domain to use with server sockets that support SSL
	 */
	protected String sslDomain;

	/**
     * constructs an SMTPProtocol with no configuration
     */
    public SMTPProtocol() {
        handlers = SMTPHandlers.instance();
    }

    public String getName() {
        return "SMTP";
    }

    // documented in super class

    //TODO: return properties as XML
    public Element getProperties() {
        return null;
    }

    public void setProperties(Element properties) {
    	log.debug("setProperties()");
        NodeList nodes = properties.getElementsByTagName("property");
        Map pproperties = new HashMap(nodes.getLength());
        for (int k = 0; k < nodes.getLength(); k++) {
            String pname = null;
            Object pval = null;
            Node node = nodes.item(k);
            NodeList subnodes = node.getChildNodes();
            for (int x = 0; x < subnodes.getLength(); x++) {
                Node subnode = subnodes.item(x);
                if (subnode.getNodeName().equals("name")) {
                    pname = subnode.getFirstChild().getNodeValue();
                } else if (subnode.getNodeName().equals("value")) {
                    pval = subnode.getFirstChild().getNodeValue();
                } else if (subnode.getNodeName().equals("values")) {
                    NodeList values = subnode.getChildNodes();
                    pval = new ArrayList(values.getLength());
                    List plist = (List) pval;
                    for (int y = 0; y < values.getLength(); y++) {
                        Node value = values.item(y);
                        if (value.getNodeName() != null
                            && value.getNodeName().equals("value")) {
                            plist.add(value.getFirstChild().getNodeValue());
                        }
                    }
                } /* else {
                                throw new RuntimeException("property and password pairs are all that are needed/permitted "+subnode.getNodeName());
                            }*/
            }
            if (pname == null || pval == null) {
                throw new RuntimeException("users need passwords and passwords need users...");
            }

            setProperty(pproperties, pname, pval);
        }
        if (domainGroup != null) {
            DomainGroupMBean bean = null;
            try {
                bean = (DomainGroupMBean) JBMailJMXUtil.getService(domainGroup);
            } catch (Exception e) {
                e.printStackTrace();
                throw new RuntimeException("could not set domain group due to JMX issues");
            }
            pproperties.put(SMTPConstants.DOMAIN_GROUP,bean);
        }
		
		//Create some default values
		if (pproperties.get(SMTPConstants.MAX_RECEIVED_HEADERS) == null){
			log.debug("Property '" + SMTPConstants.MAX_RECEIVED_HEADERS + "' not present. Defaulting to 100");
        	pproperties.put(SMTPConstants.MAX_RECEIVED_HEADERS, new Long(100));
		}
		if (pproperties.get(SMTPConstants.RECEIVED_HEADERS_THRESHOLD) == null){
			log.debug("Property '" + SMTPConstants.RECEIVED_HEADERS_THRESHOLD + "' not present. Defaulting to 20");
			pproperties.put(SMTPConstants.RECEIVED_HEADERS_THRESHOLD, new Long(20));
		}
		if (pproperties.get(SMTPConstants.MAX_OWN_RECEIVED_HEADERS) == null){
			log.debug("Property '" + SMTPConstants.MAX_OWN_RECEIVED_HEADERS + "' not present. Defaulting to 5");
			pproperties.put(SMTPConstants.MAX_OWN_RECEIVED_HEADERS, new Long(5));        	
		}
        this.properties = pproperties;
        
		log.debug("setProperties() - END");
    }

    /**
     * helper function to set properties into the map, typing them and all...
     * @param pproperties the map which we're building
     * @param pname the name of the property
     * @param pval the value to set
     */
    private void setProperty(Map pproperties, String pname, Object pval) {
        if (pname.equals(SMTPConstants.SERVER_NAME)) {
            String sval = (String) pval;
            pproperties.put(pname, sval);
		} else if (pname.equals(SMTPConstants.SERVER_NAME)) {
			String sval = (String) pval;
			pproperties.put(pname, sval);
		} else if (pname.equals(SMTPConstants.POSTMASTER)) {
			String sval = (String) pval;
			pproperties.put(pname, sval);
			System.setProperty(pname, sval);
        } else if (pname.equals(SMTPConstants.AUTH_REQUIRED)) {
            String sval = (String) pval;
            pproperties.put(pname, new Boolean(sval));
        } else if (pname.equals(SMTPConstants.AUTH_METHODS)) {
            String sval = (String) pval;
            pproperties.put(pname, sval);
        } else if (pname.equals(SMTPConstants.USER_REPOSITORY)) {
            try {
                ServiceProxy proxy =
                    (ServiceProxy) JBMailJMXUtil.getService((String) pval);
                pproperties.put(pname, proxy);
            } catch (Exception e) {
                e.printStackTrace();
                throw new RuntimeException(e);
            }
        } else if (pname.equals(SMTPConstants.VERIFY_IDENTITY)) {
            String sval = (String) pval;
            pproperties.put(pname, sval);
        } else if (pname.equals(SMTPConstants.MAX_MSG_SIZE)) {
            String sval = (String) pval;
            pproperties.put(pname, new Long(sval));
        } else if (pname.equals(SMTPConstants.MAIL_LISTENERS)) {
            List list = (List) pval;
            List plist = new ArrayList(list.size());
            Iterator i = list.iterator();
            while (i.hasNext()) {
                try {
                    MailListener listener =
                        (MailListener) JBMailJMXUtil.getService(
                            (String) i.next());
                    plist.add(listener);
                } catch (Exception e) {
                    e.printStackTrace();
                    throw new RuntimeException(e);
                }
            }
            pproperties.put(pname, plist);
        } else if (pname.equals(SMTPConstants.REQUIRE_CLIENT_CERT)) {
            String sval = (String) pval;
            pproperties.put(pname, new Boolean(sval));
        } else if (pname.equals(SMTPConstants.REQUIRE_STARTTLS)) {
            String sval = (String) pval;
            pproperties.put(pname, new Boolean(sval));
        } else if (pname.equals(SMTPConstants.BLOCK_SIZE)) {
            String sval = (String) pval;
            pproperties.put(pname, new Long(sval));
        } else if (pname.equals(SMTPConstants.TLS_SUPPORT)) {
            String sval = (String) pval;
            pproperties.put(pname, new Boolean(sval));
		} else if (pname.equals(SMTPConstants.MAX_RECEIVED_HEADERS)){
			String sval = (String) pval;
			pproperties.put(SMTPConstants.MAX_RECEIVED_HEADERS, new Long(sval));
		} else if (pname.equals(SMTPConstants.RECEIVED_HEADERS_THRESHOLD)){
			String sval = (String) pval;
			pproperties.put(SMTPConstants.RECEIVED_HEADERS_THRESHOLD, new Long(sval));
		} else if (pname.equals(SMTPConstants.MAX_OWN_RECEIVED_HEADERS)){
			String sval = (String) pval;
			pproperties.put(SMTPConstants.MAX_OWN_RECEIVED_HEADERS, new Long(sval));
		}
    }

    public ServiceProxy produceServiceProxy() {
	    return new SMTPProtocolInstance(handlers, properties);
    }


    public void setDomainGroup(ObjectName domainGroup) {
        this.domainGroup = domainGroup;
        DomainGroupMBean bean;
        if (properties != null) {
            try {
                bean = (DomainGroupMBean) JBMailJMXUtil.getService(domainGroup);
            } catch (Exception e) {
                e.printStackTrace();
                throw new RuntimeException("could not set domain group due to JMX issues");
            }
            properties.put(DOMAIN_GROUP,bean);
        }
    }

    public ObjectName getDomainGroup() {
        return domainGroup;
    }

	/** Set the JAAS security domain to be used with server sockets that support SSL
	 * @param domainName The JNDI name of the security domain used with server sockets that support SSL
	 */
	public void setSecurityDomain(String domainName) {
		try{
			log.debug("setSecurityDomain()");
			this.sslDomain = domainName;
		}
		catch(RuntimeException e){
			log.error("Error", e);
			throw e;
		}
	}

	/** Get the JAAS security domain to be used with server sockets that support SSL
	 * @return The JNDI name of the security domain used with server sockets that support SSL
	 */
	public String getSecurityDomain() {
		return sslDomain;
	}

    protected void startService() throws Exception {
    	log.debug("startService()");
		setupSSLSocketFactory();
   	}

	/** Set up the SSLSocketFactory to be used to create the wrapper SSL Socket when
	 * SSL/TLS is to be used
	 */
	private void setupSSLSocketFactory() throws RuntimeException{
		SSLSocketFactory sslSocketFactory = null;
		if (sslDomain != null) {
			ClassLoader loader = Thread.currentThread().getContextClassLoader();

			try {
				///////////////////////////////////////////////////////////
				// Here we want to use an SSLSocketFactory that understands
				// security domains. org.jboss.security.ssl.DomainSocket
				// does but doesn't exist in 3.2.3. Another implementation
				// has been provided for 3.2.3 hence this "clutter"!
				//
				// Once 3.2.4 has taken over the alternative implementation
				// should be scrapped.

				try{
					//This only exists in 3.2.4 and not in 3.2.3
					Class sfClass = loader.loadClass("org.jboss.security.ssl.DomainSocketFactory");
					sslSocketFactory = (SSLSocketFactory)sfClass.newInstance();
				}
				catch(Exception e){
					throw new RuntimeException("Could not create SSLSocketFactory", e);
				}

				Class sfClass = sslSocketFactory.getClass();
				InitialContext ctx = new InitialContext();
				SecurityDomain domain = (SecurityDomain) ctx.lookup(sslDomain);
				Class[] parameterTypes = { SecurityDomain.class };
				Method m = sfClass.getMethod("setSecurityDomain", parameterTypes);
				Object[] args = { domain };
				m.invoke(sslSocketFactory, args);

			}
			catch(NoSuchMethodException e){
				log.error("Socket factory does not support setSecurityDomain(SecurityDomain)");
				throw new RuntimeException("Could not set up security for TLS", e);
			}
			catch (Exception e) {
				log.error("Failed to setSecurityDomain=" + sslDomain + " on socket factory", e);
				throw new RuntimeException("Could not set up security for TLS", e);
			}
		}
		else{
			//If no security domain has been created, it makes no sense to
			//have enabled tls and/or set tls to required
			Boolean prop = (Boolean)properties.get(SMTPConstants.TLS_SUPPORT);
			if (prop != null && prop.booleanValue()){
				throw new RuntimeException("tls-enabled was set to true, but no SecurityDomain was specified");
			}

			prop = (Boolean)properties.get(SMTPConstants.REQUIRE_STARTTLS);
			if (prop != null && prop.booleanValue()){
				throw new RuntimeException("requireSTARTTLS was set to true, but no SecurityDomain was specified");
			}
		}

		properties.put(SMTPConstants.SSL_SOCKET_FACTORY, sslSocketFactory);
	}

}
