/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.smtp;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.net.InetAddress;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.channels.SocketChannel;

import org.jboss.mail.ServerThread;
import org.jboss.logging.Logger;
import org.jboss.mail.AbstractProtocol;
import org.jboss.mail.AbstractResponse;
import org.jboss.mail.Protocol;
import org.jboss.mail.Request;
import org.jboss.mail.Response;
import org.jboss.mail.ServiceProxy;
import org.jboss.mail.domaingroup.DomainGroupMBean;
import org.jboss.mail.message.*;
import org.jboss.mail.smtp.handlers.SMTPHandler;
import org.jboss.mail.util.DateUtil;
import org.w3c.dom.Element;

/**
 * Service Proxy created by the MBean which is used by the Server (actually ServerThread) to implement
 * the protocol.
 *
 * @author Andrew C. Oliver
 */
public class SMTPProtocolInstance
    extends AbstractProtocol
    implements Protocol, SMTPConstants, ServiceProxy {
    private static final Logger log = Logger.getLogger(SMTPProtocolInstance.class);
    /**
     * returned from SMTPHandlers.instance();
     */
    private Map handlers;

    private MailHeaders mailHeaders = new MailHeadersImpl();

    /**
     * returns the *only* instance of MailHeaders.  I'm not sure why we're doing this like this.
     * @todo why??
     * @return ...
     */
    public MailHeaders getMailHeaders(){
        return mailHeaders;
    }
    
    /**
     * create an SMTPProtocolInstance with the propert handlers and the already constructed properties map
     * @param handlers from SMTPHandlers.instance();
     * @param properties from SMTPProtocol.setProperties()
     */
    public SMTPProtocolInstance(Map handlers, Map properties) {
        super(properties);
        this.handlers = handlers;
    }

    public void greet(OutputStream stream) throws IOException {
        PrintWriter writer = AbstractResponse.getWrappedWriter(new PrintWriter(stream));
        writer.println(
            "220 "
                + getProperty(SERVER_NAME)
                + " SMTP Server (JBMAIL SMTP Server version 0.1) "
                + "ready"
                + " "
                + DateUtil.longInternetDate());
        writer.flush();
    }

    public Request parseRequest(InputStream stream, Socket socket)
        throws IOException {
        MutableSMTPRequest request = new MutableSMTPRequest(this, stream);
        String commstring = readCommand(stream);
        String[] command = parseCommand(commstring);

        request.setCommand(command[0].trim());
        request.addArgument(command[1]);
        request.setSocket(socket);
        
        if (log.isTraceEnabled() ) {
            SMTPProtocolInstance.log.trace(commstring);
        }
        return request;
    }


    /**
     * gets a response string for a request
     * @param request which is being processed
     * @return response string
     * @throws IOException in the event the response cannot be read due to some communication exception
     */
    public String readResponse(SMTPRequest request) throws IOException {
        return readCommand(request.getInputStream());
    }

    public Response handleRequest(OutputStream stream, Request request) {
        if (log.isDebugEnabled()) {
            log.debug(
            "HandleRequest called with request = "
                + (request == null ? "true" : "false"));
            log.debug(
            "HandleRequest called with command = "
                + (request == null
                    ? "null"
                    : ((SMTPRequest) request).getCommand()));
        }
        SMTPHandler handler =
            (SMTPHandler) handlers.get(((SMTPRequest) request).getCommand());
        SMTPResponse response = null;
        if (handler != null) {
            try {
                response =
                    handler.handleRequest(stream, (SMTPRequest) request, this);
                } catch (IOException e) {
                e.printStackTrace();
                handleIOError(stream);
            }
        } else {
            try {
                response =
                    new SMTPResponse((SMTPRequest) request, stream, this, secure);
                PrintWriter writer = response.getWriter();
                writer.println(
                    "500 "
                        + getProperty("servername")
                        + " Syntax error, command unrecognized: "
                        + ((SMTPRequest) request).getCommand());
                writer.flush();
            } catch (IOException ioe) {
                response = (SMTPResponse) handleIOError(stream);
            }
        }
        return response;
    }


//TODO: this smells bad..
    /**
     * this is called back to from various handlers.  It handles the actual authentication with a positive
     * or negative response
     * @param rcptAddr recepient of the org.jboss.org.jboss.mail
     * @param out output stream
     * @param request SMTPRequest which is being processed
     * @param protocol should be equal to this instance (or we'll puke)
     * @param writer printwriter for writing the response, if negative
     * @return boolean whether the user is authenticated or not
     */
    public boolean handleAuth(
        MailAddress rcptAddr,
        OutputStream out,
        SMTPRequest request,
        SMTPProtocolInstance protocol,
        PrintWriter writer) {
        DomainGroupMBean dg = (DomainGroupMBean) protocol.getProperty(SMTPConstants.DOMAIN_GROUP);
        log.debug("HANDLE AUTH");
        if (protocol != this) {
            throw new RuntimeException("THIS SHOULD NOT HAPPEN:::WRONG PROTOCOL INSTANCE TO HANDLE AUTH...I THINK I'LL CHOKE NOW...BYE");
        }
        if (!getPropertyBool(AUTH_REQUIRED)) {
            log.debug("AUTH NOT REQUIRED");
            return true;
        }

	// mail to any domain in our local group should be accepted --- that's the free
	// and open world of smtp, where anybody can email anybody. but mail submitted by an
	// unauthenticated user to an unknown domain should get denied.
	String toDomain = rcptAddr.getDomain();
	if (dg.isInGroup(toDomain)) {
	    return true; // normal open and free smtp world
	} else if (getState("USER") == null) {
	    return false; // no access for unauthenticated users to unknown domains
	}
	

        if (log.isDebugEnabled()) {
            log.debug("USER ==" + getState("USER"));
        }

	InetAddress address = (InetAddress) getState(ServerThread.STATE_CLIENT_ADDRESS);
	boolean local = address.isLinkLocalAddress() || address.isLoopbackAddress();

        if (getState("USER") == null) {

            if (dg.isInGroup(toDomain)) {
                if (log.isDebugEnabled()) {
                    log.debug("SERVERNAME != domain");                    
                    log.debug(
                    "SERVERNAME = "
                        + getProperty(SMTPConstants.SERVER_NAME)
                        + " domain = "
                        + rcptAddr.getDomain());
                }
                writer.println("530 Authentication Required");
                writer.flush();
                return false;
            }
        }

        if ( authShouldBeRequired((String)getState("USER"), rcptAddr.getDomain()) &&
            getProperty(VERIFY_IDENTITY) != null
            && getProperty(VERIFY_IDENTITY).equals("true")) {
            String authUser =
                ((String) getState("USER"));
            authUser = authUser != null ? authUser.toLowerCase(Locale.US) : null;
            MailAddress senderAddress = (MailAddress) getState(SENDER);
            boolean domainExists = false;
//TODO: also check other "local" domains
            if ((authUser != null && !authUser.equals(senderAddress.getUser()))
                || (!dg.isInGroup(senderAddress.getDomain()))) {
                writer.println(
                    "503 Incorrect Authentication for Specified Email Address");
                writer.flush();
                return false;
            }
        }

        return true;
    }

    /**
     * dumb method to simplify the nasty if statement (will be removed when we decrud authentication)
     * @param user that is logged in (or null)
     * @param domain of the to address
     * @return whether auth should have been required for the purpose of verifying identity
     */
    private boolean authShouldBeRequired(String user, String domain) {
	DomainGroupMBean dg = (DomainGroupMBean) getProperty(SMTPConstants.DOMAIN_GROUP);
	// MIKEA: should actually allow unauthenticated email to any address in our domain group, not just server name
	if (dg.isInGroup(domain)) {
	    return false;
	}
	if (user == null) {
	    return false;
	}
	return true;
    }

	/**
     * in the event an IO Error occurs (generally fatal to the request) this is called
     */
    public Response handleIOError(OutputStream stream) {
        log.error("Handle IO Error");
        return null;
    }

    public String getName() {
        return "SMTP";
    }

    public void setProperties(List properties) {
        // TODO Auto-generated method stub

    }

    public Element getProperties() {
        // TODO Auto-generated method stub
        return null;
    }

	/* (non-Javadoc)
	 * @see org.jboss.mail.Protocol#greet(java.nio.channels.SocketChannel)
	 */
	public void greet(SocketChannel ioChannel) throws IOException {
		char[] chr = new String(            "220 "
            + getProperty(SERVER_NAME)
            + " SMTP Server (JBMAIL SMTP Server version 0.1) "
            + "ready"
            + " "
            + DateUtil.longInternetDate()).toCharArray();
        ByteBuffer buffer = ByteBuffer.allocate(chr.length * 2);
        buffer.asCharBuffer().put(chr);
        ioChannel.write(buffer);        
	}
}