/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.smtp.handlers;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Iterator;
import java.util.List;

import org.jboss.logging.Logger;
import org.jboss.mail.CharTerminatedInputStream;
import org.jboss.mail.DotStuffingInputStream;
import org.jboss.mail.MailListener;
import org.jboss.mail.Protocol;
import org.jboss.mail.SizeLimitedInputStream;
import org.jboss.mail.message.LoopDetectedException;
import org.jboss.mail.message.Mail;
import org.jboss.mail.message.MailAddress;
import org.jboss.mail.message.Message;
import org.jboss.mail.message.StandardMailHeaders;
import org.jboss.mail.smtp.SMTPConstants;
import org.jboss.mail.smtp.SMTPProtocolInstance;
import org.jboss.mail.smtp.SMTPRequest;
import org.jboss.mail.smtp.SMTPResponse;

/**
 * DATA tells the server to expect the actual message.  It marks the begin of a message payload.
 *  
 * @author Andrew C. Oliver
 */
public class CmdDATA implements SMTPHandler, SMTPConstants {
    private static final Logger jblog = Logger.getLogger(CmdDATA.class);

    public final static String COMMAND = "DATA";

    /*
     *  (non-Javadoc)
     * @see org.jboss.mail.smtp.handlers.SMTPHandler#handleRequest(java.io.OutputStream, org.jboss.mail.smtp.SMTPRequest, org.jboss.mail.Protocol)
     */
    public SMTPResponse handleRequest(
        OutputStream out,
        SMTPRequest request,
        Protocol protocol)
        throws IOException {
        jblog.debug("DATA command handler called");
        SMTPResponse response = new SMTPResponse(request, out, protocol);
        SMTPProtocolInstance smtpProtocol = (SMTPProtocolInstance) protocol;
        PrintWriter writer = response.getWriter();
        String arg = getArg(request.arguments());
        if (arg != null && !arg.equals("")) {
            writer.println(
                "500 Unexpected argument provided with DATA command");
        }

        if (smtpProtocol.getState(SENDER) == null) {
            writer.println("503 No sender specified");
        } else if (protocol.getState(RCPT_LIST) == null) {
            writer.println("503 No recipients specified");
        } else {
            writer.println("354 Ok Send data ending with <CRLF>.<CRLF>");
            writer.flush();
            InputStream in = request.getInputStream();
            InputStream msgIn =
                new CharTerminatedInputStream(in, SMTPTerminator);
            msgIn =
                new SizeLimitedInputStream(
                    msgIn,
                    smtpProtocol.getPropertyLong(MAX_MSG_SIZE));
            msgIn = new DotStuffingInputStream(msgIn);

            //prepare sender and receiver        
            MailAddress sender = (MailAddress) smtpProtocol.getState(SENDER);
            List rcptsList = (List) smtpProtocol.getState(RCPT_LIST);
            MailAddress[] rcpts = new MailAddress[rcptsList.size()];
            rcpts = (MailAddress[]) rcptsList.toArray(rcpts);

            try {
                int blocksize = (int)smtpProtocol.getPropertyLong(BLOCK_SIZE);
                // if you don't specify a blocksize you get 1k
                blocksize = blocksize == 0? 1024: blocksize;
                Mail mail = Mail.create(msgIn, blocksize, sender, rcpts, smtpProtocol);
                List lstListeners = (List) smtpProtocol.getProperty(MAIL_LISTENERS);
                Iterator mailListeners = lstListeners.iterator();
                
                String[] receivedHeaders = mail.getMailHeaders().getHeader(StandardMailHeaders.RECEIVED);
                boolean success = true; // assume success for now, even though listeners don't confirm
                
                Message msg = mail;
                while (mailListeners.hasNext() && msg != null) { 
                    MailListener lsner = (MailListener) mailListeners.next();
                    jblog.info("sending mail to MailListener: " + lsner);
                    msg = lsner.send(msg);
                }
                
                if (success) {
                    writer.println("250 Message received");
                } else {
                    writer.println("559 Message not committed due to internal error"); // '9' is non-standard but OK
                }
                in.mark(2);
            } catch (LoopDetectedException e) {
                //The exception contains the reason
                writer.println("550 " + e.getMessage());
            }
        }
       
        protocol.resetState();

        writer.flush();
        return response;
    }

    /**
     * gets the actual argument...SMTP has single line arguments this we have to parse them out of one argument
     * @param iter list of arguments (one element)
     * @return String represneting actual argument...
     */ 
    private String getArg(Iterator iter) {
        String remote = null;
        if (iter.hasNext() && iter != null) {
            remote = (String) iter.next();
        }
        return remote;
    }
}
