/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.smtp.handlers;

import org.jboss.logging.Logger;
import org.jboss.mail.Protocol;

import org.jboss.mail.message.MailAddress;
import org.jboss.mail.smtp.SMTPConstants;
import org.jboss.mail.smtp.SMTPRequest;
import org.jboss.mail.smtp.SMTPResponse;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;

import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.HashMap;
import java.util.StringTokenizer;

/**
 * MAIL signifies that we have a new message to be sent.  It attaches the FROM to the 
 * message.
 * 
 * @author Andrew C. Oliver
 * @copyright SuperLink Software, Inc. 2003 All rights reserved unless otherwise granted.
 */
public class CmdMAIL implements SMTPHandler, SMTPConstants{
    private static final Logger jblog = Logger.getLogger(CmdMAIL.class);

  public final static String COMMAND = "MAIL";

  /*
   *  (non-Javadoc)
   * @see org.jboss.mail.smtp.handlers.SMTPHandler#handleRequest(java.io.OutputStream, org.jboss.mail.smtp.SMTPRequest, org.jboss.mail.Protocol)
   */
  public SMTPResponse handleRequest(OutputStream out, SMTPRequest request, Protocol protocol) throws IOException {
    jblog.debug("MAIL command handler called");
    Map mailOptions = null;
    SMTPResponse response = new SMTPResponse(request, out, protocol);
    PrintWriter writer = response.getWriter();
    String arg = getArgline(request.arguments());
    String sender = null;
    if (jblog.isDebugEnabled()) { 
        jblog.debug("arg = "+arg);
    }
    if ((arg != null) && (arg.indexOf(":") > 0)) {
         jblog.debug("arg wasn't null");
         int colonIndex = arg.indexOf(":");
         jblog.debug("colonIndex="+colonIndex);
         sender = arg.substring(colonIndex + 1);
         jblog.debug("sender="+sender);
         arg = arg.substring(0, colonIndex);
         jblog.debug("arg="+arg);
    }
    if (sender != null && arg != null && jblog.isDebugEnabled()) {
         jblog.debug("Sender= " + sender.trim() + " **arg = " +arg.trim());
    }
    if (protocol.getState(SENDER) != null) {
      writer.println("503 Sender already specified");
    } else if (arg == null || !arg.toUpperCase(Locale.US).equals("FROM") ||
               sender == null) {
      writer.println("501 Usage: MAIL FROM:<sender>");
    } else {
      sender = sender.trim();
      int lastChar = sender.indexOf('>', sender.indexOf('<'));
      if ((lastChar > 0) && (sender.length() > lastChar + 2) && (sender.charAt(lastChar +1) == ' ')) {
         String mailOptionString = sender.substring(lastChar + 2);
         sender = sender.substring(0, lastChar + 1); 
         mailOptions = parseMailOptions(mailOptionString);
      }

    } 
    if (sender == null) {
       writer.println("501 Syntax error in MAIL command"); 
    } else if (!sender.startsWith("<") || !sender.endsWith(">")) {   
       writer.println("501 Syntax error in MAIL command"); 
    } else {
    	boolean isfrom = true;
        MailAddress address = MailAddress.parseSMTPStyle(sender, isfrom);
        if (!address.isValid()) {
           writer.println("501 Syntax error in sender address");
        }  else {
           protocol.setState(SENDER, address);
           writer.println("250 Sender "+sender+" OK");
        }
    }
      
    writer.flush();
    return response;
  }

  /**
   * The message size tells the server the size of message to expect.  Here we need to check 
   * if its withing bounds
   * @param value size of the message (string)
   * @param writer PrintWriter to write the response
   * @param protocol instance containing the configuration and state
   * @return boolean signifying if the size was within bounds.
   */
  private boolean handleMailSize(String value, PrintWriter writer, Protocol protocol) {
        int size = 0;
        try {
            size = Integer.parseInt(value);
        } catch (NumberFormatException pe) {
            // This is a malformed option value.  We return an error     
            writer.println( "501 Syntactically incorrect value for SIZE parameter" );
            return false;
        }
        long maxMessageSize = ((Long)protocol.getProperty("maxMessageSize")).longValue();    
        if ((maxMessageSize > 0) && (size > maxMessageSize)) {
            // Let the client know that the size limit has been hit.
            writer.println("552 Message size exceeds fixed maximum message size");
            return false;
        } else {
            // put the message size in the message state so it can be used
            // later to restrict messages for user quotas, etc.
            protocol.setState(SMTPConstants.MSG_SIZE, new Integer(size));
         }
     return true;

  }

  private Map parseMailOptions(final String mailOptionString) {
     Map options = new HashMap();
     StringTokenizer optionTokenizer = new StringTokenizer(mailOptionString, " ");
     while (optionTokenizer.hasMoreElements()) {
        String mailOption = optionTokenizer.nextToken();
        int equalIndex = mailOptionString.indexOf('=');
        String mailOptionName = mailOption;
        String mailOptionValue = "";
        if (equalIndex > 0) {
            mailOptionName = mailOption.substring(0, equalIndex).toUpperCase(Locale.US);
            mailOptionValue = mailOption.substring(equalIndex + 1);
         }
         options.put(mailOptionName, mailOptionValue);
     }
     return options;
  }

  private String getArgline(Iterator iter) {
   String remote = null;
   if (iter.hasNext() && iter != null) {
     remote = (String)iter.next();
    }
   return remote;
  }
}
