/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.smtp.handlers;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;

import org.jboss.mail.Protocol;
import org.jboss.mail.message.MailAddress;
import org.jboss.mail.smtp.SMTPConstants;
import org.jboss.mail.smtp.SMTPProtocolInstance;
import org.jboss.mail.smtp.SMTPRequest;
import org.jboss.mail.smtp.SMTPResponse;
import org.jboss.logging.Logger;

/**
 * RCPT is for listing the users who will receive the message
 * 
 * @author Andrew C. Oliver
 */
public class CmdRCPT implements SMTPHandler, SMTPConstants {

    private static final Logger jblog = Logger.getLogger(CmdRCPT.class);
    public final static String COMMAND = "RCPT";
    public final static String SENDER = "SENDER";
    public final static String MESG_SIZE = "MESG_SIZE";

    /*
     *  (non-Javadoc)
     * @see org.jboss.mail.smtp.handlers.SMTPHandler#handleRequest(java.io.OutputStream, org.jboss.mail.smtp.SMTPRequest, org.jboss.mail.Protocol)
     */
    public SMTPResponse handleRequest(
        OutputStream out,
        SMTPRequest request,
        Protocol protocol)
        throws IOException {
        jblog.debug("RCPT command handler called");
        SMTPResponse response = new SMTPResponse(request, out, protocol);
        PrintWriter writer = response.getWriter();
        String arguments[] = parseArguments(getArgline(request.arguments()));
        String recipient = arguments[0];
        String argument = arguments[1];

        if (protocol.getState(SENDER) == null) {
            writer.println("503 Need MAIL before RCPT");
        } else if (
            recipient == null
                || !argument.toUpperCase(Locale.US).equals("TO")
                || recipient.equals("")) {
            writer.println("501 Usage: RCPT TO:<recipient>");
        } else {
	    
	    List rcpts = (List) protocol.getState(RCPT_LIST);
	    if (rcpts == null) {
		rcpts = new ArrayList();
	    }
	    Map options = parseOptions(recipient);
	    MailAddress rcpt = parseRecipient(recipient.trim());
	    
	    if (rcpt.isValid()) {
		if (((SMTPProtocolInstance) protocol)
		    .handleAuth(
				rcpt,
				out,
				(SMTPRequest) request,
				(SMTPProtocolInstance) protocol,
				response.getWriter())) {
		    rcpts.add(rcpt);
		    protocol.setState(RCPT_LIST, rcpts);
		    writer.println("250 Recipient " + recipient.trim() + " OK");
		} else {
		    // don't we need to write something here?
		    writer.println("550 Not Authorized");
		}
	    } else {
		writer.println("501 Syntax error in recipient address");
	    }
	    
        }
	
        writer.flush();
        return response;
    }

    /**
     * parse the receipient and return a MailAddress object
     * @param String representation of recipient for this mail
     * @return MailAddress representation
     */
    private MailAddress parseRecipient(String recipient) {
        int lastChar = recipient.lastIndexOf('>');
        MailAddress retval =
            MailAddress.parseSMTPStyle(recipient.substring(0, lastChar + 1));
        return retval;
    }

    /**
     * parse the various options into a map.  Theoretically there can be options associated with the recipient
     * @param recipient as a String (raw argument to CmdRCPT)
     * @return Map representing any options
     */
    private Map parseOptions(String recipient) {
        Map options = new HashMap();
        int lastChar = recipient.lastIndexOf('>');
        if ((lastChar > 0)
            && (recipient.length() > lastChar + 2)
            && (recipient.charAt(lastChar + 1) == ' ')) {
            String rcptOptionString = recipient.substring(lastChar + 2);

            StringTokenizer optionTokenizer =
                new StringTokenizer(rcptOptionString, " ");

            while (optionTokenizer.hasMoreElements()) {
                String rcptOption = optionTokenizer.nextToken();
                int equalIndex = rcptOptionString.indexOf('=');
                String rcptOptionName = rcptOption;
                String rcptOptionValue = "";

                if (equalIndex > 0) {
                    rcptOptionName =
                        rcptOption.substring(0, equalIndex).toUpperCase(
                            Locale.US);
                    rcptOptionValue = rcptOption.substring(equalIndex + 1);
                }
                options.put(rcptOptionName, rcptOptionValue);
            }
        }
        return options;
    }

    private String[] parseArguments(String argline) {
        String args[] = new String[2];
        int colIndex = argline.indexOf(":");
        args[0] = argline.substring(colIndex + 1);
        args[1] = argline.substring(0, colIndex);
        return args;
    }

    private String getArgline(Iterator iter) {
        String remote = null;
        if (iter.hasNext() && iter != null) {
            remote = (String) iter.next();
        }
        return remote;
    }
}
