package org.jboss.mail.smtp.sender;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.mail.Address;
import javax.mail.MessagingException;
import javax.mail.SendFailedException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.URLName;
import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;

import org.jboss.mail.ServiceProxy;
import org.jboss.mail.message.EnvelopedAddress;
import org.jboss.mail.message.LoopDetectedException;
import org.jboss.mail.message.Mail;
import org.jboss.mail.message.MailAddress;
import org.jboss.mail.message.MailHeaders;
import org.jboss.mail.message.StandardMailHeaders;
import org.jboss.mail.smtp.SMTPProtocolInstance;
import org.jboss.system.ServiceMBeanSupport;
import org.w3c.dom.Element;
import org.xbill.DNS.Cache;
import org.xbill.DNS.Credibility;
import org.xbill.DNS.DClass;
import org.xbill.DNS.ExtendedResolver;
import org.xbill.DNS.MXRecord;
import org.xbill.DNS.Message;
import org.xbill.DNS.Name;
import org.xbill.DNS.RRset;
import org.xbill.DNS.Rcode;
import org.xbill.DNS.Record;
import org.xbill.DNS.Resolver;
import org.xbill.DNS.SetResponse;
import org.xbill.DNS.SimpleResolver;
import org.xbill.DNS.TextParseException;
import org.xbill.DNS.Type;

import com.sun.mail.smtp.SMTPMessage;

/**
 * @see org.jboss.mail.smtp.sender.SMTPSenderMBean
 * @author <a href="mailto:julien@jboss.org">Julien Viet</a>
 * @author Andrew C. Oliver <acoliver at jboss.org> (refactored from Nukes code)
 */
public class SMTPSender
	extends ServiceMBeanSupport
	implements SMTPSenderMBean, ServiceProxy {
	private Cache cache;
	private ExtendedResolver resolver;
	private boolean useCache;
	Map resolvers;
	private static final MXRecord[] EMPTY_MXRECORD = new MXRecord[0];

	private static final Comparator mxComparator = new Comparator() {
		public int compare(Object a, Object b) {
			return ((MXRecord) a).getPriority() - ((MXRecord) b).getPriority();
		}
	};

	public SMTPSender() {
		resolvers = new HashMap();
	}

	protected void startService() throws Exception {
		this.cache = new Cache();
		if (resolvers.values().size() > 0) {
			Resolver[] r =
				(Resolver[]) resolvers.values().toArray(
					new Resolver[resolvers.values().size()]);
			this.resolver = new ExtendedResolver(r);
		} else {
			this.resolver = new ExtendedResolver();
		}

	}

	public void addDNSServer(String dns) throws UnknownHostException {
		SimpleResolver sr = new SimpleResolver(dns);
		resolver.addResolver(sr);
		resolvers.put(dns, sr);
	}

	/* (non-Javadoc)
	 * @see org.jboss.mail.smtp.sender.SMTPSenderMBean#clearDNSCache()
	 */
	public void clearDNSCache() {
		cache = new Cache();
	}

	/* (non-Javadoc)
	 * @see org.jboss.mail.smtp.sender.SMTPSenderMBean#getCacheEnabled()
	 */
	public boolean getCacheEnabled() {
		return useCache;
	}

	/* (non-Javadoc)
	 * @see org.jboss.mail.smtp.sender.SMTPSenderMBean#getDNSServers()
	 */
	public Element getDNSServers() {
		//TODO output in XML
		return null;
	}

	/* (non-Javadoc)
	 * @see org.jboss.mail.smtp.sender.SMTPSenderMBean#mxLookup(java.lang.String)
	 */
	public String[] mxLookup(String address) {
		MXRecord[] records = performMXLookup(address);
		String[] results = new String[records.length];
		for (int i = 0; i < records.length; i++) {
			results[i] = records[i].getTarget().toString();
		}
		return results;
	}

	/**
	 * @param address
	 * @return
	 */
	private MXRecord[] performMXLookup(String address) {
		// for simplicity the code for cache and no cache is just whether we set to
		//this.cache or whether we make our own local instance.
		Cache cache = null;
		if (useCache == true) {
			cache = this.cache;
		} else {
			cache = new Cache();
		}

		// Do we have a mounted resolver ?
		if (resolver == null) {
			throw new RuntimeException("No name servers are mounted!  Mount one will you?");
		}

		Name name; // Name _name = Name.fromString(name); does not work, why ?
		try {
			name = new Name(address + ".");
			//trailing dots make it "absolute" so that it is happy
		} catch (TextParseException e) {
			throw new RuntimeException("Invalid lookup ", e);
		}

		// First lookup the cache
		SetResponse cached =
			cache.lookupRecords(name, Type.MX, Credibility.NONAUTH_ANSWER);

		if (!cached.isSuccessful()) {
			// Cache failed, let's query the server
			Record question = Record.newRecord(name, Type.MX, DClass.IN);
			Message query = Message.newQuery(question);
			Message response;
			try {
				response = resolver.send(query);
			} catch (IOException ioe) {
				throw new RuntimeException("Could not resolve DNS lookup", ioe);
			}
			short rcode = response.getHeader().getRcode();
			if (rcode == Rcode.NOERROR || rcode == Rcode.NXDOMAIN) {
				cache.addMessage(response);
			}
			cached =
				cache.lookupRecords(name, Type.MX, Credibility.NONAUTH_ANSWER);
		}

		// The server was not able to give us a good answer
		if (!cached.isSuccessful()) {
			return EMPTY_MXRECORD;
		}

		// Extract the MX records from the answer
		RRset[] rrsets = cached.answers();
		int answerCount = 0;
		int index = 0;
		for (int i = 0; i < rrsets.length; i++) {
			answerCount += rrsets[i].size();
		}
		MXRecord[] answers = new MXRecord[answerCount];
		for (int i = 0; i < rrsets.length; i++) {
			for (Iterator iterator = rrsets[i].rrs(); iterator.hasNext();) {
				answers[index++] = (MXRecord) iterator.next();
			}
		}

		// Finally sort the records
		Arrays.sort(answers, mxComparator);
		return answers;
	}

	/* (non-Javadoc)
	 * @see org.jboss.mail.smtp.sender.SMTPSenderMBean#removeDNSServer(java.lang.String)
	 */
	public void removeDNSServer(String dns) {
		Resolver r = (Resolver) resolvers.get(dns);
		resolver.deleteResolver(r);
		resolvers.remove(dns);
	}

	/* (non-Javadoc)
	 * @see org.jboss.mail.smtp.sender.SMTPSenderMBean#send(org.jboss.mail.smtp.Mail, boolean)
	 */
	public SMTPResult[] send(Mail message, MailAddress[] exclude) {
		log.debug("sender - exclude:" + exclude);
		MailAddress[] addresses = message.getTo();
		ArrayList results = new ArrayList(message.getTo().length);

		//Get the addresses by domain, and send for each domain
		MailAddressesByDomain addrByDomain =
			new MailAddressesByDomain(addresses, exclude);

		for (Iterator it = addrByDomain.getDomains(); it.hasNext();) {
			String domain = (String) it.next();
			MailAddress[] addressesForDomain = addrByDomain.getAddresses(domain);
			ArrayList resultsForDomain = sendForDomain(addressesForDomain, message);
			results.addAll(resultsForDomain);
		}

		return (SMTPResult[]) results.toArray(new SMTPResult[results.size()]);
	}
	

	/* (non-Javadoc)
	 * @see org.jboss.mail.smtp.sender.SMTPSenderMBean#send(org.jboss.mail.smtp.MailAddress, org.jboss.mail.smtp.Mail)
	 */
	public SMTPResult send(MailAddress to, Mail mail) {
		return (SMTPResult) sendForDomain(new MailAddress[] { to }, mail).get(
			0);
	}

	/** Sends a mail to several recipients
	 * @param to list of recipients who must all belong to the same domain
	 * @param mail mail to be sent
	 * @return ArrayList of SMTPResult entries. One for each recipient
	 */
	private ArrayList sendForDomain(MailAddress[] to, Mail mail) {
		boolean debug = log.isDebugEnabled();

		String[] hosts = mxLookup(to[0].getDomain());
		String host = null;

		if (hosts.length > 0) {
			host = hosts[0];
		} else {
			if (debug) {
				log.debug(
					"Could not resolve MX lookup for "
						+ to[0].getDomain()
						+ ". Remote mail not sent");
			}
			return SMTPResultImpl.getResolveDomainErrorForAll(to);
		}

		//Quite a lot of different errors are thrown by the javax.mail methods called.
		//My understanding (Kab, shoot me if I'm wrong!) that SendFailedException (thrown 
		//in a chained MessagingException by Transport.sendMessage()) means that sending 
		//got through to all recipients, while all other exceptions cause the mail to be 
		//aborted for all recipients.
		Transport transport = null;
		SMTPMessage message = null;
		try {
			//TODO doesn't handle i18n l10n properly..assumes US-ASCII for everything...thats wrong.
			MailAddress from = mail.getFrom();

			InternetAddress[] internetTos = new InternetAddress[to.length];	
			for (int i = 0 ; i < to.length ; i++){
				internetTos[i] = new InternetAddress(to[i].toString());
			}
		


			Properties props = new Properties();
			props.put("mail.debug", "true");

			Session session = Session.getDefaultInstance(props);
			URLName urlname = new URLName("smtp://" + host);
			transport = session.getTransport(urlname);
			transport.connect();

			//Down here will fail for everybody

			//SMTPMessage has a few handy methods that don't exist in MimeMessage
			message = new SMTPMessage(session);

			//HashMap containing the headers added for us by JavaMail, which we don't want to add on our
			//own further down as they would then be duplicated. Use upper-cae for everything
			HashMap excludeHeaders = new HashMap();
			excludeHeaders.put(StandardMailHeaders.SUBJECT.toUpperCase(), StandardMailHeaders.SUBJECT);
			excludeHeaders.put(StandardMailHeaders.TO.toUpperCase(), StandardMailHeaders.TO);
			excludeHeaders.put(StandardMailHeaders.CC.toUpperCase(), StandardMailHeaders.CC);
			excludeHeaders.put(StandardMailHeaders.BCC.toUpperCase(), StandardMailHeaders.BCC);

			if (!from.isEmpty()) {
				message.setFrom(new InternetAddress(mail.getFrom().toString()));
				excludeHeaders.put(StandardMailHeaders.FROM.toUpperCase(),StandardMailHeaders.FROM);
			} else {
				//Set envelope from to empty, we want to keep the From header
				message.setEnvelopeFrom("<>");
			}

			MailAddress[] tos = mail.getTo();

			Address[][] toAddy = createMessageRecps(tos);

			// Down here fail for everybody            
			if (toAddy[0].length > 0) {
				message.addRecipients(
					javax.mail.Message.RecipientType.TO,
					toAddy[0]);
			}
			if (toAddy[1].length > 0) {
				message.addRecipients(
					javax.mail.Message.RecipientType.CC,
					toAddy[1]);
			}
			if (toAddy[2].length > 0) {
				message.addRecipients(
					javax.mail.Message.RecipientType.BCC,
					toAddy[2]);
			}

			// KAB: Changed to ignore To, CC, BCC and From (in addition to the already existing 
			// Subject. Otherwise JavaMail adds them again.
			// The standard headers are added when loading the Mail instance from the DATA stream
			// as they are also needed when delivering to the local mailboxes.
			//
			// This was the previous comment:
			// --------------------------------------
			// NEED TO EXTRACT MORE HEADERS FROM Mail object, and also add standard ones that
			// are the responsibility of the server.
			//

			String[] subjects = mail.getMailHeaders().getHeader("Subject");
			String subject =
				(subjects != null && subjects[0] != null) ? subjects[0] : null;
			message.setSubject(subject);
			byte[] bbody = mail.getBody();
			String sbody =
				(bbody != null && bbody.length > 0)
					? new String(bbody, "US-ASCII")
					: "";
			if (debug) {
				log.debug("text = " + sbody);
			}
			message.setText(sbody, "US-ASCII");

			MailHeaders headers = mail.getMailHeaders();
			Iterator headerLines = headers.getAllHeaderLines();
			while (headerLines.hasNext()) {
				String headerLine = headerLines.next().toString();
				// no guarantee of class?? may want type-safety here

				String[] nameValue = headerLine.split(": ");

				if (excludeHeaders.get(nameValue[0].toUpperCase()) == null) {
					//subject, to, cc, bcc and from set above
					message.addHeader(nameValue[0], nameValue[1]);
				}
			}

			
			//Wrapper around Transport.send() to deal with its crazy exception behaviour
			Sender sender = new Sender(transport, message, internetTos);
			return sender.sendMessage();
		}catch(Throwable t){
				log.error("Cannot send mail" + t.toString(), t);
				return SMTPResultImpl.getErrorForAll(to);
		}finally{
        	if (transport != null) {
	            try{
	                transport.close();
	            }catch(MessagingException ignore) {
	                    ignore.printStackTrace();
	            }
        	}
        }
	}		



	/**
	 * @param addresses
	 * @return
	 */
	private InternetAddress[][] createMessageRecps(MailAddress[] addresses)
		throws AddressException {
		InternetAddress[][] retval = new InternetAddress[3][];
		int to = 0;
		int cc = 0;
		int bcc = 0;
		int ito = 0;
		int icc = 0;
		int ibcc = 0;
		for (int idx = 0; idx < addresses.length; idx++) {
			EnvelopedAddress addy = (EnvelopedAddress) addresses[idx];
			if (addy.getType() == javax.mail.Message.RecipientType.TO) {
				to++;
			} else if (addy.getType() == javax.mail.Message.RecipientType.CC) {
				cc++;
			} else if (
				addy.getType() == javax.mail.Message.RecipientType.BCC) {
				bcc++;
			}
		}
		retval[0] = new InternetAddress[to];
		retval[1] = new InternetAddress[cc];
		retval[2] = new InternetAddress[bcc];
		for (int idx = 0; idx < addresses.length; idx++) {
			EnvelopedAddress addy = (EnvelopedAddress) addresses[idx];
			if (addy.getType() == javax.mail.Message.RecipientType.TO) {
				retval[0][ito] = new InternetAddress(addy.toString());
				ito++;
			} else if (addy.getType() == javax.mail.Message.RecipientType.CC) {
				retval[1][icc] = new InternetAddress(addy.toString());
				icc++;
			} else if (
				addy.getType() == javax.mail.Message.RecipientType.BCC) {
				retval[2][ibcc] = new InternetAddress(addy.toString());
				ibcc++;
			}
		}
		return retval;
	}

	public void testSend(
		String to,
		String from,
		String subject,
		String message) {
		try {
			String themail =
				"\r\n"
					+ "User-Agent: JBMail 0.1 STMPSender\r\n"
					+ "Date: Thu, 01 Jan 2004 17:41:09 -0500\r\n"
					+ "Subject: "
					+ subject
					+ "\r\n"
					+ "From: Andrew Oliver <"
					+ from
					+ ">\r\n"
					+ "To: <"
					+ to
					+ ">\r\n"
					+ "Message-ID: <BC1A0C35.20FF7%acoliver@localhost>\r\n"
					+ "Mime-version: 1.0\r\n"
					+ "Content-type: text/plain; charset=\"US-ASCII\"\r\n"
					+ "Content-transfer-encoding: 7bit\r\n"
					+ "\r\n"
					+ message
					+ "\r\n.\r\n";
			
			byte[] bmail = themail.getBytes();
			InputStream is = new ByteArrayInputStream(bmail);
			MailAddress mafrom = MailAddress.parseSMTPStyle("<" + from + ">");
			Mail mail =
				Mail.create(
					is,
					1000,
					mafrom,
					new MailAddress[] { MailAddress.parseSMTPStyle("<" + to + ">")},
					new SMTPProtocolInstance(new HashMap(), null));
			send(mail.getTo()[0], mail);
		} catch (LoopDetectedException e) {
			//Won't happen
		}
	}

	/* (non-Javadoc)
	 * @see org.jboss.mail.smtp.sender.SMTPSenderMBean#setCacheEnabled(boolean)
	 */
	public void setCacheEnabled(boolean enabled) {
		this.useCache = enabled;
	}

	/* (non-Javadoc)
	 * @see org.jboss.mail.smtp.sender.SMTPSenderMBean#setDNSServers(org.w3c.dom.Element)
	 */
	public void setDNSServers(Element servers) {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
	 * @see org.jboss.mail.ServiceProxyFactory#produceServiceProxy()
	 */
	public ServiceProxy produceServiceProxy() {
		return this;
	}

//	Wrapper around Transport.send() to deal with its crazy exception behaviour
  class Sender{
	  final static int RETRIES = 3;
	  Transport transport;
	  SMTPMessage message;

	  //Addresses we have not yet sent to
	  Address[] internetTos;
	
	  int nestingLevel;
	  Set failedAddrs = new HashSet();
	  Set succeededAddrs = new HashSet();
	
	
	  Sender(Transport transport, SMTPMessage message, InternetAddress[] internetTos){
		  this.transport = transport;
		  this.message = message;
		  this.internetTos = internetTos;
	  }
	
	  ArrayList sendMessage(){
		  recursiveSend();	
		
		  //Create an arraylist of failed and succeeded SMTPResults
		  ArrayList results = new ArrayList(failedAddrs.size() + succeededAddrs.size());
		
		  //These have failed
		  for (Iterator it = failedAddrs.iterator() ; it.hasNext() ; ){
			  InternetAddress address = (InternetAddress)it.next();
			  EnvelopedAddress addr = getEnvelopedAddress(address);
			  results.add(new SMTPResultImpl(addr, SMTPResult.STATUS_ERROR));
		  }
		
		  //Set the ones not sent yet to failed too
		  if (internetTos != null){
			  for (int i = 0 ; i < internetTos.length ; i++){
				  EnvelopedAddress addr = getEnvelopedAddress((InternetAddress)internetTos[i]);
				  results.add(new SMTPResultImpl(addr, SMTPResult.STATUS_ERROR));
			  }
		  }
				
		  //Yoohoo some of them worked		
		  for (Iterator it = succeededAddrs.iterator() ; it.hasNext() ; ){
			  InternetAddress address = (InternetAddress)it.next();
			  EnvelopedAddress addr = getEnvelopedAddress(address);
			  results.add(new SMTPResultImpl(addr, SMTPResult.STATUS_OK));
		  }
		
		  return results;
	  }
			  
	  /* All mails sent here are for remote delivery, so it should be
	   * safe to crate a remote EnvelopedAddress 
	   */
	  private EnvelopedAddress getEnvelopedAddress(InternetAddress ia){
		  MailAddress addr = MailAddress.parseSMTPStyle(ia.getAddress());
		  EnvelopedAddress ea = new EnvelopedAddress(addr);
		  ea.setLocal(false);
		  return ea;
	  }

	  private void recursiveSend(){
				
		  try {
			  if (nestingLevel < RETRIES && internetTos != null && internetTos.length > 0){
				  //Send to all addresses we have not yet sent
				  transport.sendMessage(message, internetTos);
				
				  //Set internetTos to be the good ones we have not yet sent (i.e. none since we have sent all)
				  addToSet(succeededAddrs, internetTos);	
				  internetTos = null;					
			  }
		  } catch (SendFailedException e) {
			  //If the addresses for the domain contained both invalid addresses and valid addresses,
			  //it seems that the valid addresses will not be sent either. We try resending the valid 
			  //addresses again here. After a few retries with the dodgy addresses removed the valid 
			  //ones should get sent
			  nestingLevel++;
			  addToSet(failedAddrs, e.getInvalidAddresses());
			  addToSet(succeededAddrs, e.getValidSentAddresses());

			  //Set internetTos to be the good ones we have not yet sent 
			  internetTos = e.getValidUnsentAddresses();
			  recursiveSend();						
		  }
		  catch (Exception e){
			log.error("Error sending mail", e);
		  }
	  }
	
	  private void addToSet(Set set, Address[] addresses){
		  if (addresses != null){
			  for (int i = 0 ; i < addresses.length ; i++){
				  set.add(addresses[i]);
			  }
		  }
	  }
	
  }
}

class MailAddressesByDomain {
	HashMap addressesByDomain = new HashMap();

	public MailAddressesByDomain(MailAddress[] addresses, MailAddress[] exclude) {
		for (int index = 0; index < addresses.length; index++) {
			EnvelopedAddress ea = new EnvelopedAddress(addresses[index]);
			
			if (!included(ea, exclude) && !ea.getLocal()) {
				String domain = ea.getDomain();
				Set addrs = (Set) addressesByDomain.get(domain);

				if (addrs == null) {
					addrs = new HashSet();
					addressesByDomain.put(domain, addrs);
				}

				addrs.add(ea);
			}
		}
	}

	/** Checks to see if the address is in the list of addresses
	* not efficient for large sets, but good for now.
	* @param address to check for
	* @param list to check in
	* @return inclusion state (yes or no)
	*/
   private boolean included(MailAddress address, MailAddress[] list) {
	   boolean retval=false;
	   for (int k = 0; k < list.length; k++) {
		   if (address.equals(list[k])) {
			   retval = true;
			   break;
		   }
	   }
	   return retval;
   }

	public Iterator getDomains() {
		return addressesByDomain.keySet().iterator();
	}

	public MailAddress[] getAddresses(String domain) {
		Set addrs = (Set) addressesByDomain.get(domain);

		if (addrs != null) {
			return (MailAddress[]) addrs.toArray(new MailAddress[addrs.size()]);
		} else {
			return new MailAddress[] {
			};
		}

	}

		
}

