/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.message;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import junit.framework.TestCase;

import org.jboss.mail.CharTerminatedInputStream;
import org.jboss.mail.DotStuffingInputStream;
import org.jboss.mail.ServerThread;
import org.jboss.mail.SizeLimitedInputStream;
import org.jboss.mail.smtp.SMTPConstants;
import org.jboss.mail.smtp.SMTPProtocolInstance;

/**
 * @author Kabir Khan
 */
public class TestMail extends TestCase{
	SMTPProtocolInstance smtpProtocolInstance;
	final static MailAddress FROM = MailAddress.parseSMTPStyle("sender@origin.com");
	final static MailAddress TO1 = MailAddress.parseSMTPStyle("receiver@destination.com");
	final static MailAddress TO2 = MailAddress.parseSMTPStyle("receiver2@destination.com");

	InetAddress remoteAddress;
	InetAddress localAddress;

    public static void main(String[] args) {
        junit.textui.TestRunner.run(TestMail.class);
    }

    public TestMail(String name){
        super(name);
    }

	/* (non-Javadoc)
	 * @see junit.framework.TestCase#setUp()
	 */
	protected void setUp() throws Exception {
		super.setUp();
		smtpProtocolInstance = new SMTPProtocolInstance(new HashMap(), createMap());

		remoteAddress 	= InetAddress.getByName("smtp.tiscali.com");
		localAddress	= InetAddress.getByName("smtp.jboss.org");

		smtpProtocolInstance.setState(ServerThread.STATE_CLIENT_ADDRESS, remoteAddress);
		smtpProtocolInstance.setState(ServerThread.STATE_LOCAL_ADDRESS, localAddress);
		smtpProtocolInstance.setState(SMTPConstants.HELO_DOMAIN, "kabkhan.somewhere.com");
	}

	/* (non-Javadoc)
	 * @see junit.framework.TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * create a Map of properties
	 */
	private Map createMap() {
		Map props = new HashMap();
		props.put(org.jboss.mail.smtp.SMTPConstants.SERVER_NAME,"localhost");
		props.put(org.jboss.mail.smtp.SMTPConstants.MAX_RECEIVED_HEADERS, new Long(100));
		props.put(org.jboss.mail.smtp.SMTPConstants.RECEIVED_HEADERS_THRESHOLD, new Long(20));
		props.put(org.jboss.mail.smtp.SMTPConstants.MAX_OWN_RECEIVED_HEADERS, new Long(5));        	
		
		return props;
	}

	public void testFullMail() throws Exception{
		MailAddress from = FROM;
		MailAddress[] to = {TO1, TO2};

		StringBuffer headers = getBasicMailHeaderBuffer(to);
		headers = addTraceInfoToHeaders(headers, from);
		headers = addFromToHeaders(headers, from);
		headers = addDateToHeaders(headers);

		Mail mail = createMail(headers, from, to, smtpProtocolInstance);


		assertNotNull("Mail object was null", mail);
		MailHeaders ih = mail.getMailHeaders();

		assertNotNull("From headers were null",
				ih.getHeader(StandardMailHeaders.FROM));
		assertNotNull("Date header was null",
				ih.getHeader(StandardMailHeaders.DATE));

		assertNotNull("Return-Path header was null",
				ih.getHeader(StandardMailHeaders.RETURN_PATH));

		Iterator it = ih.getAllHeaderLines();
		String header = it.next().toString();
		String expected = "Return-Path: " + FROM.toString();
		assertTrue("First header Expected: '" + expected + "'; got:'" + header + "'",
				expected.equals(header));

		header = it.next().toString();
		expected = getExpectedReceived();
		assertTrue("Second header Expected: '" + expected + "..." + "'; got:'" + header + "'",
				header.startsWith(expected));

		header = it.next().toString();
		expected = "Received: ";
		assertTrue("Third header expected to be Received line" + header,
				header.startsWith(expected));

		header = it.next().toString();
		expected = "Received: ";
		assertTrue("Fourth header expected to be Received line; got:" + header,
				header.startsWith(expected));

		header = it.next().toString();
		expected = "Received: ";
		assertTrue("There should be only three received lines; got:" + header,
				!header.startsWith(expected));
	}

	public String getExpectedReceived(){
		return "Received: from kabkhan.somewhere.com ("
			+ remoteAddress.getHostName() + " " + remoteAddress.getHostAddress()
			+ ") by localhost (" + localAddress.getHostAddress() + ")\r\n\twith SMTP id";
	}


	public void testMailMissingFromAndDateNotOriginator() throws Exception{
		MailAddress from = FROM;
		MailAddress[] to = {TO1, TO2};

		StringBuffer headers = getBasicMailHeaderBuffer(to);
		headers = addTraceInfoToHeaders(headers, from);

		//If we are not the originator we should not be adding Date and From headers
		Mail mail = createMail(headers, from, to, smtpProtocolInstance);

		assertNotNull("Mail object was null", mail);
		MailHeaders ih = mail.getMailHeaders();

		String[] header = ih.getHeader(StandardMailHeaders.FROM);
		assertNull("From headers were not null", header);

		header = ih.getHeader(StandardMailHeaders.DATE);
		assertNull("Date header was not null",  header);
	}

	public void testMailMissingFromAndDateOriginator() throws Exception{
		MailAddress from = FROM;
		MailAddress[] to = {TO1, TO2};

		StringBuffer headers = getBasicMailHeaderBuffer(to);
		headers = addTraceInfoToHeaders(headers, from);

		//If we are the originator we should be adding Date and From headers
		smtpProtocolInstance.setState(SMTPConstants.USER, "x");
		Mail mail = createMail(headers, from, to, smtpProtocolInstance);

		assertNotNull("Mail object was null", mail);
		MailHeaders ih = mail.getMailHeaders();

		String[] header = ih.getHeader(StandardMailHeaders.TO);
		assertNotNull("To header was null",
				header);
		assertTrue("Expected first To was: " + TO1.toString()
				+ " got " + header[0], header[0].equals(TO1 + ", " + TO2));


		header = ih.getHeader(StandardMailHeaders.FROM);
		assertNotNull("From headers were null",
				header);
		assertTrue("Expected From From was: " + FROM.toString()
				+ " got " + header[0], header[0].trim().equals(FROM.toString()));

		header = ih.getHeader(StandardMailHeaders.DATE);
		assertNotNull("Date header was null",
				header[0]);
	}

	public void testMailMissingTraceInfoNotOriginator() throws Exception{
		MailAddress from = FROM;
		MailAddress[] to = {TO1, TO2};

		StringBuffer headers = getBasicMailHeaderBuffer(to);

		//If we are the originator we should not be adding the Return-Path header,
		//but should be adding a Received line
		Mail mail = createMail(headers, from, to, smtpProtocolInstance);
		MailHeaders ih = mail.getMailHeaders();

		assertNull("Return-Path header was not null",
				ih.getHeader(StandardMailHeaders.RETURN_PATH));

		String[] receivedHeader =
				ih.getHeader(StandardMailHeaders.RECEIVED);
		assertNotNull("Received header was null", receivedHeader);
		assertEquals("Received header has wrong length",
				receivedHeader.length, 1);
	}

	public void testMailMissingTraceInfoOriginator() throws Exception{
		MailAddress from = FROM;
		MailAddress[] to = {TO1, TO2};

		StringBuffer headers = getBasicMailHeaderBuffer(to);

		//If we are the originator we should be adding the Return-Path header
		//as well as a Received line
		smtpProtocolInstance.setState(SMTPConstants.USER, "x");
		headers = getBasicMailHeaderBuffer(to);
		Mail mail = createMail(headers, from, to, smtpProtocolInstance);

		assertNotNull("Mail object was null", mail);
		MailHeaders ih = mail.getMailHeaders();

		String[] header = ih.getHeader(StandardMailHeaders.TO);
		assertNotNull("To header was null",
				header);
		assertTrue("Expected first To was: " + TO1.toString()
				+ " got " + header[0], header[0].equals(TO1 + ", " + TO2));


		header = ih.getHeader(StandardMailHeaders.DATE);
		assertNotNull("Date header was null",
				header[0]);

		String[] receivedHeader =
				ih.getHeader(StandardMailHeaders.RECEIVED);

		assertNotNull("Received header was null", receivedHeader);
		assertEquals("Received header has wrong length",
				receivedHeader.length, 1);

		assertNotNull("Return-Path header was null",
				ih.getHeader(StandardMailHeaders.RETURN_PATH));


		Iterator it = ih.getAllHeaderLines();
		String hdr = it.next().toString();
		String expected = "Return-Path: " + FROM.toString();
		assertTrue("First header Expected: '" + expected + "'; got:'" + hdr + "'",
				expected.equals(hdr));

		hdr = it.next().toString();
		expected = getExpectedReceived();
		assertTrue("Second header Expected: '" + expected + "..." + "'; got:'" + header + "'",
				hdr.startsWith(expected));

		hdr = it.next().toString();
		expected = "Received: ";
		assertTrue("There should be only three received lines; got:" + header,
				!hdr.startsWith(expected));
	}


	/** Adds the trace headers to the headers
	 * @param headers StringBuffer containing the headers
	 * @param from The sender
	 * @return StringBuffer containing the headers with trace info added
	 */
	private StringBuffer addTraceInfoToHeaders(StringBuffer headers, MailAddress from){
		headers.append("Return-Path: ");
		headers.append(from.toString());
		headers.append("\r\n");

		headers.append("Received: from mailgw1.tiscali.no (213.142.64.211) by cpmail.dk.tiscali.com (6.7.018)");
		headers.append("id 3FF9222C0005C83E for kabirkhan@tiscali.no; Fri, 16 Jan 2004 16:54:44 +0100");
		headers.append("\r\n");
		headers.append("Received: by mailgw1.tiscali.no (Postfix, from userid 0)");
		headers.append("id 4AAECB55F; Fri, 16 Jan 2004 16:55:08 +0100 (CET)");
		headers.append("\r\n");

		return headers;
	}

	/** Adds the From header to the headers
	 * @param headers StringBuffer containing the headers
	 * @param from The sender
	 * @return StringBuffer containing headers with the From header added
	 */
	private StringBuffer addFromToHeaders(StringBuffer headers, MailAddress from){
		headers.append("From: ");
		headers.append(from.toString());
		headers.append("\r\n");
		return headers;
	}


	/** Adds the Date header to the headers
	 * @param headers StringBuffer containing the headers
	 * @return StringBuffer containing headers with the Date header added
	 */
	private  StringBuffer addDateToHeaders(StringBuffer headers){
		headers.append("Date: Mon, 15 Mar 2004 15:09:20 -0500 (EST)");
		headers.append("\r\n");
		return headers;
	}

	/**
	 * @return A set of headers without From, To, Date or trace headers set.
	 */
	private StringBuffer getBasicMailHeaderBuffer(MailAddress[] to){
		StringBuffer headers = new StringBuffer();
		headers.append("Message-ID: <33490261.1079381360857.Javaheaders.root@nukes.jboss.org>");
		headers.append("\r\n");
		headers.append("Subject: [headers Services] - TODO: Script Runner");
		headers.append("\r\n");
		headers.append("To: ");
		for (int i = 0 ; i < to.length ; i++){
			if (i > 0){
				headers.append(",");
			}
			headers.append(" ");
			headers.append(to[i].toString());
		}
		headers.append("\r\n");


		return headers;
	}


	/** Takes a set of headers and creates a mail object with a standard body
	 * @param headers The headers
	 * @param from The sender
	 * @param to The receivers
	 * @param smtpProtocol Initialised SMTPProtocolInstance
	 * @return
	 */
	private Mail createMail(StringBuffer headers, MailAddress from, MailAddress[] to, SMTPProtocolInstance smtpProtocol)throws LoopDetectedException{
		headers.append("\r\n");
		headers.append("Hello\r\n");
		headers.append(".\r\n");

		byte[] bytes = headers.toString().getBytes();
		InputStream in = new BufferedInputStream(new ByteArrayInputStream(bytes));

		in = new CharTerminatedInputStream(in,
				org.jboss.mail.smtp.SMTPConstants.SMTPTerminator);
		in  = new SizeLimitedInputStream(in, 10000);
        in = new DotStuffingInputStream(in);

		Mail mail = Mail.create(in, 300, from, to, smtpProtocol);
		return mail;
	}
}
