/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.smtp;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.net.Socket;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import junit.framework.TestCase;

import org.jboss.mail.MBeanServerUtil;
import org.jboss.mail.MailListener;
import org.jboss.mail.Protocol;
import org.jboss.mail.ServiceProxy;
import org.jboss.mail.message.Message;
import org.jboss.mail.message.MailAddress;
import org.jboss.mail.userrepository.UserRepository;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;

/**
 * Tests the SMTPProtocol Object.  Basic testing is done here, more extensive testing is done
 * in the individual command handlers.  As the implementation progresses some scenario testing
 * should also be done, but might be best done in other tests to ensure manageability.
 * @author Andrew C. Oliver
 */
public class TestSMTPProtocol extends TestCase {
    SMTPProtocol protocolmbean;
    SMTPProtocolInstance protocol;

    public TestSMTPProtocol(String name){
        super(name);
    }

	public static void main(String[] args) {
		junit.textui.TestRunner.run(TestSMTPProtocol.class);
	}

	/*
	 * @see TestCase#setUp()
	 */
	protected void setUp() throws Exception {
		System.out.println("SMTPProtocol setup");
		MBeanServerUtil.configureMBeanServerFactory();
		MBeanServerUtil.registerStaticUserRepository();
		MBeanServerUtil.registerSysOutMailListener();
		protocolmbean = new SMTPProtocol();
        protocolmbean.setProperties(createProps());
        protocol = (SMTPProtocolInstance) protocolmbean.produceServiceProxy();

        System.out.println("SMTPProtocol setup done");
	}


	/**
     * @return
     */
    private Element createProps() {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        try {
            builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {

            e.printStackTrace();
        }
        DOMImplementation impl = builder.getDOMImplementation();

        DocumentType DOCTYPE = impl.createDocumentType("non", "non", "non");
        Document doc = impl.createDocument("non", "non", DOCTYPE);
        Element retval = doc.createElement("properties");
        Element property = getProperty(doc, SMTPConstants.SERVER_NAME, "localhost");
        retval.appendChild(property);
        property = getProperty(doc, SMTPConstants.AUTH_REQUIRED, "true");
        retval.appendChild(property);
        property = getProperty(doc, SMTPConstants.AUTH_METHODS, "AUTH LOGIN PLAIN");
        retval.appendChild(property);
        property = getProperty(doc, SMTPConstants.USER_REPOSITORY, "jboss.mail:type=MailServices,name=UserRepository");
        retval.appendChild(property);
        property = getProperty(doc, SMTPConstants.VERIFY_IDENTITY, "true");
        retval.appendChild(property);
        property = getProperty(doc, SMTPConstants.MAX_MSG_SIZE, "1000000");
        retval.appendChild(property);
        property = getProperty(doc, SMTPConstants.MAIL_LISTENERS, new String[]{"jboss.mail:type=MailServices,name=MailListener"});
        retval.appendChild(property);
        

        return retval;
    }
    
    /**
     * @param doc
     * @param string
     * @param strings
     * @return
     */
    private Element getProperty(Document doc, String keyval, String[] strings) {
        Element property = doc.createElement("property");
         Element key = doc.createElement("name");
         key.appendChild(doc.createTextNode(keyval));
         Element values = doc.createElement("values");
         for (int k = 0; k < strings.length; k++) {
            Element value = doc.createElement("value");
            value.appendChild(doc.createTextNode(strings[k]));
            values.appendChild(value);
         }         
         property.appendChild(key);
         property.appendChild(values);
         return property;

    }

    private Element getProperty(Document doc, String keyval, String val) {
        Element property = doc.createElement("property");
         Element key = doc.createElement("name");
         key.appendChild(doc.createTextNode(keyval));
         Element value = doc.createElement("value");
         value.appendChild(doc.createTextNode(val));  
         property.appendChild(key);
         property.appendChild(value);           
         return property;
    }

    /*
	 * @see TestCase#tearDown()
	 */
	protected void tearDown() throws Exception {
		super.tearDown();
	}


    public void testSetProperties() throws Exception {
        String servername = (String)protocol.getProperty(SMTPConstants.SERVER_NAME);
        boolean authreq = ((Boolean)protocol.getProperty(SMTPConstants.AUTH_REQUIRED)).booleanValue();
        String authmeth = (String)protocol.getProperty(SMTPConstants.AUTH_METHODS);
        UserRepository res = (UserRepository)protocol.getProperty(SMTPConstants.USER_REPOSITORY);
        String verify = (String)protocol.getProperty(SMTPConstants.VERIFY_IDENTITY);
        long maxmsg = protocol.getPropertyLong(SMTPConstants.MAX_MSG_SIZE);
        List maillist = (List)protocol.getProperty(SMTPConstants.MAIL_LISTENERS);
        
        assertEquals("localhost", servername);
        assertEquals(true, authreq);
        assertEquals("AUTH LOGIN PLAIN", authmeth);
        assertTrue(res instanceof ServiceProxy);
        assertTrue(res instanceof Serializable);
        assertEquals("true",verify);
        assertEquals(1000000,maxmsg);
        Iterator i = maillist.iterator();
        while (i.hasNext()) {
            Object listener = i.next();
            assertTrue(listener instanceof MailListener);
            assertTrue(listener instanceof Serializable);
            assertTrue(listener instanceof ServiceProxy);
        }
        
    }

    /**
     * An SMTP greeting should contain the "220 servername SMTP Server (softwareversion) ready"
     * string followed by a long date (not checked).  This verifies the greet method behaves as
     * expected.
     * @throws IOException
     */
	public void testGreet() throws IOException {
		System.out.println("in greet");
		String compare = "220 localhost SMTP Server (JBMAIL SMTP Server version 0.1) ready";
		ByteArrayOutputStream stream = new ByteArrayOutputStream();
		protocol.greet(stream);
		byte[] b = (stream.toByteArray());
		String s = new String(b);
		System.out.println(s);
		assertTrue("Expected "+compare+"\n got"+s, s.substring(0,compare.length()).equals(compare));
	}

    /**
     * Test to verify that a properly formed request is parsed into a request object properly
     * @throws IOException
     */
	public void testParseRequest() throws IOException {
		String request = "EHLO me\n";
		ByteArrayInputStream stream = new ByteArrayInputStream(request.getBytes());
		SMTPRequest r = (SMTPRequest) protocol.parseRequest(stream,createSocket());
		assertTrue("Expected EHLO got "+r.getCommand(),
		            request.substring(0,4).equals(r.getCommand()));
	}

    /**
     * tests that a response to a written command can be read off of the stream
     * @throws IOException
     */
	public void testReadResponse() throws IOException {
		String request = "Test response string\n";
		ByteArrayInputStream stream = new ByteArrayInputStream(("\r"+request).getBytes());
	    String input = protocol.readResponse(createRequest(null,stream));
	    assertTrue("Expected '"+request+"' got "+input, input.equals(request));
	}

	/*
	 * not done here.  instead we'll have a test for each handler. 
	 */
	/*public void testHandleRequest() {
		doELHO();		
	}*/
	
	public void testHandleAuth() {
		System.out.println("New handleauth method needed");
		/* this test is bogus
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		ByteArrayInputStream in = new ByteArrayInputStream("".getBytes());
		
		boolean auth = protocol.handleAuth(createMailAddr("<jondoe@localhost>"),
		                                   out, 
		                                   createRequest("AUTH",in),
		                                   protocol, 
		                                   new PrintWriter(out));
		assertTrue("expected user to be authorized ", auth);
		*/
		//TODO: test various ways for him to fail and succeed authroization
	}

    /**
     * does nothing at the moment... The only thing it shouldn't do is throw an exception.
     * Later we should log or something
     *
     */
	public void testHandleIOError() {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		protocol.handleIOError(bos);
		OutputStream ios = createInvalidOutputStream();
	}
	
	/**
	 * creates an invalid input stream to simulate a closed one.
	 */
	private OutputStream createInvalidOutputStream() {
		return new OutputStream() {
			public void close() throws IOException {
				super.close();
			}

			public void flush() throws IOException {
				throw new IOException("Invalid stream test (stream closed)");
			}

			public void write(byte[] arg0, int arg1, int arg2)
				throws IOException {
					throw new IOException("Invalid stream test (stream closed)");

			}

			public void write(byte[] arg0) throws IOException {
				throw new IOException("Invalid stream test (stream closed)");

			}

			public void write(int arg0) throws IOException {
				throw new IOException("Invalid stream test (stream closed)");

			}
		};
		
	}

	public Socket createSocket() {
		Socket s= new Socket() {
			
		};
		return s;
	}

	/**
	 * @param string
	 * @return
	 */
	private MailAddress createMailAddr(String string) {
		MailAddress addy = MailAddress.parseSMTPStyle(string);
		return addy;
	}

	

	/**
	 * @param stream
	 * @return
	 */
	private SMTPRequest createRequest(final String command, final ByteArrayInputStream stream) {
		SMTPRequest request = new SMTPRequest() {
			/* (non-Javadoc)
			 * @see org.jboss.mail.smtp.SMTPRequest#getCommand()
			 */
			public String getCommand() {
				return command;
			}

			/* (non-Javadoc)
			 * @see org.jboss.mail.smtp.SMTPRequest#arguments()
			 */
			public Iterator arguments() {
				return null;
			}

			/* (non-Javadoc)
			 * @see org.jboss.mail.Request#getProtocol()
			 */
			public Protocol getProtocol() {
				return null;
			}

			/* (non-Javadoc)
			 * @see org.jboss.mail.smtp.SMTPRequest#getInputStream()
			 */
			public InputStream getInputStream() {
				return stream;
			}

			/* (non-Javadoc)
			 * @see org.jboss.mail.Request#getRemoteAddr()
			 */
			public String getRemoteAddr() {
				return null;
			}

			public String[] getArguments() {
				// TODO Auto-generated method stub
				return null;
			}
		};		
		return request;
	}


	/**
	 * create a Map of properties
	 */
	private Map createMap() {
		Map props = new HashMap();
		props.put(org.jboss.mail.smtp.SMTPConstants.SERVER_NAME,"localhost");
		props.put(org.jboss.mail.smtp.SMTPConstants.AUTH_REQUIRED, "true");
		props.put(org.jboss.mail.smtp.SMTPConstants.STATUS, "ready");
		props.put(org.jboss.mail.smtp.SMTPConstants.AUTH_METHODS, "AUTH LOGIN PLAIN");
		props.put(org.jboss.mail.smtp.SMTPConstants.USER_REPOSITORY, createFakeUR());
		props.put(org.jboss.mail.smtp.SMTPConstants.MAX_MSG_SIZE, new Long(1000000));
		List lsnrs = new ArrayList();
		lsnrs.add(new ConcreteMailListener());
		props.put(org.jboss.mail.smtp.SMTPConstants.MAIL_LISTENERS,lsnrs);
		return props;
	}

	/**
	 * create a mock UserRepository that always says TRUE
	 */
	private UserRepository createFakeUR() {
		return new UserRepository() {
		   public boolean test(String user, String password) {
			  System.out.println("FAKE UR "+user+", "+password);
			  return true;
		   }

		   public boolean test(String user) {
			  System.out.println("FAKE UR "+user);
			  return true;
		   }
		};		
	}



}

class ConcreteMailListener implements MailListener {
   public Message send( Message msg ) {
      System.out.println("Concrete Mail Listener sending "+msg);
      //return true;
      return msg;
   }
}
