/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.smtp.handlers;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import junit.framework.TestCase;

import org.jboss.mail.MailListener;
import org.jboss.mail.Protocol;
import org.jboss.mail.message.Message;
import org.jboss.mail.smtp.SMTPProtocolInstance;
import org.jboss.mail.smtp.SMTPRequest;
import org.jboss.mail.smtp.SMTPResponse;
import org.jboss.mail.userrepository.UserRepository;

/**
 * base test for Handlers, gives us the basic setup
 * @author Andrew C. Oliver
 */
public abstract class HandlerBase extends TestCase {
    SMTPProtocolInstance protocol;
    SMTPHandler handler;
    
    public static final String newline = "\r\n";

    public HandlerBase(String name){
        super(name);
    }

    /*
     * @see TestCase#setUp()
     */
    protected void setUp() throws Exception {
        protocol = new SMTPProtocolInstance(SMTPHandlers.instance(),createMap());
    }
    
    public abstract void testHandleRequest() throws Exception;

    protected SMTPResponse handleRequest(OutputStream out, SMTPRequest request) 
    throws Exception {
    	return handler.handleRequest(out,request,protocol);
    }
    
    /**
     * testcases that require state can set it here
     * @param key name of the state variable
     * @param value of the state
     */
    protected void setProtocolState(String key, Object value) {
        protocol.setState(key, value);		
    
    }

	/**
	 * create a Map of properties
	 */
	private Map createMap() {
		Map props = new HashMap();
		props.put(org.jboss.mail.smtp.SMTPConstants.SERVER_NAME,"localhost");
		props.put(org.jboss.mail.smtp.SMTPConstants.AUTH_REQUIRED, "true");
		props.put(org.jboss.mail.smtp.SMTPConstants.STATUS, "ready");
		props.put(org.jboss.mail.smtp.SMTPConstants.AUTH_METHODS, "AUTH LOGIN PLAIN");
		props.put(org.jboss.mail.smtp.SMTPConstants.USER_REPOSITORY, createFakeUR());
		props.put(org.jboss.mail.smtp.SMTPConstants.MAX_MSG_SIZE, new Long(1000000));
		props.put(org.jboss.mail.smtp.SMTPConstants.MAX_RECEIVED_HEADERS, new Long(100));
		props.put(org.jboss.mail.smtp.SMTPConstants.RECEIVED_HEADERS_THRESHOLD, new Long(20));
		props.put(org.jboss.mail.smtp.SMTPConstants.MAX_OWN_RECEIVED_HEADERS, new Long(5));        	
		
		List lsnrs = new ArrayList();
		lsnrs.add(new ConcreteMailListener());
		props.put(org.jboss.mail.smtp.SMTPConstants.MAIL_LISTENERS,lsnrs);
		return props;
	}

	/**
	 * create a mock UserRepository that always says TRUE
	 */
	private UserRepository createFakeUR() {
		return new UserRepository() {
		   public boolean test(String user, String password) {
			  System.out.println("FAKE UR "+user+", "+password);
			  return true;
		   }

		   public boolean test(String user) {
			  System.out.println("FAKE UR "+user);
			  return true;
		   }
		};		
	}

	/**
	 * @param stream
	 * @return
	 */
	protected SMTPRequest createRequest(final String command, 
	                                    final ByteArrayInputStream stream,
	                                    final String[] parms) {
		SMTPRequest request = new SMTPRequest() {
			/* (non-Javadoc)
			 * @see org.jboss.mail.smtp.SMTPRequest#getCommand()
			 */
			public String getCommand() {
				return command;
			}
	
			/* (non-Javadoc)
			 * @see org.jboss.mail.smtp.SMTPRequest#arguments()
			 */
			public Iterator arguments() {
				List plist = new ArrayList(parms.length);
				for (int k = 0; k < parms.length;k++) {
					plist.add(parms[k]);
				}
				return plist.iterator();
			}
	
			/* (non-Javadoc)
			 * @see org.jboss.mail.Request#getProtocol()
			 */
			public Protocol getProtocol() {
				return null;
			}
	
			/* (non-Javadoc)
			 * @see org.jboss.mail.smtp.SMTPRequest#getInputStream()
			 */
			public InputStream getInputStream() {
				return stream;
			}
	
			/* (non-Javadoc)
			 * @see org.jboss.mail.Request#getRemoteAddr()
			 */
			public String getRemoteAddr() {
				return null;
			}

			public String[] getArguments() {
				// TODO Auto-generated method stub
				return null;
			}
		};		
		return request;
	}
}

class ConcreteMailListener implements MailListener {
   public Message send( Message msg ) {
       System.out.println("Concrete Mail Listener sending "+msg);
       //return true;
       return msg;
   }
}
