/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.mail.maillistener;

import java.util.ArrayList;
import java.util.Date;

import javax.jms.ObjectMessage;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.management.ObjectName;
import javax.naming.InitialContext;

import org.apache.log4j.Logger;
import org.jboss.mail.MailListener;
import org.jboss.mail.ServiceProxy;
import org.jboss.mail.ServiceProxyFactory;
import org.jboss.mail.domaingroup.DomainGroupMBean;
import org.jboss.mail.message.BouncedMail;
import org.jboss.mail.message.EnvelopedAddress;
import org.jboss.mail.message.Mail;
import org.jboss.mail.message.MailAddress;
import org.jboss.mail.message.Message;
import org.jboss.mail.message.MailRetryWrapper;
import org.jboss.mail.util.JBMailJMXUtil;
import org.jboss.system.ServiceMBeanSupport;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 * JMSMailListener - proof of concept code which posts mails to queues or topics.  It has both
 * onServer and offServer settings thus you can divide the processing up such that posts for users
 * in domains handled by this SMTP server are "onServer" and handled by one set of MDBeans and the
 * others are handled by the offServer queue and processed seperately.  If you don't want them
 * seperate, just set them both to the same queue.
 * 
 * Because JMS is asynchronous, the status cannot be returned, you must put it last in the chain.  It
 * will always say "yes I'm doing it"...
 *
 * @author Andrew C. Oliver
 */
public class JMSMailListener
    extends ServiceMBeanSupport
    implements JMSMailListenerMBean, ServiceProxyFactory, MailListener {

    private static final Logger jblog = Logger.getLogger(JMSMailListener.class);

    /**
     * queue or topic
     */
    private String destType;
    /**
     * onServer queue/topic name
     */
    private String onServer;
    /**
     * offServer queue/topic name
     */
    private String offServer;
    /*
     * domains which we will consider "local"
     */
    //private List domains;
    /**
     * Name of the DomainGroup MBean containing the domains to be considered as "local"
     */
    ObjectName domainGroup;
    /**
     * DomainGroup MBean containing the domains to be considered as "local"
     */
    DomainGroupMBean domainGroupMBean;
    /**
     * QCF instance from JNDI (null unless destType="queue")
     */
    private QueueConnectionFactory qcf;
    /**
     * TCF instance from JNDI (null unless destType="topic")
     */
    private TopicConnectionFactory tcf;
    /**
     * onServer queue (null unless destType="queue")
     */
    private Queue onServerQueue;
    /**
     * onServer topic (null unless destType="topic")
     */
    private Topic onServerTopic;
    /**
     * offServer queue (null unless destType="queue")
     */
    private Queue offServerQueue;
    /**
     * offServer topic (null unless destType="topic")
     */
    private Topic offServerTopic;

    private String queueConnectionFactoryName;
    private String topicConnectionFactoryName;

    private long[] localRetryIntervals = new long[0];
    private long[] remoteRetryIntervals = new long[0];

    public JMSMailListener(String queueConnectionFactoryName, String topicConnectionFactoryName) {
        this.queueConnectionFactoryName = queueConnectionFactoryName;
        this.topicConnectionFactoryName = topicConnectionFactoryName;
    }

    /* (non-Javadoc)
     * @see org.jboss.mail.ServiceProxyFactory#produceServiceProxy()
     */
    public ServiceProxy produceServiceProxy() {
        JMXMailListenerProxy proxy = new JMXMailListenerProxy(this.getServiceName());
        return proxy;
    }

    public void startService() throws Exception {
        InitialContext ctx = new InitialContext();
        if (destType.equals("topic")) {
          //TODO: make configurable
          tcf = (TopicConnectionFactory)ctx.lookup(topicConnectionFactoryName);
          onServerTopic = (Topic)ctx.lookup("topic/"+onServer);
          offServerTopic = (Topic)ctx.lookup("topic/"+offServer);
        } else {
          //TODO: make configurable
          qcf = (QueueConnectionFactory)ctx.lookup(queueConnectionFactoryName);
          onServerQueue = (Queue)ctx.lookup("queue/"+onServer);
          offServerQueue = (Queue)ctx.lookup("queue/"+offServer);
        }
    }

    public void stopService() throws Exception {

    }

   public void setDestinationType(String destType) {
        this.destType = destType;
    }

    public String getDestinationType() {
        return this.destType;
    }

    public void setOnServerPostDestination(String postDest) {
        this.onServer = postDest;
    }

    public void setOffServerPostDestination(String postDest) {
        this.offServer = postDest;
    }

    public String getOnServerPostDestination() {
        return onServer;
    }

    public String getOffServerPostDestination() {
        return offServer;
    }

    public void setLocalRetryTimes(Element localRetryTimes){
        NodeList nodes = localRetryTimes.getElementsByTagName("retryTime");
        
        int length = nodes.getLength();
        this.localRetryIntervals = new long[length];
        
        for (int i = 0; i < length ; i++) {
            Node node = nodes.item(i);
            String time = node.getFirstChild().getNodeValue();
            localRetryIntervals[i] = Long.parseLong(time);
        }
    }
    
    public Element getLocalRetryTimes(){
        //TODO: return xml representations of local retry times
        return null;
    }

    public void setRemoteRetryTimes(Element remoteRetryTimes){
        NodeList nodes = remoteRetryTimes.getElementsByTagName("retryTime");
        
        int length = nodes.getLength();
        this.remoteRetryIntervals = new long[length];
        
        for (int i = 0; i < length ; i++) {
            Node node = nodes.item(i);
            String time = node.getFirstChild().getNodeValue();
            remoteRetryIntervals [i] = Long.parseLong(time);
        }
    }

    public Element getRemoteRetryTimes(){
        //TODO: return xml representations of local retry times
        return null;
    }

    public void setDomainGroup(ObjectName domainGroup) {
        //TODO This has been copied from SMTPProtocol, move to common code?
        this.domainGroup = domainGroup;

        try {
            domainGroupMBean = (DomainGroupMBean) JBMailJMXUtil.getService(domainGroup);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException("could not set domain group due to JMX issues");
        }
    }

    public ObjectName getDomainGroup() {
        return domainGroup;
    }


    public String getDestinationForMailAddress(MailAddress ma) {
        String domain = ma.getDomain();

        if (domainGroupMBean.isInGroup(domain)){
            return onServer;
        }

        return offServer;
    }

    public Message send(Message msg) {
        try { 

            if (destType.equals("topic")) {
                sendMessageTopic(msg);
            }   else {
                sendMessageQueue(msg);
            }
        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException (e);
        }
        return msg;
    }

    /**
     * does the real work for sendMessage in the event of a "queue" destType
     * @param msg the message to be sent
     * @throws Exception if there is any problem (JMS related or otherwise)
     */
    private void sendMessageQueue(Message msg) throws Exception {
        Mail mail = (Mail)msg;
        
        PreparedMail preparedMail = prepareMail(mail);
        
        QueueConnection qc = qcf.createQueueConnection();
        QueueSession qs = qc.createQueueSession(true, 0);
        
        try{
            if (preparedMail != null){        
                if (preparedMail.getSendOnServer()){
                    putMessageOnQueue(qs, onServerQueue, preparedMail);
                }
    
                if (preparedMail.getSendOffServer()){
                    putMessageOnQueue(qs, offServerQueue, preparedMail);
                }
            }
        }finally{
            qs.commit();
        }
    }    

    private void putMessageOnQueue(QueueSession qs, Queue queue, PreparedMail preparedMail) throws Exception
    {
            QueueSender sender = qs.createSender(queue);
            
            log.info("PUT MESSAGE ON QUEUE:" + preparedMail.getMail());
            
            ObjectMessage om = qs.createObjectMessage(preparedMail.getMail());
            long secondsDelay = preparedMail.getRetrySeconds();
            
            if (secondsDelay > 0){
                long deliveryTime = new Date().getTime() + (secondsDelay * 1000); 
                //Note: will only work in JBoss >= 3.2.2 with the JBoss JMS provider
                om.setLongProperty("JMS_JBOSS_SCHEDULED_DELIVERY", deliveryTime);            
            }
            sender.send(om);
            sender.close();
    }


    /**
     * does the real work for sendMessage in the event of a "topic" destType
     * @param msg the message to be sent
     * @throws Exception if there is any problem (JMS related or otherwise)
     */
    private void sendMessageTopic(Message msg) throws Exception {
        Mail mail = (Mail)msg;
        
        PreparedMail preparedMail = prepareMail(mail);
        
        TopicConnection tc = tcf.createTopicConnection();
        TopicSession ts = tc.createTopicSession(false, 0);
        
        try{
            if (preparedMail != null){        
                if (preparedMail.getSendOnServer()){
                    putMessageOnTopic(ts, onServerTopic, preparedMail);
                }
    
                if (preparedMail.getSendOffServer()){
                    putMessageOnTopic(ts, offServerTopic, preparedMail);
                }
            }
        }finally{
            ts.commit();
        }
    }


    private void putMessageOnTopic(TopicSession ts, Topic topic, PreparedMail preparedMail) throws Exception
    {
        TopicPublisher pub = ts.createPublisher(topic);
        ObjectMessage om = ts.createObjectMessage(preparedMail.getMail());        
        long secondsDelay = preparedMail.getRetrySeconds();
             
        if (secondsDelay > 0){
            long deliveryTime = new Date().getTime() + (secondsDelay * 1000); 
            //Note: will only work in JBoss >= 3.2.2 with the JBoss JMS provider
            om.setLongProperty("JMS_JBOSS_SCHEDULED_DELIVERY", deliveryTime);            
        }
        
        pub.publish(om);
        pub.close();
    }


    /** Prepares the mail for sending by figuring out things like the retry count,
     * the time to retry, what queue to put the message on etc. If too many retries
     * have taken place a bounce message is generated for the correct queue. 
     * @param mail The mail to be sent
     * @return A PreparedMail structure containing the mail and other useful info. Null if we should not send.
     */
    private PreparedMail prepareMail(Mail mail){
        if (mail instanceof MailRetryWrapper){
            //MailRetryWrapper are created by the queues if a mail fails, hence this is 
            //a retried mail
            return prepareResentMail((MailRetryWrapper)mail);
        }
        else{
            //MailRetryWrapper are created by the queues if a mail fails, hence this is 
            //a new mail
            return prepareNewMail(mail);
        }        
    }
    
    /** Prepares the mail for retry by figuring out things like the retry count,
     * the time to retry, what queue to put the message on etc. If too many retries
     * have taken place a bounce message is generated for the correct queue. 
     * @param mail The mail to be sent
     * @return A PreparedMail structure containing the mail and other useful info. Null if we should not send.
     */
    private PreparedMail prepareResentMail(MailRetryWrapper mail){
        //The message is being retried
        //We have already been through all the addresses and need to look at the extra information in
        //the MailWrapper to see what to do with the mail
        
        boolean mailIsLocal = (mail.getRedeliveryDestination() == MailRetryWrapper.REDELIVERY_LOCAL);
            
        //Get the interval to wait for a retry
        int retryNumber = mail.getRetryNumber();
        try{
            PreparedMail preparedMail = new PreparedMail(mail);
            
            if (mailIsLocal){
                preparedMail.setRetrySeconds(localRetryIntervals[retryNumber]);
                preparedMail.setSendOnServer(true);
            } else{
                preparedMail.setRetrySeconds(remoteRetryIntervals[retryNumber]);
                preparedMail.setSendOffServer(true);
            }
            mail.increaseRetries();
            return preparedMail;
            
        }catch(ArrayIndexOutOfBoundsException e){
            //We have exhausted the number of retries, generate a bounce reply to 
            //the receivers who have not been sent to
            return prepareBounceMessageForMail(mail, mailIsLocal);    
        }
    }
    
    /** Prepares a bounce message. 
     * @param mail The mail to be sent
     * @return A PreparedMail structure containing the mail and other useful info. Null if 
     * we should not send (i.e. this message is a bounce mail itself)
     */
    private PreparedMail prepareBounceMessageForMail(MailRetryWrapper mail, boolean mailIsLocal){
        if (!mail.getFrom().isEmpty() ){
                
            ArrayList failed = new ArrayList();
                    
            //If this message was meant for local delivery, only failed local ones should be reported and vice versa
            MailAddress[] rcpts = mail.getTo();
            for (int i = 0 ; i < rcpts.length ; i++){
                // MIKEA: rcpts[i] is a MailAddress, NOT an EnvelopedAddress!
                 boolean local = ((EnvelopedAddress)rcpts[i]).getLocal(); // CLASSCAST EXCEPTION!!!!
                        
                 if (local == mailIsLocal){
                    failed.add(rcpts[i]);
                 }
            }
                    
            if (failed.size() > 0){
                MailAddress[] failedaddrs = (MailAddress[])failed.toArray(new MailAddress[failed.size()]);
                BouncedMail bouncedMail = new BouncedMail(mail, failedaddrs);
                PreparedMail preparedMail = new PreparedMail(bouncedMail);
                preparedMail = determineDestinationsInMail(preparedMail);
                return preparedMail;
            }
        }
        
        return null;
    }
    
    private PreparedMail prepareNewMail(Mail mail){
        PreparedMail preparedMail = new PreparedMail(mail);
        return determineDestinationsInMail(preparedMail);
    }
    
    private PreparedMail determineDestinationsInMail(PreparedMail preparedMail){
        MailAddress[] rcpts = preparedMail.getMail().getTo();
        //Go through and modify the addresses to be on or off server and determine
        //the queues to be used
        for (int k = 0; k < rcpts.length ; k++) {
           String dest = getDestinationForMailAddress(rcpts[k]);
            EnvelopedAddress r = (EnvelopedAddress)rcpts[k];
            if (dest.equals(onServer)) {
                r.setLocal(true);
                preparedMail.setSendOnServer(true);
            } else if (dest.equals(offServer)) {
                r.setLocal(false);
                preparedMail.setSendOffServer(true);
            }
        }
        
        return preparedMail;
    }

}

/**
 * Class to hold data passed around within JMSMailListener 
 * @author Kabir Khan
 *
 */
class PreparedMail
{
    /** Should mail be sent on server */
    private boolean sendOnServer;
    /** Should mail be sent off server */
    private boolean sendOffServer;
    /** Mail to be sent */
    private Mail mail;    
    /** Delay before JMS send queue */
    private long retrySeconds = 0;
    
    PreparedMail(Mail mail){
        this.mail = mail;
    }

    public Mail getMail() {
        return mail;
    }

    public long getRetrySeconds() {
        return retrySeconds;
    }

    public boolean getSendOffServer() {
        return sendOffServer;
    }

    public boolean getSendOnServer() {
        return sendOnServer;
    }

    public void setRetrySeconds(long l) {
        retrySeconds = l;
    }

    public void setSendOffServer(boolean b) {
        sendOffServer = b;
    }

    public void setSendOnServer(boolean b) {
        sendOnServer = b;
    }

}
